/* -*- Mode: C; tab-width: 4 -*-
   renderer.c -- font renderers.
   Created: Chris Toshok <toshok@hungry.com>, 23-Sep-1997
*/
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "yfonts_p.h"
#include "ytypes.h"
#include "ydll.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <assert.h>

typedef YFontRenderer* (*YFontRenderer_getFunc)();

#define MAX_FONT_RENDERERS 200
YFontRenderer *font_renderers[ MAX_FONT_RENDERERS ];
int num_font_renderers;

/* Return true if file is dynamic library. Only skips directories */
static int
yhw_dynalib(
#if defined(linux) || defined(hpux)
			const
#endif
			struct dirent *file)
{
  int len = strlen(file->d_name);

  if (len > 3
	  && file->d_name[len - 3] == '.'
	  && file->d_name[len - 2] == 's'
	  && file->d_name[len - 1] == 'o')
	return 1;
  else
	return 0;
}

static char **
y_generateRendererPaths()
{
  char *dir_path = strdup(FONT_RENDERER_PATH "/");
  struct dirent **namelist;
  int i;
  int num_entries;
  char **renderer_paths;

  num_entries = scandir(dir_path, &namelist, yhw_dynalib, alphasort);

  if (num_entries <= 0)
    return NULL;
  
  renderer_paths = (char**)calloc(num_entries + 1, sizeof(char*));
  for (i = 0; i < num_entries; i ++)
    {
      renderer_paths[i] = (char*)calloc(strlen(dir_path) 
										+ strlen(namelist[i]->d_name) + 1,
										sizeof(char));
      strcpy(renderer_paths[i], dir_path);
      strcat(renderer_paths[i], namelist[i]->d_name);
    }

  for (i = 0; i < num_entries; i ++)
	free(namelist[i]);

  free(namelist);
  free(dir_path);

  return renderer_paths;
}


YFontRenderer *Y_GetFontRendererByName (char *name)
{
  int i;

  for (i=0; i<num_font_renderers; i++)
	{
	  const char *i_name = YGetFontRendererName (font_renderers[ i ]);
	  printf ("  matching '%s' against '%s'\n", name, i_name);
	  if (strcmp (name, i_name) == 0)
		return font_renderers[ i ];
	}

  fprintf (stderr, "Can't find font renderer with name '%s'\n", name);
  return NULL;
}


void
Y_registerFontRenderers()
{
  char **renderers = y_generateRendererPaths();
  char *renderer;
  int i = 0;
  YFontRenderer *theRenderer;

  assert(NULL != renderers);
  num_font_renderers = 0;

  while ((renderer = renderers[i++]) != NULL)
	{
	  YDLL *renderer_lib;
	  YFontRenderer_getFunc get_func;

	  if (num_font_renderers == MAX_FONT_RENDERERS)
		{
		  fprintf (stderr, "MAX_FONT_RENDERERS reached...\n");
		  break;
		}

	  renderer_lib = YDLL_LoadLibrary(renderer);
	  get_func = (YFontRenderer_getFunc)
		YDLL_LookupSymbol(renderer_lib, "YFontRenderer_get");

	  if (!renderer_lib || !get_func)
		continue;

	  theRenderer = (*get_func)();
	  printf ("+ Found Font Renderer '%s', version '%s'\n",
			  YGetFontRendererName(theRenderer),
			  YGetFontRendererVersion(theRenderer));

	  font_renderers[ num_font_renderers ++ ] = theRenderer;

	  /*current_font_renderer = theRenderer;*/
	}
}

const char *
YGetFontRendererName(YFontRenderer *renderer)
{
  return (*renderer->_vtable.getName)(renderer);
}

const char *
YGetFontRendererVendor(YFontRenderer *renderer)
{
  return (*renderer->_vtable.getVendor)(renderer);
}

const char *
YGetFontRendererVersion(YFontRenderer *renderer)
{
  return (*renderer->_vtable.getVersion)(renderer);
}


int
YFontRenderer_countFonts(YFontRenderer *renderer)
{
  if (renderer->_vtable.countFonts)
	return (*renderer->_vtable.countFonts)(renderer);
  else
	return 0;
}

int
YFontRenderer_listFonts(YFontRenderer *renderer,
						const char ***font_specs)
{
  if (renderer->_vtable.listFonts)
	return (*renderer->_vtable.listFonts)(renderer, font_specs);
  else
	return 0;
}

YFont *
YFontRenderer_loadFont(YFontRenderer *renderer,
					   const char *spec)
{
  return (*renderer->_vtable.loadFont)(renderer, spec);
}

void
YFontRenderer_renderCharacter(YFontRenderer *renderer,
							  char *string, int length,
							  YGC *gc, YPoint *start_at)
{
  (*renderer->_vtable.renderCharacter)(renderer, string, length,
									  gc, start_at);
}

int 
YFontRenderer_maxAscent(YFontRenderer *renderer, YFont *font)
{
  return (*renderer->_vtable.maxAscent)(renderer, font);
}

int 
YFontRenderer_maxDescent(YFontRenderer *renderer, YFont *font)
{
  return (*renderer->_vtable.maxDescent)(renderer, font);
}

int 
YFontRenderer_maxWidth(YFontRenderer *renderer, YFont *font)
{
  return (*renderer->_vtable.maxWidth)(renderer, font);
}

int
YFontRenderer_advancement(YFontRenderer *renderer,
						  YFont *font,
						  char character)
{
  return (*renderer->_vtable.advancement)(renderer,
										  font,
										  character);
}

int 
YFontRenderer_stringWidth(YFontRenderer *renderer,
						  YFont *font,
						  char *string, 
						  int length)
{
  return (*renderer->_vtable.stringWidth)(renderer,
										  font,
										 string, length);
}

void
YFontRenderer_destroy(YFontRenderer *renderer)
{
  (*renderer->_vtable.destroy)(renderer);

  free(renderer);
}
