/*--------------------------------------------------------------------------
  ----- File:        t1base.c 
  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
  ----- Date:        11/30/97
  ----- Description: This file is part of the t1-library. It contains basic
                     basic routines to initialize the data structures used
		     by the t1-library.
  ----- Copyright:   This program is copyright (c) Rainer Menzner, 1996. 
                     Everyone is allowed to use the library and to make
		     changes to the source code, provided that the authors
		     name is left where it appears and that changes are
		     metioned in some notice. The library must be
		     distributed including sources, you are not allowed to
		     take money for this.
		     The parseAFM-package is copyrighted by Adobe Systems
		     Inc.
		     The type1 rasterizer is copyrighted by IBM and the
		     X11-consortium.
  ----- Warranties:  Of course, there's NO WARRANTY OF ANY KIND :-)
  ----- Credits:     I want to thank IBM and the X11-consortium for making
                     their rasterizer freely available.
		     Also thanks to Piet Tutelaers for his ps2pk, from
		     which I took the rasterizer sources in a format
		     independ from X11.
                     Thanks to all people who make free software living!
--------------------------------------------------------------------------*/

#define T1BASE_C


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <ctype.h>


#include "../type1/ffilest.h" 
#include "../type1/types.h"
#include "parseAFM.h" 
#include "../type1/objects.h"
#include "../type1/spaces.h"
#include "../type1/util.h"
#include "../type1/fontfcn.h"
#include "../type1/regions.h"


#include "t1base.h"
#include "t1types.h"
#include "t1global.h"
#include "t1env.h"
#include "t1delete.h"


static int T1_pad=0;


/* This function is to be called by the user to initialize
   the font mechanism */
void *T1_InitLib( int log)
{
  int result;
  
  unsigned char *usershome;
  unsigned char *logfilepath;
  
  
  /* Assign pointer to global struct and set entry padding: */
  pFontBase=&FontBase;
  if (T1_pad)
    pFontBase->bitmap_pad=T1_pad;
  else
    pFontBase->bitmap_pad=T1GLYPH_PAD;


  /* Setup . to be the defaults for search paths and config files: */
  PFAB_ptr=pfab;
  AFM_ptr=afm;
  ENC_ptr=enc;
  FDB_ptr=fontdatabase;

  /* Open log-file: */
  if (log){
    /* Try first opening in current directory: */
    if ((t1lib_log_file=fopen( T1_LOG_FILE, "w"))==NULL){

      usershome=getenv("HOME");
      logfilepath=(unsigned char *)malloc((strlen(usershome) +
					   strlen(T1_LOG_FILE) + 1
					   ) * sizeof(unsigned char));
      strcpy( logfilepath, usershome);
      strcat( logfilepath, DIRECTORY_SEP);
      strcat( logfilepath, T1_LOG_FILE);
      if ((t1lib_log_file=fopen( logfilepath, "w"))==NULL){
	print_msg( "T1_InitLib()", "Warning: Unable to open log-file");
      }
      free( logfilepath);
    }
    T1_PrintLog( "T1_InitLib()", "Initialization started",
	       T1LOG_STATISTIC);
  }
  
  /* Check for representation of data in memory: */
  if ((pFontBase->endian=CheckEndian())){
    T1_PrintLog( "T1_InitLib()", "Using Big Endian data presentation (MSBFirst)",
		 T1LOG_STATISTIC);
    pFontBase->endian=1;
  }
  else{
    T1_PrintLog( "T1_InitLib()", "Using Little Endian data presentation (LSBFirst)",
		 T1LOG_STATISTIC);
    pFontBase->endian=0;
  }

  /* Save padding value in log file */
  sprintf( err_warn_msg_buf, "Glyphs are padded to %d bits",
	   pFontBase->bitmap_pad);
  T1_PrintLog( "T1_InitLib()", err_warn_msg_buf,
	       T1LOG_STATISTIC);
  

  if ((result=ScanConfigFile( &PFAB_ptr, &AFM_ptr, &ENC_ptr, &FDB_ptr))==0)
    T1_PrintLog( "T1_InitLib()", "Warning t1lib configuration file not found!",
	       T1LOG_WARNING);
  
  /* Default-values for device in case the user does not set them: */
  DeviceSpecifics.x_resolution=DEFAULT_RES;
  DeviceSpecifics.y_resolution=DEFAULT_RES;
  DeviceSpecifics.scale_x=DEFAULT_RES/BIGPOINTSPERINCH;
  DeviceSpecifics.scale_y=DEFAULT_RES/BIGPOINTSPERINCH;


  if ((result=scanFontDBase(FDB_ptr))==-1){
    print_msg( "T1_InitLib()", "Fatal error scanning Font Database File");
    return(NULL);
  }
  if (result == 0){
    print_msg( "T1_InitLib()", "No fonts from Font Database File found");
    return(NULL);
  }
  if (result < pFontBase->no_fonts){
    sprintf( err_warn_msg_buf, "%d fonts from %d fonts in Font Database File not found",
		 pFontBase->no_fonts - result, pFontBase->no_fonts);
    print_msg( "T1_InitLib()", err_warn_msg_buf);
    
  }

  /* Initialize the no_fonts... values */
  pFontBase->no_fonts_ini=pFontBase->no_fonts;
  pFontBase->no_fonts_limit=pFontBase->no_fonts;

  T1_Up=1;         /* System has been initialized ! */
  T1_PrintLog( "T1_InitLib()", "Initialization succesfully finished",
	     T1LOG_STATISTIC);

  return((void *) pFontBase);
}

/* scanFontDBase():
   - opens the file with the font definitions,
   - reads the number of fonts defined and saves this in FontBase,
   - allocates memory for all the filenames of the Type1 files
   - tests for .pfa und .pfb files and saves the name found
   - initializes an array that allows to acces these names by an
     index number, the font_ID
   - returns -1 on fatal error and the number of fonts located
     successfullly
   */
int scanFontDBase( unsigned char *filename)
{
  int fd;
  int filesize, i, j, k, l;
  int found=0;
  unsigned char **FontFileNameIDArray;
  unsigned char *filebuffer, *FontFileNameArray;
  unsigned char *sort_ptr;
  
  
  if ((fd=open( filename, O_RDONLY))<3){
    print_msg( "scanFontDBase()", "Font Database File not found");
    return(-1);
  }
  
  /* Get the file size */
  filesize=lseek( fd, 0, 2);
  /* Reset fileposition to start */
  lseek (fd, 0, 0); 

  if ((filebuffer=(unsigned char *)malloc(filesize*sizeof(unsigned char)
					  )) == NULL){
    print_msg( "scanFontDBase()", "Couldn't allocate memory for loading font database file");
    return(-1);
  }
    
  /* this is a safe choose, will later be realloc'ed: */
  if ((FontFileNameArray=(unsigned char *)malloc(filesize*sizeof(unsigned char)
						 )) == NULL){
    print_msg( "scanFontDBase()", "Couldn't allocate memory for font filename array");
    return(-1);
  }
  sort_ptr=(unsigned char *)FontFileNameArray;
  i=read( fd, filebuffer, filesize);
  close(fd);   /* Close font database file */
   
  i=j=0;
  
  while (i<filesize){
    if (filebuffer[i]=='\n'){ /* We are at the end of line */
      if (j==0){  /* Read the first line as the number of fonts */
	filebuffer[i]=0;
	sscanf( &filebuffer[0], "%d", &FontBase.no_fonts);
	filebuffer[i]='\n';  /* Because it gives a better feeling */
	if ((FontFileNameIDArray=(unsigned char **)
	     malloc(FontBase.no_fonts*sizeof(char *))) == NULL){
	  print_msg( "scanFontDBase()", "Couldn't allocate memory for filename index array");
	  return(-1);
	}
	/* Store address of ID-array */
	FontBase.pFontFileNameIDArray=FontFileNameIDArray;
	/* Allocate memory for 'no_fonts' structures: */ 
	if ((FontBase.pFontArray=(FONTPRIVATE *)
	     calloc( FontBase.no_fonts, sizeof(FONTPRIVATE))) == NULL){
	  print_msg( "scanFontDBase()", "Failed to allocate memory for FONTPRIVATE-area");
	  return(-1);
	}
      }
      else{       /* We are in the second or higher line */
	k=i;
	while (isspace(filebuffer[k])){
	  k--;
	}
	/* We are at the last printable character of a line; let's 
	   step back (over the [afm/sfm]) to the period and replace it
	   by an ASCII 0 */
	while (filebuffer[k]!='.'){
	  k--;
	}
	filebuffer[k]=0; /* termination for string reading */
	while (!isspace(filebuffer[k])){
	  k--;
	}
	sscanf( &filebuffer[k+1], "%s", &linebuf[0]);
	/* We print error string before testing because after the call
	   to test_for_t1_file() filename is substituted by an emty
	   string if the file was not found: */
	sprintf( err_warn_msg_buf, "Type 1 Font file %s.[pfa/pfb] not found (FontID=%d)",
		 linebuf, j-1);
	if (test_for_t1_file( &linebuf[0]))
	  print_msg( "scanFontDBase()", err_warn_msg_buf);
	else
	  found++;
	/* linebuf contains now the valid Type1 filename; let's
	   now copy this string into the appropriate place in the
	   FontFileNameArray: */
	FontFileNameIDArray[j-1]=sort_ptr; /* Save the pointer in the
					      ID-Array */
	l=0;
	while (linebuf[l]){
	  sort_ptr[l]=linebuf[l];
	  ++l;
	}
	sort_ptr[l]=0;  /* the terminating zero */
	sort_ptr +=l+1; /* sort_ptr points now to the next element */
      }
      j++; /* Advance line counter */
    }
    if (j>FontBase.no_fonts) /* to ignore especially white space at end */
      break;
    i++;   /* Step further in file position */
  }
  /* Let's now realloc the memory for FontFileNameArray */
  FontFileNameArray=(unsigned char *) realloc( FontFileNameArray,
					       (size_t) (sort_ptr-FontFileNameArray));
  l=0;
  /* Now FontFileNameIDArray[i] have to reinitialized to the reallocated values,
     since it may happen that the base address of the fontfilename changes
     by reallocation. Forgetting this got me 5 nice hours of Bug searching :-/ ,
     because the errors appeared only if the base address changes during
     reallocation. Moreover some special configuration in the font initialization
     file makes this error seemingly disappear */
  FontFileNameIDArray[0]=FontFileNameArray;
  for (j=1;j<FontBase.no_fonts;j++){
    while(FontFileNameArray[l]){
      l++;
    }
    l++;
    FontFileNameIDArray[j]=&FontFileNameArray[l];
  }
  /* Return the memory for file reading */
  free(filebuffer); 
  return( found);
}


/* T1_CloseLib(): Close the library and free all associated memory */
int T1_CloseLib( void)
{

  int i, j, error=0;

  if (T1_Up){
    for (i=pFontBase->no_fonts; i; i--){
      if ((j=T1_DeleteFont( i-1))){
	error=1;
	sprintf( err_warn_msg_buf, "T1_DeleteFont() returned %d for Font %d",
		 j, i-1);
	T1_PrintLog( "T1_CloseLib()", err_warn_msg_buf, T1LOG_ERROR);
      }
    }
    /* Free filenamestrings area: */
    if (pFontBase->pFontFileNameIDArray[0])
      free( pFontBase->pFontFileNameIDArray[0]);
    else
      error=1;
    /* Free pointer area */
    if (pFontBase->pFontFileNameIDArray)
      free( pFontBase->pFontFileNameIDArray);
    else
      error=1;
    /* Free the FONTPRIVATE area */
    if (pFontBase->pFontArray)
      free( pFontBase->pFontArray);
    else
      error=1;
    
    /* Indicate Library is no more initialized */
    pFontBase=NULL;
    T1_Up=0;
    T1_PrintLog( "T1_CloseLib()", "Library closed", T1LOG_STATISTIC);
    fclose(t1lib_log_file);
    
  }
  return( error);
}



/* This function prints a message to stderr and places an entry in the log
   file */
void print_msg( unsigned char *func_ident, unsigned char *msg_txt)
{
  fprintf(stderr,"%s: %s\n", func_ident, msg_txt);
  T1_PrintLog( func_ident, msg_txt, T1LOG_ERROR);
  return;
  
}


/* T1_PrintLog generates entries in the log file: */
void T1_PrintLog( unsigned char *func_ident, unsigned char *msg_txt, int level)
{
  static unsigned char levelid[4]={ 'E', 'W', 'S', 'D'};
  time_t s_clock, *tp;
  
  if (t1lib_log_file==NULL)
    return;
  if ((level>t1lib_log_level) || (level<1)){
    return;
  }
  else{
    tp=&s_clock;
    s_clock=time( tp);
    fprintf( t1lib_log_file, "(%c) (%.24s) %s: %s \n",
	     levelid[level-1], ctime(&s_clock), func_ident, msg_txt );
    fflush(t1lib_log_file); 
    return;
  }
}


/* T1_SetLogLevel(): Set the level of whihc a message must have so
   that it is printed into the logfile. This function may be called
   before T1_InitLib(). */
void T1_SetLogLevel( int level)
{
  if ((level>0) && (level<5))
    t1lib_log_level=level;
  return;
}



/* CheckForInit(): If no initialization of font mechanism has been
   done, return -1, indicating an error; also, write an error message to
   stderr. */
int CheckForInit(void)
{
  if(T1_Up)
    return(0);
  else{
    fprintf(stderr,
	    "CheckForInit: Font-mechanism has not been initialized\n");
    return(-1);
  }
}



/* CheckForFontID(): Checks the font mechanism concerning the specified
   ID. It returns:
                   0  if font belonging to FontID has not yet been loaded
		   1  if font belonging to FontID has already been loaded
		   -1 if FontID is an invalid specification
		   */
int CheckForFontID( int FontID)
{

  /* FontID is invalid */
  if ((FontID<0)||(FontID>(pFontBase->no_fonts - 1)))
    return(-1);
  
  if (pFontBase->pFontArray[FontID].pType1Data==NULL)
    return(0);     /* has not yet been loaded */
  else
    return(1);     /* has already been loaded */
}

 


/* test_for_t1_file returns zero if a file "name.pfa" or "name.pfb"
   was found. Else, -1 is returned. If succesful, buffer contains the
   found filename string */
int test_for_t1_file( unsigned char *buffer )
{
  int i=0;
  unsigned char *FullName;
  
  
  while (buffer[i]!=0){
    i++;
  }
  buffer[i]='.';
  buffer[i+1]='p';
  buffer[i+2]='f';
  buffer[i+4]=0;

  
  /* First case: A PostScript Font ASCII File */
  buffer[i+3]='a';
  if ((FullName=Env_GetCompletePath(buffer,PFAB_ptr))!=NULL){
    free(FullName);
    return(0);
  }
  /* Second case: A PostScript Font Binary File */
  buffer[i+3]='b';
  if ((FullName=Env_GetCompletePath(buffer,PFAB_ptr))!=NULL){
    free(FullName);
    return(0);
  }
  
  /* If we get here no file was found => Set buffer
     to an empty string and return -1 */
  
  buffer[0]=0;
  return(-1);
}


/* T1_GetFontFileName() returns a pointer to the filename of the font,
   associated with FontID. This filename does not contain a full path.
   */

unsigned char *T1_GetFontFileName( int FontID)
{

  
  if (CheckForInit())return(NULL);

  /* Check first for valid FontID */
  if ((FontID<0) || (FontID>FontBase.no_fonts))
    return(NULL);
    return((unsigned char *)FontBase.pFontFileNameIDArray[FontID]);
}



/* T1_Get_no_fonts(): Return the number of declared fonts */

int  T1_Get_no_fonts(void)
{
  if (CheckForInit())return(-1);
  return(FontBase.no_fonts);
}

	

/* T1_SetDeviceResolutions( x_res, y_res): Set the device's physical
   resolution in horizontal and vertical direction, mesured in DPI
   (Dots Per Inch). This should be done before the first font is
   loaded! */
int T1_SetDeviceResolutions( float x_res, float y_res)
{
  
  if (CheckForInit())return(-1);
  /* Save resolutions and calculate horizontal and vertical
     scale factors to map desired bp to device pixel */
  DeviceSpecifics.x_resolution=(float) x_res;
  DeviceSpecifics.y_resolution=(float) y_res;
  DeviceSpecifics.scale_x=(float)(((float)x_res)/BIGPOINTSPERINCH);
  DeviceSpecifics.scale_y=(float)(((float)y_res)/BIGPOINTSPERINCH);
  return(0);
}


/* T1_QueryX11Support(): Check at runtime to see if t1lib was compiled
   with X11 interface: */
int T1_QueryX11Support( void)
{
#ifndef T1LIB_NO_X11_SUPPORT
  return(1);
#else
  return(0);
#endif
}




/* int T1_CopyFont(): Copies the font associated with FontID to another
   location. The pointers to type1- , afm- and encoding data as well as
   the matrices remain completely untouched. However, size dependent data
   is not copied. The produced font is marked as a "logical" font.
   If no memory is available in the FONTPRIVATE-array, there's realloc'ed
   some more memory. The FontID which is assigned to the newly generated
   font is given as the return value, or < 0 if an error occurs. Also,
   the refcount entry of the source font is incremented by one.

   Return value -1: invalid FontID specified
                -2: source font is not a "physical" font
		-3: no memory for reallocation of FONTPRIVATEs
   */
int T1_CopyFont( int FontID)
{
  FONTPRIVATE *save_ptr;
  int new_ID;
  
  
  /* Check for a valid source font */
  if (CheckForFontID(FontID)!=1)
    return(-1);


  /* Check if the font in question is a "physical" font, otherwise it may
     not be copied */
  if (pFontBase->pFontArray[FontID].physical==0)
    return(-2);
  
  /* Check if free space for a new FONTPRIVATE is available; if not,
     realloc memory twice as large  */
  save_ptr=pFontBase->pFontArray;
  if (pFontBase->no_fonts==pFontBase->no_fonts_limit){
    pFontBase->pFontArray=(FONTPRIVATE *)realloc(pFontBase->pFontArray,
						 2 * pFontBase->no_fonts_limit
						 * sizeof(FONTPRIVATE));
    if (pFontBase->pFontArray==NULL){
      /* Restore pointer */
      pFontBase->pFontArray=save_ptr;
      return(-3);
    }
    pFontBase->no_fonts_limit *=2;
  }
  /* no_fonts-1 was the largest allowed font ID */
  new_ID=pFontBase->no_fonts;
  /* Copy FONTPRIVATE-structure: */
  pFontBase->pFontArray[new_ID]=pFontBase->pFontArray[FontID];
  /* (Re)Set some values explicitly, others remain untouched: */  
  pFontBase->pFontArray[new_ID].pFontSizeDeps=NULL;
  pFontBase->pFontArray[new_ID].physical=0;
  /* New font is logical --> indicate to which physical font it
     refers by setting refcount: */
  pFontBase->pFontArray[new_ID].refcount=FontID;
  
  /* Now the struct is setup; increment no_fonts by 1 because
     new_ID is a valid font specification from now on. */
  pFontBase->no_fonts++;
  /* Increment refcount in source font */
  pFontBase->pFontArray[FontID].refcount++;

  return(new_ID);
}



/* T1_SetBitmapPad(): Set the value to which bitmap-scanlines are padded.
   This has to be done before initialization because it is a very
   rudimentary operation.
   */
int T1_SetBitmapPad( int pad)
{
  if (T1_Up)
    /* Library is initialized --> return error */
    return(-1);
  else{
    switch (pad){
    case 8: 
      T1_pad=8;
      return(0);
    case 16:
      T1_pad=16;
      return(0);
    case 32:
      T1_pad=32;
      return(0);
    default:
      return(-1);
    }
  }
}



/* T1_GetBitmapPad(): Read the value to which scanlines of bitmap are
   padded. This can be done before or after initialization. */
int T1_GetBitmapPad( void)
{
  if (pFontBase) /* T1lib initialized --> return value from struct */
    return( pFontBase->bitmap_pad);
  else{
    if (T1_pad) 
      return(T1_pad);  /* pad is explicitly set --> return that value */
    else
      return( T1GLYPH_PAD); /* not expl. set --> return compilation default */
  }
}



/* bin_dump(): Print a binary dump of a byte, short and
   long variable (used for debug purposes only): */
void bin_dump_c(unsigned char value)
{
  int i,j;
  
  for (i=0;i<=7;i++){
    if ((j=((value)>>i)&0x01))
      printf("X");
    else
      printf(".");
  }
  printf(" ");

}

void bin_dump_s(unsigned short value)
{
  int i,j;

  if (CheckEndian()){
    for (i=8;i<=15;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
    for (i=0;i<=7;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
  }
  else{
    for (i=0;i<=15;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
  }
  printf(" ");
  
}

void bin_dump_l(unsigned long value)
{
  int i,j;
  
  if (CheckEndian()){
    for (i=24;i<=31;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
    for (i=16;i<=23;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
    for (i=8;i<=15;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
    for (i=0;i<=7;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
  }
  else{
    for (i=0;i<=31;i++){
      if ((j=((value)>>i)&0x01))
	printf("X");
      else
	printf(".");
    }
  }
  printf(" ");

}



/* CheckEndian(): Checks whether the current machine is of little or big
   endian architecture. This is important for concatenating bitmaps.
   Function returns 0 if LittleEndian and 1 if BigEndian representation
   is used  on the current hardware.
   */
int CheckEndian()
{
  unsigned char *charptr;

  /* Generate test value */
  unsigned short test=0x0001;

  /* Read out memory as unsigned char */
  charptr=(unsigned char *)(&test)+1;

  /* Return value will be 1 if Big- and 0 if Little Endian */
  return((int) *charptr);
  
}
