/*--------------------------------------------------------------------------
  ----- File:        t1enc.c 
  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
  ----- Date:        11/09/97
  ----- Description: This file is part of the t1-library. It contains
                     functions for reading encoding files at runtime.
  ----- Copyright:   This program is copyright (c) Rainer Menzner, 1996. 
                     Everyone is allowed to use the library and to make
		     changes to the source code, provided that the authors
		     name is left where it appears and that changes are
		     metioned in some notice. The library must be
		     distributed including sources, you are not allowed to
		     take money for this.
		     The parseAFM-package is copyrighted by Adobe Systems
		     Inc.
		     The type1 rasterizer is copyrighted by IBM and the
		     X11-consortium.
  ----- Warranties:  Of course, there's NO WARRANTY OF ANY KIND :-)
  ----- Credits:     I want to thank IBM and the X11-consortium for making
                     their rasterizer freely available.
		     Also thanks to Piet Tutelaers for his ps2pk, from
		     which I took the rasterizer sources in a format
		     independ from X11.
                     Thanks to all people who make free software living!
--------------------------------------------------------------------------*/
  
#define T1ENC_C


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>


#include "../type1/ffilest.h" 
#include "../type1/types.h"
#include "parseAFM.h" 
#include "../type1/objects.h"
#include "../type1/spaces.h"
#include "../type1/util.h"
#include "../type1/fontfcn.h"
#include "../type1/regions.h"


#include "t1types.h"
#include "t1extern.h"
#include "t1enc.h"
#include "t1env.h"
#include "t1base.h"
#include "t1finfo.h"


/* ScanEncodingFile(): Read an encoding file of an appropriate format
   and prepare the read data for usage with the type1 rasterizer, i.e.
   generate an array char *enc[256]. Return the pointer to the data area
   or NULL in case of an error.
   */
unsigned char **ScanEncodingFile( unsigned char *FileName) 
{
  
  unsigned char *linebuf;
  unsigned char *charnames;
  unsigned char save_char;
  unsigned char **encoding;
  
  FILE *enc_fp;
  int filesize, i, j, k, charname_count;

  if ((enc_fp=fopen( FileName,"r"))==NULL){
    return(NULL);  /* file could not be opened 
		      => no encoding read */
  }
  
  
  /* enc_fp points now to a (hopefully) valid encoding file */
  /* Get the file size */
  fseek( enc_fp, 0, SEEK_END);
  filesize=ftell(enc_fp);
  /* Reset fileposition to start */
  fseek( enc_fp, 0, SEEK_SET);
  
  if ((linebuf=(unsigned char *)calloc( filesize,
					sizeof(unsigned char)))==NULL)
    return(NULL);
  /* Allocate space for character names, assume the worst case and realloc
     later: */
  if ((charnames=(unsigned char *)calloc( filesize,
					  sizeof(unsigned char)))==NULL)
    return(NULL);
  fread((char *)linebuf, sizeof(char), filesize, enc_fp);
  fclose(enc_fp);

  i=0;
  
  while(i<filesize){
    j=i;     /* Save index of beginning of line */
    while (isspace(linebuf[i])==0)
      i++;
    save_char=linebuf[i];
    linebuf[i]=0;  /* replace ' ' by ASCII-0 */
    if (strcmp( "Encoding=", &linebuf[j])==0){
      /* set index to start of next line */
      if (save_char=='\n')
	i++;   
      else {
	while (linebuf[i]!='\n')
	  i++;
	i++;
      }
      /* keyword found => now, 256 lines should follow, each
	 specifying a character name and optionally some comment
	 to enhance readability: */
      break;
    }
    i++;
  }
  
  k=0;
  charname_count=0;
  
  while((i<filesize) && (charname_count<256)){
    j=i;     /* Save index of beginning of line */
    while (isspace(linebuf[i])==0)
      i++;
    save_char=linebuf[i];
    linebuf[i]=0;  /* replace whitespace by ASCII-0 */
    /* Copy Name to *char_names-array */
    while (linebuf[j])
      charnames[k++]=linebuf[j++];
    /* Append ASCII-0 */
    charnames[k++]=linebuf[j++];
    /* set index to start of next line */
    if (save_char=='\n')
      i++;   
    else {
      while (linebuf[i]!='\n')
	i++;
      i++;
    }
    /* Increment character counter */
    charname_count++;
  }

  free(linebuf);

  /* Check if exactly 256 characters have been defined, if not,
     return NULL: */
  if (charname_count!=256){
    free(charnames);
    return(NULL);
  }
  
  if ((encoding=(unsigned char **)malloc(256*sizeof(char **)))==NULL){
    free(charnames);
    return(NULL);
  }
  /* k should still contain the number of characters copied, so let's
     now realloc charnames */
  charnames=(unsigned char *)realloc( charnames, k*sizeof(unsigned char));
  /* Now initialize the array with the start-addresses of the character
     name strings */
  i=0;
  j=0;
  
  while (i<256){
    encoding[i]=&charnames[j];
    while (charnames[j])
      j++;
    j++;
    i++;
  }
  return(encoding);
}



/* T1_LoadEncoding(): Load an encoding file to have a new encoding
   available. If succesful, the pointer to the encoding array is
   returned. In case of an error, the return value is NULL.
   */
unsigned char **T1_LoadEncoding( unsigned char *FileName)
{
  unsigned char **Encoding;
  unsigned char *EncFileName;
  
  if ((EncFileName=Env_GetCompletePath( FileName, ENC_ptr))==NULL)
    return(NULL);
  Encoding=ScanEncodingFile(EncFileName);
  free(EncFileName);
  return(Encoding);
}

  

/* T1_DeleteEncoding() free a previously loaded encoding */
int T1_DeleteEncoding( unsigned char **encoding)
{
  if (encoding){
    /* First free character names memory */
    free( encoding[0]);
    /* then, free pointer array */
    free( encoding);
  }
  return(0);
  
}



/* T1_ReencodeFont(): Assign a new encoding to an existent font. This is
   only allowed if no size dependent data exists for the font in question.
   Moreover, the font must be loaded already since must get the position
   of the space-character. Function returns 0 if succesful, and -1 otherwise.
   */
int T1_ReencodeFont( int FontID, unsigned char **Encoding)
{
  int i, j, k;
  unsigned char *charname;
  PairKernData *pkd;
  KERN_ENTRY *kern_tbl;
  int char1, char2;
  
  
  /* First, check for valid font ID residing in memory: */
  if (CheckForFontID(FontID)!=1)
    return(-1);
    
  /* Second, check whether size-dependent data exists: */
  if (pFontBase->pFontArray[FontID].pFontSizeDeps != NULL)
    return(-1);

  pFontBase->pFontArray[FontID].pFontEnc=Encoding;

  /* We have to update the space_position-entry in the FONTPRIVATE. 
     If space is not found (not encoded), set it to -1: */
  pFontBase->pFontArray[FontID].space_position=-1;
  i=0;
  if (Encoding){ /* external encoding */
    while (i<256){
      if (strcmp( (unsigned char *)pFontBase->pFontArray[FontID].pFontEnc[i],
		  "space")==0){
	/* space found at position i: */
	pFontBase->pFontArray[FontID].space_position=i;
	break;
      }
      i++;
    }
  }
  else{ /* reencoding to internal encoding */
    while (i<256){
      if (strcmp( (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[i].data.arrayP,
		  "space")==0){
	/* space found at position i: */
	pFontBase->pFontArray[FontID].space_position=i;
	break;
      }
      i++;
    }
  }

  /* Now update afm index mapping: */
  if (pFontBase->pFontArray[FontID].pAFMData != NULL){
    for (i=0; i<256; i++)
      pFontBase->pFontArray[FontID].pAFMData->enc_map[i]=-1;
    for (i=0; i<256; i++){
      charname=T1_GetCharName( FontID, i);
      for ( j=0; j<pFontBase->pFontArray[FontID].pAFMData->numOfChars; j++){
	if (strcmp( charname,
		    pFontBase->pFontArray[FontID].pAFMData->cmi[j].name)==0){
	  pFontBase->pFontArray[FontID].pAFMData->enc_map[i]=j;
	}
      }
    }
    
    /* Update flags array and kerning table */
    /* Reset flags-table */
    memset( pFontBase->pFontArray[FontID].pAFMData->flags, 0, 256);
    /* Update kerning table */
    k=pFontBase->pFontArray[FontID].pAFMData->numOfPairs;
    kern_tbl=pFontBase->pFontArray[FontID].pAFMData->kern_tbl;
    pkd=pFontBase->pFontArray[FontID].pAFMData->pkd;
    j=0;
    for ( i=0; i<k; i++){
      if ((char1=T1_GetEncodingIndex( FontID, pkd[i].name1))==-1)
	/* pair is not relevant in current encoding */
	continue;
      if ((char2=T1_GetEncodingIndex( FontID, pkd[i].name2))==-1)
	/* pair is not relevant in current encoding */
	continue;
      /* Since we get here we have a relevant pair -->
	 Put char1 in higher byte and char2 in LSB:
	 */
      kern_tbl[j].chars=(char1 << 8) | char2;
      /* We only make use of horizontal kerning */
      kern_tbl[j].hkern=pkd[i].xamt;
      j++;
      /* Set the kerning flag for this char1: */
      pFontBase->pFontArray[FontID].pAFMData->flags[char1] |=KERN_FLAG;
    }
    /* Reset the remaining of the table in case there were
       irrelevant pairs: */
    for ( ; j<k; j++){
      kern_tbl[j].chars=0;
      kern_tbl[j].hkern=0;
    }
  }
  
  return(0);
}


