/*--------------------------------------------------------------------------
  ----- File:        t1env.c 
  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
  ----- Date:        11/25/97
  ----- Description: This file is part of the t1-library. It implements
                     the reading of a configuration file and path-searching
		     of type1-, afm- and encoding files.
  ----- Copyright:   This program is copyright (c) Rainer Menzner, 1996. 
                     Everyone is allowed to use the library and to make
		     changes to the source code, provided that the authors
		     name is left where it appears and that changes are
		     metioned in some notice. The library must be
		     distributed including sources, you are not allowed to
		     take money for this.
		     The parseAFM-package is copyrighted by Adobe Systems
		     Inc.
		     The type1 rasterizer is copyrighted by IBM and the
		     X11-consortium.
  ----- Warranties:  Of course, there's NO WARRANTY OF ANY KIND :-)
  ----- Credits:     I want to thank IBM and the X11-consortium for making
                     their rasterizer freely available.
		     Also thanks to Piet Tutelaers for his ps2pk, from
		     which I took the rasterizer sources in a format
		     independ from X11.
                     Thanks to all people who make free software living!
--------------------------------------------------------------------------*/
  

#define T1ENV_C


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>


#include "../type1/types.h"
#include "parseAFM.h" 
#include "../type1/objects.h" 
#include "../type1/spaces.h"  
#include "../type1/util.h" 
#include "../type1/fontfcn.h"
#include "../type1/fontmisc.h"

#include "t1types.h"
#include "t1extern.h"
#include "t1env.h"
#include "t1misc.h"
#include "t1base.h"


/* ScanConfigFile(): Read a configuration file and scan and save the
   environment strings used for searching pfa/pfb-, afm- and encoding
   files as well as the name of the font database file. */
int ScanConfigFile( unsigned char **pfabenv_ptr, 
		    unsigned char **afmenv_ptr,
		    unsigned char **encenv_ptr,
		    unsigned char **fontdatabase_ptr)
{
  
  unsigned char *environ;
  unsigned char *linebuf;
  unsigned char *usershome;
  unsigned char *cnffilepath;
  unsigned char *globalcnffilepath;
  
  
  
  FILE *cfg_fp;
  int filesize, i, j, k;
  
  
  /* First, get the string stored in the environment variable: */
  environ=getenv(ENV_CONF_STRING);


  if (environ==NULL){
    /* environment variable not set, try to open default file
       in user's home directory and afterwards global config file */
    usershome=getenv("HOME");
    cnffilepath=(unsigned char *)malloc((strlen(usershome) +
					 strlen(T1_CONFIGFILENAME) + 1
					 ) * sizeof(unsigned char));
    strcpy( cnffilepath, usershome);
    strcat( cnffilepath, DIRECTORY_SEP);
    strcat( cnffilepath, T1_CONFIGFILENAME);

    globalcnffilepath=(unsigned char*)malloc((strlen(GLOBAL_CONFIG_DIR) +
					      strlen(GLOBAL_CONFIG_FILE) + 1
					      ) * sizeof(unsigned char));
    strcpy( globalcnffilepath, GLOBAL_CONFIG_DIR);
    strcat( globalcnffilepath, DIRECTORY_SEP);
    strcat( globalcnffilepath, GLOBAL_CONFIG_FILE);

    if ((cfg_fp=fopen( cnffilepath, "r"))==NULL){
      sprintf( err_warn_msg_buf, "Could not open %s", cnffilepath);
      T1_PrintLog( "ScanConfigFile()", err_warn_msg_buf, T1LOG_STATISTIC);
      /* Try global config file */
      if ((cfg_fp=fopen( globalcnffilepath, "r"))==NULL){
	sprintf( err_warn_msg_buf, "Could not open %s", globalcnffilepath);
	T1_PrintLog( "ScanConfigFile()", err_warn_msg_buf, T1LOG_WARNING);
      }
      else{
	sprintf( err_warn_msg_buf, "Using %s as Configfile (global)",
		 cnffilepath);
	T1_PrintLog( "ScanConfigFile()", err_warn_msg_buf, T1LOG_STATISTIC);
      }
    }
    else{
      sprintf( err_warn_msg_buf, "Using %s as Configfile (user's)",
	       cnffilepath);
      T1_PrintLog( "ScanConfigFile()", err_warn_msg_buf, T1LOG_STATISTIC);
    }
    free( cnffilepath);
    free( globalcnffilepath);
    if (cfg_fp==NULL){
      T1_PrintLog( "ScanConfigFile()",
		   "Neither user's nor global Configfile has been found",
		   T1LOG_WARNING);
      return(0);
    }
  }
  else{
    /* open specified file for reading the configuration */
    if ((cfg_fp=fopen(environ,"r"))==NULL){
      T1_PrintLog( "ScanConfigFile()",
		   "Configfile as specified by Environment has not been found",
		   T1LOG_WARNING);
      return(0);  /* specified file could not be openend
		     => no config paths read */
    }
    else{
      sprintf( err_warn_msg_buf, "Using %s as Configfile (environment)",
	       environ);
      T1_PrintLog( "ScanConfigFile()", err_warn_msg_buf, T1LOG_STATISTIC);
    }
  }
  

  /* cfg_fp points now to a valid config file */
  /* Get the file size */
  fseek( cfg_fp, 0, SEEK_END);
  filesize=ftell(cfg_fp);
  /* Reset fileposition to start */
  fseek( cfg_fp, 0, SEEK_SET);
  
  if ((linebuf=(char *)calloc( filesize,
			       sizeof(unsigned char)))==NULL)
    return(-1);
  fread((char *)linebuf, sizeof(char), filesize, cfg_fp);
  fclose(cfg_fp);
  
  i=0;
  
  while(i<filesize){
    j=i;     /* Save index of beginning of line */
    while ((linebuf[i]!='=')&&(linebuf[i]!='\n'))
      i++;
    linebuf[i]=0;  /* replace '=' by ASCII-0 */
    if (strcmp( "ENCODING", &linebuf[j])==0){
      k=i+1;       /* points to assigned path string */
      while (linebuf[i] != '\n')
	i++;
      linebuf[i]=0;  /* replace linefeed by ASCII-0 */
      *encenv_ptr=&linebuf[k];
    }  
    if (strcmp( "TYPE1", &linebuf[j])==0){
      k=i+1;       /* points to assigned path string */
      while (linebuf[i] != '\n')
	i++;
      linebuf[i]=0;  /* replace linefeed by ASCII-0 */
      *pfabenv_ptr=&linebuf[k];
    }  
    if (strcmp( "AFM", &linebuf[j])==0){
      k=i+1;       /* points to assigned path string */
      while (linebuf[i] != '\n')
	i++;
      linebuf[i]=0;  /* replace linefeed by ASCII-0 */
      *afmenv_ptr=&linebuf[k];
    }  
    if (strcmp( "FONTDATABASE", &linebuf[j])==0){
      k=i+1;       /* points to assigned path string */
      while (!isspace(linebuf[i]))
	i++;
      if (linebuf[i]=='\n'){
	linebuf[i]=0;  /* replace linefeed by ASCII-0 */
      }
      else{
	linebuf[i]=0;  /* replace linefeed by ASCII-0 */
	/* Step to end of line */
	while (linebuf[i] != '\n')
	  i++;
	linebuf[i]=0;
      }
      *fontdatabase_ptr=&linebuf[k];
    }
    i++;
  }
  /*file should now be read in */
  return(i);
  
}



/* Env_GetCompletePath( ): Get a full path name from the file specified by
   argument 1 in the environment specified by argument 2. Return the pointer
   to the path string or NULL if no file was found.*/
unsigned char *Env_GetCompletePath( unsigned char *FileName,
				    unsigned char *env_ptr )
{
  struct stat filestats;    /* A structure where fileinfo is stored */
  int fnamelen, enamelen, i, j;
  unsigned char save_char;
  unsigned char *FullPathName;
  unsigned char *res_ptr;
  
  
  fnamelen=strlen(FileName);
  enamelen=strlen(env_ptr);
  i=0;
  while (i<enamelen){
    j=i;      /* Save start of current path string */
    while ((env_ptr[i]!=':')&&(env_ptr[i]!=0))
      i++;
    /* Save the character that indicated end of path */
    save_char=env_ptr[i];
    env_ptr[i]=0;   /* Set limit for this path element */
    FullPathName=(unsigned char *)malloc((i-j+2+fnamelen)*sizeof(char));
    /* Copy current path element: */
    res_ptr=strcpy( FullPathName, &env_ptr[j]);
    /* Add the directory separator: */
    res_ptr=strcat( FullPathName, DIRECTORY_SEP);
    /* And finally the filename: */
    res_ptr=strcat( FullPathName, FileName);
    
    /* Check for existence of the path: */
    if (!stat( FullPathName, &filestats)){
      /* restore the replaced character in order to leave environment
	 string unmodified */
      env_ptr[i]=save_char;
      return(FullPathName);
    }
    
    /* We didn't find the file --> try next path entry */
    free(FullPathName);
    /* restore the replaced character in order to leave environment string
       unmodified */
    env_ptr[i]=save_char;
    i++;
  }
  /* If we get here, no file was found at all, so return a NULL-pointer */
  return(NULL);
}


