/*--------------------------------------------------------------------------
  ----- File:        t1finfo.c 
  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
  ----- Date:        11/09/97
  ----- Description: This file is part of the t1-library. It contains
                     functions for accessing afm-data and some other
		     fontinformation data.
  ----- Copyright:   This program is copyright (c) Rainer Menzner, 1996. 
                     Everyone is allowed to use the library and to make
		     changes to the source code, provided that the authors
		     name is left where it appears and that changes are
		     metioned in some notice. The library must be
		     distributed including sources, you are not allowed to
		     take money for this.
		     The parseAFM-package is copyrighted by Adobe Systems
		     Inc.
		     The type1 rasterizer is copyrighted by IBM and the
		     X11-consortium.
  ----- Warranties:  Of course, there's NO WARRANTY OF ANY KIND :-)
  ----- Credits:     I want to thank IBM and the X11-consortium for making
                     their rasterizer freely available.
		     Also thanks to Piet Tutelaers for his ps2pk, from
		     which I took the rasterizer sources in a format
		     independent from X11.
                     Thanks to all people who make free software living!
--------------------------------------------------------------------------*/
  
#define T1FINFO_C


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "../type1/ffilest.h" 
#include "../type1/types.h"
#include "parseAFM.h" 
#include "../type1/objects.h"
#include "../type1/spaces.h"
#include "../type1/util.h"
#include "../type1/fontfcn.h"
#include "../type1/regions.h"


#include "t1types.h"
#include "t1extern.h"
#include "t1finfo.h"
#include "t1base.h"
#include "t1misc.h"
#include "t1set.h"
#include "t1load.h"



/* The following variable controls the computation of the bbox internal
   to T1_GetMetricsInfo(). Its influence may be overridden by the
   global variable ForceAFMBBox: */
static int ForceAFMBBoxInternal=0;


/* int T1_GetKerning(): This function returns the amount of kerning that
   is specified in the afm-file for the supplied character-pair. If an
   an extension has been applied to the font in question, this is taken
   into account.
   If for whatever reason there's no afm information available (that's not
   deadly), simply 0 is returned, indicating that no kerning should be used.
   The value returned is meant to be in character space coordinates. Thus,
   it must be transformed to be applicable in device space.
   A note more: This implementation implicates that the kerning pairs are
   alphabetically ordered in the afm-file. The function may not work
   correct if an afm-file includes kerning pairs not alphabetically
   sorted!
   */
int T1_GetKerning( int FontID, unsigned char char1, unsigned char char2)
{

  int i, found, kern;
  int chars=0;
  KERN_ENTRY *kern_tbl;
  

  /* Check whether font is loaded: */
  if (CheckForFontID(FontID)!=1){
    return(0);
  }
  
  if (pFontBase->pFontArray[FontID].pAFMData==NULL){
    return(0);  /* font is loaded, but no afm data available */
  }
  
  /* Check for char1 having kerning info */
  found=0;
  if ((pFontBase->pFontArray[FontID].pAFMData->flags[char1]) & KERN_FLAG){
    chars=(char1<<8) | char2;
    kern_tbl=pFontBase->pFontArray[FontID].pAFMData->kern_tbl;
    for ( i=0; i < pFontBase->pFontArray[FontID].pAFMData->numOfPairs; i++){
      if (kern_tbl[i].chars == chars){
	found=1;
	break;
      }
    } 
    if (found){
      kern=kern_tbl[i].hkern * pFontBase->pFontArray[FontID].extend;
      return(kern);
    }
    else{
      return(0);
    }
  }
  else{
    return(0);
  }
}



/* int T1_GetCharWidth(): This function returns the characterwidth
   specified in the .afm-file. If no .afm-file is loaded for that font,
   0 is returned. Note that if one tries to raster strings, afm data
   should always be available. The returned character width is corrected
   using  a possibly applied font extension!
   */
int T1_GetCharWidth( int FontID, unsigned char char1)
{
  
  /* Check whether font is loaded: */
  if (CheckForFontID(FontID)!=1)
    return(0);
  
  if (pFontBase->pFontArray[FontID].pAFMData==NULL)
    return(0);  /* font is loaded, but no afm data available */

  /* Check if character is encoded */
  if (pFontBase->pFontArray[FontID].pAFMData->enc_map[(int) char1]==-1)
    return(0);
  
  return((int) ((pFontBase->pFontArray[FontID].pAFMData->cmi[pFontBase->pFontArray[FontID].pAFMData->enc_map[(int) char1]].wx) * pFontBase->pFontArray[FontID].extend));
  
}



/* T1_GetCharBBox(): Get the BoundingBox of specified character. If an
   extension factor has been applied to the font in question, this
   is taken into account. However, a slant factor which has been applied
   to the font, also affects the bounding box of a character. The
   only way to determine its influence on the character bounding box
   is to compute the exact shape of that slanted character. There's no
   simple way to extract the new bounding box from the former bounding
   box. Thus, if a font has been slanted, the characters outline itself
   is examined. Since this must be done at 1000 bp it takes considerably
   longer than reading afm data. */
BBox T1_GetCharBBox( int FontID, unsigned char char1)
{

  struct region *area;
  struct XYspace *S;    
  int mode=0;
  
  extern int ForceAFMBBox;
  BBox NullBBox= { 0, 0, 0, 0}; /* A bounding box containing all 0's. */
  BBox ResultBox= { 0, 0, 0, 0}; /* The Box returned if char is found */
  
  
  /* Check whether font is loaded: */
  if (CheckForFontID(FontID)!=1)
    return(NullBBox);

  if (pFontBase->pFontArray[FontID].pAFMData==NULL)
    return(NullBBox);  /* font is loaded, but no afm data available */

  /* Check if character is encoded */
  if (pFontBase->pFontArray[FontID].pAFMData->enc_map[(int) char1]==-1)
    return(NullBBox);
  
  /* Check for a font slant */
  if ((pFontBase->pFontArray[FontID].slant!=0.0)
      &&(ForceAFMBBox==0)
      &&(ForceAFMBBoxInternal==0)){
    /* We have a font slant -> character outline must be examined in order
       to determine bounding box */
    /* Set up an identity charspace matrix 
       and take a slant and an extension into account */
    S=(struct XYspace *)IDENTITY;
    S=(struct XYspace *)Transform(S,
				  pFontBase->pFontArray[FontID].FontTransform[0],
				  pFontBase->pFontArray[FontID].FontTransform[1],
				  pFontBase->pFontArray[FontID].FontTransform[2],
				  pFontBase->pFontArray[FontID].FontTransform[3]);
    /* Genrate an edgelist for the current character at size 1000bp
       using current transformation and encoding: */
    area=fontfcnB( S,
		   pFontBase->pFontArray[FontID].pFontEnc,
		   (int) char1, &mode,
		   pFontBase->pFontArray[FontID].pType1Data);
    /* Read out bounding box */
    ResultBox.llx =area->xmin;
    ResultBox.urx =area->xmax;
    ResultBox.lly =area->ymin;
    ResultBox.ury =area->ymax;
    
    /* Reset AFM-switch and return BBox */
    ForceAFMBBoxInternal=0;
    return(ResultBox);

  }
  else{
    /* Assign bounding box ... */
    ResultBox=(pFontBase->pFontArray[FontID].pAFMData->cmi[pFontBase->pFontArray[FontID].pAFMData->enc_map[(int) char1]].charBBox);
    
    /* .. and apply transformations: */
    ResultBox.llx *=pFontBase->pFontArray[FontID].extend;
    ResultBox.urx *=pFontBase->pFontArray[FontID].extend;
    
    return(ResultBox);
  }
    
}


/* int T1_GetUnderlinePosition(): Return underline position of specified
   font in charspace units. If 0 is returned, it indicated that the font
   is not yet loaded into memory. or an invalid ID has been specified. */
float T1_GetUnderlinePosition( int FontID)
{
  if (CheckForFontID(FontID)!=1)
    return(0.0);
  return((float)(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[UNDERLINEPOSITION].value.data.real));
}



/* int T1_GetUnderlineThickness(): Return underline thickness of specified
   font in charspace units. If 0 is returned, it indicated that the font
   is not yet loaded into memory. or an invalid ID has been specified. */
float T1_GetUnderlineThickness( int FontID)
{
  if (CheckForFontID(FontID)!=1)
    return(0.0);
  return((float)(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[UNDERLINETHICKNESS].value.data.real));
}


/* unsigned char *T1_GetPostScriptName( FontID): Get the PostScriptName of
   the  font dictionary associated with thre specified font, or NULL if
   an error occurs. */
unsigned char *T1_GetPostScriptName( int FontID)
{
  static unsigned char fontname[MAXPSNAMELEN];
  
  if (CheckForFontID(FontID)!=1)
    return(NULL);
  
  strncpy(fontname,
	  (unsigned char *)(pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTNAME].value.data.nameP),
	  pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTNAME].value.len);
  fontname[pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTNAME].value.len]=0;
    
  return(fontname);
  
}



/* unsigned char *T1_GetCharName(): Get the PostScript character name of
   the  character indexed by char1. */
unsigned char *T1_GetCharName( int FontID, unsigned char char1)
{
  static unsigned char cc_name1[256];
  unsigned char *c1;
  

  if (CheckForFontID(FontID)!=1)
    return(NULL);
  
  if (pFontBase->pFontArray[FontID].pFontEnc==NULL){
    /* We have to get the names from the fonts internal encoding */
    c1= (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].data.arrayP;
    strncpy(cc_name1,
	    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].data.arrayP,
	    pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].len);
    cc_name1[pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].len]=0;
  }
  else{
    /* Take names from explicitly loaded and assigned encoding */
    c1=pFontBase->pFontArray[FontID].pFontEnc[char1];
    strcpy(cc_name1,c1);
  }

  /* Return address of charname */
  return(cc_name1);
  
}



/* T1_QueryLigs(): Get the number of ligatures defined in the font FontID for
   the character which is located at position char1 in the current encoding
   vector!
   Function returns the number of defined ligs (including 0) or -1 if an
   error occured.
   */
int T1_QueryLigs( int FontID,
		  unsigned char char1,
		  unsigned char **successors,
		  unsigned char **ligatures)
{

  FontInfo *afm_ptr;
  CharMetricInfo *m_ptr;
  unsigned char *c_name;
  unsigned char cc_name[128];
  unsigned char succ[MAX_LIGS];
  unsigned char succ_index;
  unsigned char lig[MAX_LIGS];
  unsigned char lig_index;
  
  Ligature *ligs;
  int i,j;
  
  /* Check whether font is loaded: */
  if (CheckForFontID(FontID)!=1)
    return(-1);

  if (pFontBase->pFontArray[FontID].pAFMData==NULL)
    return(-1);  /* font is loaded, but no afm data available */

  /* All OK, ... */
  afm_ptr=pFontBase->pFontArray[FontID].pAFMData;
  m_ptr=afm_ptr->cmi;

  /* Get the name of the character: */
  if (pFontBase->pFontArray[FontID].pFontEnc==NULL){
    /* We have to get the name from the fonts internal encoding */
    c_name=(unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].data.arrayP;
    strncpy(cc_name,
	    (unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].data.arrayP,
	    pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].len);
    cc_name[pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[char1].len]=0;
  }
  else{
    /* Take name from explicitly loaded and assigned encoding */
    c_name=pFontBase->pFontArray[FontID].pFontEnc[char1];
    strcpy(cc_name,c_name);
  }

  for (i=0; i<afm_ptr->numOfChars; i++)
    if (strcmp(m_ptr[i].name,cc_name)==0)
      break;

  if (i==afm_ptr->numOfChars) /* we didn't find the characters name */
    return(-1);
  
  ligs=m_ptr[i].ligs;

  j=0;
  if (ligs==NULL)
    return(0);
  
  while (ligs!=NULL){
    /* Get indices of the two characters: */
    succ_index=T1_GetEncodingIndex( FontID, (unsigned char*) ligs->succ);
    lig_index=T1_GetEncodingIndex( FontID, (unsigned char*) ligs->lig);
    succ[j]=(unsigned char)succ_index;
    lig[j]=(unsigned char)lig_index;
    j++;
    ligs=ligs->next;
  }
    
  *successors=succ;
  *ligatures=lig;
  
  return(j);
}

      

/* T1_GetEncodingIndex(): Return the Index of char1 in the current
   encoding vector of font FontID */
int T1_GetEncodingIndex( int FontID, unsigned char *char1)
{
  int i;
  int len1;
  int result_index;
  
  if (CheckForFontID(FontID)!=1)
    return(-1);

  len1=strlen( char1);
  
  /* The default return-value if character is not found: */
  result_index=-1;

  if (pFontBase->pFontArray[FontID].pFontEnc==NULL){
    /* We have to search the fonts internal encoding */
    for (i=0;i<256;i++){
      if (len1==pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[i].len)
	if (strncmp((unsigned char *)pFontBase->pFontArray[FontID].pType1Data->fontInfoP[ENCODING].value.data.arrayP[i].data.arrayP,
		    char1, len1)==0){
	  result_index=i;
	  break;
	}
    }
  }
  else{
    /* Take name from explicitly loaded and assigned encoding */
    for (i=0;i<256;i++){
      if (len1==strlen(pFontBase->pFontArray[FontID].pFontEnc[i]))
	if (strncmp(pFontBase->pFontArray[FontID].pFontEnc[i],
		    char1, len1)==0){
	  result_index=i;
	  break;
	}
    }
  }
  
  return(result_index);
}


/* int T1_GetStringWidth(): This function returns the width of string
   in .afm-file units. If no .afm-file is loaded for font FontID,
   0 is returned. Note that if one tries to raster strings, afm data
   should always be available. The returned character width is corrected
   using  a possibly applied font extension!
   */
int T1_GetStringWidth( int FontID, unsigned char *string,
		       int len,  long spaceoff, int kerning)
{

  int no_chars;      /* Number of chars in string */
  int i;
  int *kern_pairs;
  int *charwidths;
  int spacewidth; 
  int stringwidth;

  
  /* First, check for a correct ID */
  i=CheckForFontID(FontID);
  if (i==-1)
    return(0);
  /* if necessary load font into memory */
  if (i==0)
    if (T1_LoadFont(FontID))
      return(0);

  /* Get length of string: */
  if (len<0)  /* invalid length */
    return(0);
  if (len==0) /* should be computed assuming "normal" 0-terminated string */
    no_chars=strlen(string);
  else        /* use value given on command line */
    no_chars=len;

  /* Allocate room for temporary arrays of kerning and width arrays: */
  kern_pairs=(int *)calloc(no_chars -1, sizeof(int));
  charwidths=(int *)calloc(no_chars, sizeof(int));
  
  /* If kerning is requested, get kerning amounts and fill the array: */
  if (kerning){
    for (i=0; i<no_chars -1; i++){
      kern_pairs[i]=T1_GetKerning( FontID, string[i], string[i+1]);
    }
  }
  
  /* Compute the correct spacewidth value (in charspace units): */
  spacewidth=T1_GetCharWidth(FontID,pFontBase->pFontArray[FontID].space_position)+spaceoff;
  
  /* Fill the width-array:  */
  for (i=0; i<no_chars; i++){
    if (string[i]==pFontBase->pFontArray[FontID].space_position)
      charwidths[i]=(int)spacewidth;
    else
      charwidths[i]=T1_GetCharWidth(FontID,string[i]);
  }
  
  /* Accumulate width: */
  stringwidth=0;
  for (i=0; i<no_chars-1; i++){
    stringwidth += kern_pairs[i];
  }
  for (i=0; i<no_chars; i++){
    stringwidth += charwidths[i];
  }
  
  /* free memory: */
  free( charwidths);
  free( kern_pairs);

  /* .. and return result: */
  return( stringwidth);

}

    

/* int T1_GetStringBBox(): This function returns the bounding box of string
   in .afm-file units. If no .afm-file is loaded for font FontID,
   0 is returned. Note that if one tries to raster strings, afm data
   should always be available. The returned character width is corrected
   using  a possibly applied font extension!
   */
BBox T1_GetStringBBox( int FontID, unsigned char *string,
		       int len,  long spaceoff, int kerning)
{

  BBox NullBBox= { 0, 0, 0, 0}; /* A bounding box containing all 0's. */
  BBox tmp_BBox= { 0, 0, 0, 0}; 
  BBox ResultBBox= { 0, 0, 0, 0}; /* The resulting BBox */
  int i;
  int no_chars;
  int stringwidth;
  
  
  /* First, check for a correct ID */
  i=CheckForFontID(FontID);
  if (i==-1)
    return(NullBBox);
  /* if necessary load font into memory */
  if (i==0)
    if (T1_LoadFont(FontID))
      return(NullBBox);

  /* Get length of string: */
  if (len<0)  /* invalid length */
    return(NullBBox);
  if (len==0) /* should be computed assuming "normal" 0-terminated string */
    no_chars=strlen(string);
  else        /* use value given on command line */
    no_chars=len;

  /* First, get horizontal information: */
  stringwidth=T1_GetStringWidth( FontID, string, len, spaceoff, kerning);
  ResultBBox.llx=T1_GetCharBBox( FontID, string[0]).llx;
  ResultBBox.urx=T1_GetCharBBox( FontID, string[no_chars-1]).urx + stringwidth
    - T1_GetCharWidth( FontID, string[no_chars -1]);

  /* Now, get vertical maximum and minimum. In this case it is sufficient
     to have the vertical metrics correct --> AFM-info is sufficient:*/
  ForceAFMBBoxInternal=1;
  for ( i=0; i<no_chars; i++){
    tmp_BBox=T1_GetCharBBox( FontID, string[i]);
    if (ResultBBox.lly>tmp_BBox.lly)
      ResultBBox.lly=tmp_BBox.lly;
    if (ResultBBox.ury<tmp_BBox.ury)
      ResultBBox.ury=tmp_BBox.ury;
  }
  /* Reset the switch: */
  ForceAFMBBoxInternal=0;
  
  return( ResultBBox);
  
}


/* T1_GetMetricsInfo(): Return a structure containing metrics information
   about the string to the user. */
METRICSINFO T1_GetMetricsInfo( int FontID, unsigned char *string,
			       int len, long spaceoff, int kerning)
{

  BBox NullBBox= { 0, 0, 0, 0}; /* A bounding box containing all 0's. */
  BBox tmp_BBox= { 0, 0, 0, 0}; 

  extern int ForceAFMBBox;

  int i;
  int no_chars;
  int spacewidth;
  
  static METRICSINFO metrics={ 0, {0, 0, 0, 0}, 0, NULL};

  
  /* Reset struct: */
  metrics.width=0;
  metrics.bbox=NullBBox;
  metrics.numchars=0;
  if (metrics.charpos != NULL){
    free( metrics.charpos);
    metrics.charpos=NULL;
  }


  /* First, check for a correct ID */
  i=CheckForFontID(FontID);
  if (i==-1)
    return(metrics);
  /* if necessary load font into memory */
  if (i==0)
    if (T1_LoadFont(FontID))
      return(metrics);

  /* Get length of string: */
  if (len<0)  /* invalid length */
    return(metrics);
  if (len==0) /* should be computed assuming "normal" 0-terminated string */
    no_chars=strlen(string);
  else        /* use value given on command line */
    no_chars=len;

  /* Compute the correct spacewidth value (in charspace units): */
  spacewidth=T1_GetCharWidth(FontID,pFontBase->pFontArray[FontID].space_position)+spaceoff;
  /* Allocate memory for character positions array: */
  metrics.charpos=(int *)calloc(no_chars, sizeof(int));

  metrics.numchars=no_chars;
  /* A loop for all characters in string: */
  for ( i=0; i<no_chars; i++){
    metrics.charpos[i]=metrics.width;
    /* Advance by characters width or by space width */
    if (string[i]==pFontBase->pFontArray[FontID].space_position)
      metrics.width += spacewidth;
    else
      metrics.width += T1_GetCharWidth( FontID, string[i]);
    
      /* Get optional kerning value and add it */
    if ((i<no_chars-1)&&(kerning))
      metrics.width += T1_GetKerning( FontID, string[i], string[i+1]);

    /* Get maximum and minimum for bounding box: */
    /* Accuracy in x-direction is only needed for first and last
       character -> force usage of afm info in all other cases
       for performance reasons: */
    if ((i==0)||(i==no_chars-1))
      ForceAFMBBoxInternal=0;
    else
      ForceAFMBBoxInternal=1;
    tmp_BBox=T1_GetCharBBox( FontID, (unsigned char) string[i]);
    
    /* vertical info for all characters: */
    if (metrics.bbox.lly>tmp_BBox.lly)
      metrics.bbox.lly=tmp_BBox.lly;
    if (metrics.bbox.ury<tmp_BBox.ury)
      metrics.bbox.ury=tmp_BBox.ury;
    /* horizontal info only for first and last character required */
    if (i==0)
      metrics.bbox.llx=tmp_BBox.llx;
    if (i==no_chars-1)
      metrics.bbox.urx=tmp_BBox.urx + metrics.charpos[i];
    
  }
  
  return( metrics);
  
}



/* T1_GetFontBBox(): Return the font's bounding box. Note: The font
   BBox is taken is taken from the font file rather than from afm
   file since I have seen some afm with rather inaccurate BBoxes.: */
BBox T1_GetFontBBox( int FontID)
{
  
  BBox outbox= { 0, 0, 0, 0}; 

  /* return Null-box if font not loaded */
  if (CheckForFontID(FontID)!=1)
    return(outbox);
  
  outbox.llx=pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTBBOX].value.data.arrayP[0].data.integer; 
  outbox.lly=pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTBBOX].value.data.arrayP[1].data.integer;
  outbox.urx=pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTBBOX].value.data.arrayP[2].data.integer;
  outbox.ury=pFontBase->pFontArray[FontID].pType1Data->fontInfoP[FONTBBOX].value.data.arrayP[3].data.integer;
  
  return( outbox);
}

       

/* T1_GetAllCharNames(): Get a list of all defined character names in
   in the font FontID: */
unsigned char **T1_GetAllCharNames( int FontID)
{
  static unsigned char **bufmem=NULL;
  register unsigned char *namedest;
  psdict *pCharStrings;
  int len, i, j;
  long nameoffset;
  
  int bufmemsize=0;
  
  /* return NULL if font not loaded */
  if (CheckForFontID(FontID)!=1)
    return( NULL);
  
  pCharStrings=pFontBase->pFontArray[FontID].pType1Data->CharStringsP;

  /* First, get number of charstrings: */
  len=pCharStrings[0].key.len;

  for ( i=1; i<=len; i++){
    /* calculate room for each characters name plus the prepending \0 */ 
    bufmemsize += pCharStrings[i].key.len + 1;
  }
  /* Now we reserve memory for the pointers (including final NULL) */
  nameoffset=(len+1)*sizeof( unsigned char *);
  bufmemsize += nameoffset;

  /* Now allocate memory, copy strings and initialize pointers */
  if (bufmem!=NULL)
    free(bufmem);
  if ((bufmem=(unsigned char **)malloc( bufmemsize))==NULL)
    return(NULL);

  namedest=(unsigned char *)((long)bufmem + nameoffset);
  j=0;
  for ( i=0; i<len; i++){
    bufmem[i]=&(namedest[j]);
    strncpy( &(namedest[j]), pCharStrings[i+1].key.data.nameP,
	     pCharStrings[i+1].key.len);
    j += pCharStrings[i+1].key.len;
    namedest[j++]='\0';
  }
  bufmem[i++]=NULL;
  
  return( bufmem);
  
}

