/*--------------------------------------------------------------------------
  ----- File:        t1set.c 
  ----- Author:      Rainer Menzner (rmz@neuroinformatik.ruhr-uni-bochum.de)
  ----- Date:        11/29/1997
  ----- Description: This file is part of the t1-library. It contains
                     functions for setting characters and strings of
		     characters.
  ----- Copyright:   This program is copyright (c) Rainer Menzner, 1996. 
                     Everyone is allowed to use the library and to make
		     changes to the source code, provided that the authors
		     name is left where it appears and that changes are
		     metioned in some notice. The library must be
		     distributed including sources, you are not allowed to
		     take money for this.
		     The parseAFM-package is copyrighted by Adobe Systems
		     Inc.
		     The type1 rasterizer is copyrighted by IBM and the
		     X11-consortium.
  ----- Warranties:  Of course, there's NO WARRANTY OF ANY KIND :-)
  ----- Credits:     I want to thank IBM and the X11-consortium for making
                     their rasterizer freely available.
		     Also thanks to Piet Tutelaers for his ps2pk, from
		     which I took the rasterizer sources in a format
		     independent from X11.
                     Thanks to all people who make free software living!
--------------------------------------------------------------------------*/
  
#define T1SET_C


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "../type1/ffilest.h" 
#include "../type1/types.h"
#include "parseAFM.h" 
#include "../type1/objects.h"
#include "../type1/spaces.h"
#include "../type1/util.h"
#include "../type1/fontfcn.h"
#include "../type1/regions.h"


#include "t1types.h"
#include "t1extern.h"
#include "t1set.h"
#include "t1load.h"
#include "t1finfo.h"
#include "t1misc.h"
#include "t1base.h"


static int T1_bit=T1GLYPH_BIT;
static int T1_byte;
static int T1_pad;
static int T1_wordsize;



/* T1_SetChar(...): Generate the bitmap for a character */
GLYPH *T1_SetChar( int FontID, unsigned char charcode, float size, float angle)
{
  int i;
  int mode;
  struct region *area;
  struct XYspace *Current_S;
  int cache_flag=1;
  int rot_flag=0;
  
  FONTSIZEDEPS *font_ptr;
  
  int memsize;
  LONG h,w;
  LONG paddedW;
 

  static GLYPH glyph={NULL,{0,0,0,0,0},NULL,1};
  

  
  /* Reset character glyph, if necessary */
  if (glyph.bits!=NULL){
    free(glyph.bits);
    glyph.bits=NULL;
  }
  glyph.metrics.leftSideBearing=0;
  glyph.metrics.rightSideBearing=0;
  glyph.metrics.characterWidth=0;
  glyph.metrics.ascent=0;
  glyph.metrics.descent=0;
  glyph.pFontCacheInfo=NULL;
  glyph.bpp=1;  
  
  /* First, check for a correct ID */
  i=CheckForFontID(FontID);
  if (i==-1)
    return(NULL);
  /* if necessary load font into memory */
  if (i==0)
    if (T1_LoadFont(FontID))
      return(NULL);

  /* Assign padding value */
  T1_pad=pFontBase->bitmap_pad;
  if (pFontBase->endian)
    T1_byte=1;
  else
    T1_byte=0;
  T1_wordsize=T1_pad;
  /* Check if rotation is requested */
  if (angle!=0.0){
    /* there's rotation requested => do not cache the resulting bitmap */
    rot_flag=1;
    cache_flag=0;
    /* Convert angle to radians: */
    angle=angle*PI/180.0;
  }
  else {
    rot_flag=0;
    cache_flag=1;
  }
  
    
  /* font is now loaded into memory =>
     Check for size: */
  if ((font_ptr=QueryFontSize( FontID, size, NO_ANTIALIAS))==NULL){
    font_ptr=CreateNewFontSize( FontID, size, NO_ANTIALIAS);
  }
  else {/* size is already existent in cache */
    /* If no rotation, try to get character from cache */
    if (rot_flag==0){
      if (font_ptr->pFontCache[charcode].bits != NULL){
	/* Character is already in Chache -> create a copy of cache
	   and return a pointer to the result: */
	memcpy( &glyph, &(font_ptr->pFontCache[charcode]), sizeof(GLYPH));
	paddedW = PAD((glyph.metrics.rightSideBearing-glyph.metrics.leftSideBearing),
		      T1_pad);
	memsize = (glyph.metrics.ascent+glyph.metrics.descent)*paddedW / 8;
	glyph.bits = (char *)malloc(memsize*sizeof( char));
	if (glyph.bits == NULL) 
	  return(NULL);
	memcpy( glyph.bits, font_ptr->pFontCache[charcode].bits, memsize);
	return(&(glyph));
      }
    }
  }
  
  /* If rotation is requested create a local copy of the Charspace matrix */
  if (rot_flag){
    Current_S=(struct XYspace *)Permanent(CopySpace(font_ptr->pCharSpaceLocal));
    Current_S=(struct XYspace *)Transform( Current_S,
					   cos((double)angle), - sin((double)angle),
					   sin((double)angle), cos((double)angle));
  }
  else{
    /* Set the pointer directly to the CharSpace matrix */
    Current_S=font_ptr->pCharSpaceLocal;
  }

  
  /* fnt_ptr now points to the correct FontSizeDeps-struct =>
     lets now raster the character */
  mode=0;
  area=fontfcnB( Current_S,
		 pFontBase->pFontArray[FontID].pFontEnc,
		 charcode, &mode,
		 pFontBase->pFontArray[FontID].pType1Data);

  /* fill the glyph-structure */
  if (mode < 0) {
    sprintf( err_warn_msg_buf, "fontfcnB() set mode=%d", mode);
    T1_PrintLog( "T1_SetChar()", err_warn_msg_buf, T1LOG_WARNING);
    return(NULL);
  }
  if (area == NULL){
    T1_PrintLog( "T1_SetChar()", "area=NULL returned by fontfcnB()", T1LOG_WARNING);
    return(NULL);
  }
  h = area->ymax - area->ymin;
  w = area->xmax - area->xmin;
  paddedW = PAD(w, T1_pad);
  if ((h==-65535)&&(w==-65535)){
    /* There was a character like .notdef or space, that didn't
       produce any black pixels on the bitmap! -> pathological */
    sprintf( err_warn_msg_buf,
	     "No black pixels found for character %d from font %d, returning NULL",
	     charcode, FontID);
    T1_PrintLog( "T1_SetChar()", err_warn_msg_buf, T1LOG_WARNING);
    return(NULL);
  }
  if (h > 0 && w > 0) {
    memsize = h * paddedW / 8;
    /* This is for the users copy of the character, for security-reasons
       the original pointer to the cache area is not used. The entry glyph.bits
       is free'ed every time this function is called: */
    glyph.bits = (char *)malloc(memsize*sizeof( char));
    if (glyph.bits == NULL) 
      return(NULL);
  }
  else {
    h = w = 0;
    area->xmin = area->xmax = 0;
    area->ymax = area->ymax = 0;
  }
  
  glyph.metrics.leftSideBearing  = area->xmin;
  glyph.metrics.characterWidth   = NEARESTPEL(area->ending.x - area->origin.x);
  glyph.metrics.rightSideBearing = w + area->xmin;
  glyph.metrics.descent          = area->ymax - NEARESTPEL(area->origin.y);
  glyph.metrics.ascent           = h - glyph.metrics.descent;

  
  if (h > 0 && w > 0) {
    (void) memset(glyph.bits, 0, memsize);
    fill(glyph.bits, h, paddedW, area, T1_byte, T1_bit, T1_wordsize );
  }
  
  
  /* Cache glyph if requested */
  if (cache_flag){
    /* Copy glyph to cache but be careful not to modify the pixmap
       entry */
    font_ptr->pFontCache[charcode].metrics.leftSideBearing=glyph.metrics.leftSideBearing;
    font_ptr->pFontCache[charcode].metrics.rightSideBearing=glyph.metrics.rightSideBearing;
    font_ptr->pFontCache[charcode].metrics.characterWidth=glyph.metrics.characterWidth;
    font_ptr->pFontCache[charcode].metrics.ascent=glyph.metrics.ascent;
    font_ptr->pFontCache[charcode].metrics.descent=glyph.metrics.descent;
    font_ptr->pFontCache[charcode].pFontCacheInfo=glyph.pFontCacheInfo;
    font_ptr->pFontCache[charcode].bpp=glyph.bpp;  
    font_ptr->pFontCache[charcode].bits = (char *)malloc(memsize*sizeof( char));
    if (font_ptr->pFontCache[charcode].bits == NULL) 
      return(NULL);
    /* ... and copy bitmap to cache area: */
    memcpy( font_ptr->pFontCache[charcode].bits, glyph.bits, memsize);
  }
  
  return(&glyph);
}



/* T1_SetString(...): Generate the bitmap for a string of characters */
GLYPH *T1_SetString( int FontID, unsigned char *string, int len, 
		     long spaceoff, int kerning, float size, float angle)
{
  int i, j, k;
  int mode;
  struct region *area;
  struct XYspace *Current_S;
  int cache_flag=1;
  int rot_flag=0;
  int *kern_pairs;       /* use for accessing the kern pairs if kerning is
			    requested */
  int *charwidths;
  int no_chars=0;        /* The number of characters in the string */
  float factor;
  long spacewidth;       /* This is given to fontfcnb_string() */
   
  
  FONTSIZEDEPS *font_ptr;

  int memsize;

  /* The following are for bitmap transfer */
  unsigned short BitBuf_c;
  unsigned long BitBuf_s;
  long long BitBuf_l;
  long BitShift;
  long ByteOffset;

  
  long h,w;
  long paddedW, char_paddedW;
  long h_anchor, v_anchor, h_anchor_corr, pixel_h_anchor_corr;
  long overallwidth=0;
  long overallascent=0;
  long overalldescent=0;

  
  static GLYPH string_glyph={NULL,{0,0,0,0,0},NULL,1};
  GLYPH *currchar;
  
  unsigned char *p_c;
  unsigned short *p_s;
  unsigned long *p_l;
  unsigned char  *Target_c;
  unsigned short *Target_s;
  unsigned long  *Target_l;
  
  
  /* Reset string glyph, if necessary */
  if (string_glyph.bits!=NULL){
    free(string_glyph.bits);
    string_glyph.bits=NULL;
  }
  string_glyph.metrics.leftSideBearing=0;
  string_glyph.metrics.rightSideBearing=0;
  string_glyph.metrics.characterWidth=0;
  string_glyph.metrics.ascent=0;
  string_glyph.metrics.descent=0;
  string_glyph.pFontCacheInfo=NULL;
  string_glyph.bpp=1;
  
  /* We don't want to cache the resulting bitmap: */
  cache_flag=0;
  
  /* First, check for a correct ID */
  i=CheckForFontID(FontID);
  if (i==-1)
    return(NULL);
  /* if necessary load font into memory */
  if (i==0)
    if (T1_LoadFont(FontID))
      return(NULL);

  /* Assign padding value */
  T1_pad=pFontBase->bitmap_pad;
  if (pFontBase->endian)
    T1_byte=1;
  else
    T1_byte=0;
  T1_wordsize=T1_pad;
  /* Check if rotation is requested */
  if (angle!=0.0){
    /* there's rotation requested => do not cache the resulting bitmap */
    rot_flag=1;
    cache_flag=0;
    /* Convert angle to radians: */
    angle=angle*PI/180.0;
  }
  
  /* font is now loaded into memory =>
     Check for size: */
  if ((font_ptr=QueryFontSize( FontID, size, NO_ANTIALIAS))==NULL){
    font_ptr=CreateNewFontSize( FontID, size, NO_ANTIALIAS);
  }
  
  /* Now comes string specific stuff: Get length of string and create an
     array of integers where to store the kerning pairs and character
     widths: */
  if (len<0)  /* invalid length */
    return(NULL);
  if (len==0) /* should be computed assuming "normal" 0-terminated string */
    no_chars=strlen(string);
  else        /* use value given on command line */
    no_chars=len;
  
  kern_pairs=(int *)calloc(no_chars -1, sizeof(int));
  charwidths=(int *)calloc(no_chars, sizeof(int));
  
  /* If kerning is requested, get kerning amounts and fill the array: */
  if (kerning){
    for (i=0; i<no_chars -1; i++){
      kern_pairs[i]=T1_GetKerning( FontID, string[i], string[i+1]);
    }
  }
  
  /* If rotation is requested create a local copy of the Charspace matrix */
  if (rot_flag){
    Current_S=(struct XYspace *)Permanent(CopySpace(font_ptr->pCharSpaceLocal));
    Current_S=(struct XYspace *)Transform( Current_S,
					   cos((double)angle), - sin((double)angle),
					   sin((double)angle), cos((double)angle));
  }
  else{
    /* Set the pointer directly to the CharSpace matrix */
    Current_S=font_ptr->pCharSpaceLocal;
  }
  
  /* Compute the correct spacewidth value (in charspace units). The
     value supplied by the user is interpreted as an offset in
     char space units:
     */
  spacewidth=T1_GetCharWidth(FontID,pFontBase->pFontArray[FontID].space_position)+spaceoff;
  
  
  /* The following if (rot_flag==0){ ... } contains the whole code
     to execute when the string is oriented horizontal. Then there's
     the possibility of concatenating most probably existent bitmaps
     from the Cache, which should be much faster than every time raster
     all the characters.
     */
  /* First, ensure that all needed characters are in the Cache; if not,
     generate them */
  if ((rot_flag==0)){  
    
    overallwidth=0;
    for (i=0; i<no_chars; i++){
      if (font_ptr->pFontCache[string[i]].bits==NULL){
	mode=0;
	/*  Don't try to generate space-character: */
	if (string[i]!=pFontBase->pFontArray[FontID].space_position){
	  area=fontfcnB( Current_S,
			 pFontBase->pFontArray[FontID].pFontEnc,
			 string[i], &mode,
			 pFontBase->pFontArray[FontID].pType1Data);

	  /* fill the glyph-structure */
	  if (mode < 0) {
	    sprintf( err_warn_msg_buf, "fontfcnB() set mode=%d", mode);
	    T1_PrintLog( "T1_SetString()", err_warn_msg_buf, T1LOG_WARNING);
	    return(NULL);
	  }
	  if (area == NULL){
	    T1_PrintLog( "T1_SetString()", "area=NULL returned by fontfcnB()", T1LOG_WARNING);
	    return(NULL);
	  }
	  h = area->ymax - area->ymin;
	  w = area->xmax - area->xmin;
	  
	  if ((h==-65535)&&(w==-65535)&&strcmp("space",T1_GetCharName(FontID,string[i]))){
	    /* There was a character like .notdef, that didn't
	       produce any black pixels on the bitmap! -> pathological
	       Note: space is valid and deserves special treatment! */
	    sprintf( err_warn_msg_buf,
		     "No black pixels in string[%d] (=character %d) from font %d, result may be damaged",
		     i, string[i], FontID);
	    T1_PrintLog( "T1_SetString()", err_warn_msg_buf, T1LOG_WARNING);
	  }
	  
	  paddedW = PAD(w, T1_pad);
	  if (h > 0 && w > 0) {
	    memsize = h * paddedW / 8;
	    font_ptr->pFontCache[string[i]].bits = (char *)malloc(memsize*sizeof( char));
	    if (font_ptr->pFontCache[string[i]].bits == NULL) 
	      return(NULL);
	  }
	  else {
	    h = w = 0;
	    area->xmin = area->xmax = 0;
	    area->ymax = area->ymax = 0;
	  }
	  font_ptr->pFontCache[string[i]].metrics.leftSideBearing  = area->xmin;
	  font_ptr->pFontCache[string[i]].metrics.characterWidth   = NEARESTPEL(area->ending.x - area->origin.x);
	  font_ptr->pFontCache[string[i]].metrics.rightSideBearing = w + area->xmin;
	  font_ptr->pFontCache[string[i]].metrics.descent          = area->ymax - NEARESTPEL(area->origin.y);
	  font_ptr->pFontCache[string[i]].metrics.ascent           = h - font_ptr->pFontCache[string[i]].metrics.descent;
	  
	  
	  if (h > 0 && w > 0) {
	    (void) memset(font_ptr->pFontCache[string[i]].bits, 0, memsize);
	    fill(font_ptr->pFontCache[string[i]].bits, h, paddedW, area,
		 T1_byte, T1_bit, T1_wordsize );
	  }
	}
      } /* end of if ( ... ), needed characters are now in cache! */
      /* Accumulate the width information and get the maximum amount of
	 ascent and descent of all characters in string. Note: To avoid
	 rounding errors, width information is taken from afm data in
	 charspace units. */
      if (string[i]==pFontBase->pFontArray[FontID].space_position)
	charwidths[i]=(int)spacewidth;
      else
	charwidths[i]=T1_GetCharWidth(FontID,string[i]);
      overallwidth +=charwidths[i];
      
      /* Vertical metrics are handle in pixel units since there may not appear
	 any rounding error effects: */
      if (font_ptr->pFontCache[string[i]].metrics.ascent>overallascent)
	overallascent=font_ptr->pFontCache[string[i]].metrics.ascent;
      if (font_ptr->pFontCache[string[i]].metrics.descent>overalldescent)
	overalldescent=font_ptr->pFontCache[string[i]].metrics.descent;
    } /* end of for (i<no_chars) loop */


    /* We've to compute the scale factor: */
    factor=size*DeviceSpecifics.scale_x/1000;
    
    /* Correct width by kerning values */
    for (i=0; i<no_chars-1; i++){
      overallwidth +=kern_pairs[i]; 
    }


    /* Now we have the over all characterwidth accumulated; store it as
       the overall width (including margins) of the resulting bitmap: */
    string_glyph.metrics.characterWidth=(int)ceil(overallwidth*factor);

    /* Store metrics for string bitmap */
    /* Left margin: */
    string_glyph.metrics.leftSideBearing=font_ptr->pFontCache[string[0]].metrics.leftSideBearing;
    /* ascent and descent: */
    string_glyph.metrics.ascent=overallascent;
    string_glyph.metrics.descent=overalldescent;
    /* Width is taken on charspace units base and right side bearing of last
       character in string is subtracted on pixel base: */
    string_glyph.metrics.rightSideBearing=(int)ceil(overallwidth*factor)
      - font_ptr->pFontCache[string[no_chars-1]].metrics.characterWidth
      + font_ptr->pFontCache[string[no_chars-1]].metrics.rightSideBearing;

    /* Now, allocate memory for the resulting bitmap and set it to 0: */
    h=overallascent + overalldescent;
    w=string_glyph.metrics.rightSideBearing-string_glyph.metrics.leftSideBearing;
    paddedW = PAD(w, T1_pad);
    if ( h > 0 && w > 0){
      memsize = h * paddedW / 8;
      string_glyph.bits = (char *)malloc( memsize);
      if (string_glyph.bits == NULL) 
	return(NULL);
    }
    else
      return(NULL);
    memset(string_glyph.bits, 0, memsize);

    /* h_anchor is the horizontal position in the resulting bitmap;
       it is incremented on the fly and of course initially 0! */
    h_anchor=0;
    /* Now comes the loop for placing the bitmaps and the kernings: */
    for (i=0;i<no_chars;i++){
      /* Get pointer to character number i of string: */
      currchar=&(font_ptr->pFontCache[string[i]]);
      /* Compute vertical anchor for current char-bitmap: */
      v_anchor=overallascent - currchar->metrics.ascent;
      char_paddedW=PAD( currchar->metrics.rightSideBearing
			- currchar->metrics.leftSideBearing , T1_pad);
      /* Compute pixel value; correct position by adding left sidebearing
	 value of that character, except it is the first character: */
      h_anchor_corr=h_anchor;
      pixel_h_anchor_corr=((int)floor(h_anchor_corr*factor+0.5));
      if (i>0){
	pixel_h_anchor_corr -= font_ptr->pFontCache[string[0]].metrics.leftSideBearing;
	pixel_h_anchor_corr += currchar->metrics.leftSideBearing;
      }
      /* We have to check for Big Endian. In that case, we have to
	 act on byte-level */
      if (T1_byte){
	BitShift =  pixel_h_anchor_corr % 8;
	ByteOffset = pixel_h_anchor_corr / 8;
      }
      else { 
	BitShift =  pixel_h_anchor_corr % T1_pad;
	ByteOffset = pixel_h_anchor_corr / T1_pad;
	if (T1_pad==32)
	  ByteOffset *=4;
	else if (T1_pad==16)
	  ByteOffset *=2;
      }  
      
      if (T1_pad==32 && T1_byte==0){
	/* The following loop steps through the lines of the character bitmap: */
	for (j=0;j<currchar->metrics.ascent+currchar->metrics.descent;j++){
	  Target_l= (unsigned long *)(string_glyph.bits +((v_anchor+j)*paddedW/8)
				      +ByteOffset);
	  /* The following loop copies the scanline of a character bitmap: */
	  p_l = (unsigned long *)(currchar->bits+(char_paddedW/8*j));
	  if (BitShift == 0) {
	    for (k=char_paddedW >> 5; k; k--)
	      *Target_l++ |= *p_l++;
	  } else {
	    for (k=0; k < char_paddedW / 32 ; k++){
	      BitBuf_l= ((long long)(*p_l++)) << BitShift;
	      *Target_l++ |= BitBuf_l;
	      *Target_l |= BitBuf_l>>32;
	    } /* End of for ( .. ) stepping through columns */
	  }
	} /* End of for( .. ) steppin' through lines of char bitmap */
      }/* end if (T1_pad==32) */
      else if (T1_pad==16 && T1_byte==0){
	/* The following loop steps through the lines of the character bitmap: */
	for (j=0;j<currchar->metrics.ascent+currchar->metrics.descent;j++){
	  Target_s= (unsigned short *)(string_glyph.bits +((v_anchor+j)*paddedW/8)
				       +ByteOffset);
	  /* The following loop copies the scanline of a character bitmap: */
	  p_s = (unsigned short *)(currchar->bits+(char_paddedW/8*j));
	  if (BitShift == 0) {
	    for (k=char_paddedW >> 4; k; k--)
	      *Target_s++ |= *p_s++;
	  }
	  else{
	    for (k=char_paddedW >> 4; k; k--){
	      BitBuf_s= ((long)(*p_s++))<<BitShift;
	      *Target_s++ |= BitBuf_s;
	      *Target_s |= BitBuf_s>>16;
	    } /* End of for ( .. ) stepping through columns */
	  }
	} /* End of for( .. ) steppin' through lines of char bitmap */
      }/* end if (T1_pad==16 */
      else{ /* T1_pad = 8 or Big Endian machine */
	/* The following loop steps through the lines of the character bitmap: */
	for (j=0;j<currchar->metrics.ascent+currchar->metrics.descent;j++){
	  Target_c= (unsigned char *)(string_glyph.bits +((v_anchor+j)*paddedW/8)
				      +ByteOffset);
	  /* The following loop copies the scanline of a character bitmap: */
	  p_c = (unsigned char *)(currchar->bits+(char_paddedW/8*j));
	  if (BitShift == 0){
	    for (k=char_paddedW >> 3; k; k--)
	      *Target_c++ |= *p_c++;
	  }
	  else{
	    for (k=char_paddedW >> 3; k; k--){
	      BitBuf_c = ((short)*p_c++) << BitShift;
	      *Target_c++ |= BitBuf_c;
	      *Target_c |= BitBuf_c>>8;
	    } /* End of for ( .. ) stepping through columns */
	  }
	} /* End of for( .. ) steppin' through lines of char bitmap */
      } /* end if (T1_pad==8) and/or BigEndian  */
      h_anchor +=charwidths[i];
      if (i<no_chars -1)
	h_anchor +=kern_pairs[i];
    }
    /* Free memory for kerning pairs */
    free(kern_pairs);
    free(charwidths);
    return(&string_glyph);
  } /* end of "if (rot_flag==0.0)" */


  /* fnt_ptr now points to the correct FontSizeDeps-struct =>
     lets now raster the character */
  mode=0;
  area=fontfcnB_string( Current_S,
			pFontBase->pFontArray[FontID].pFontEnc,
			string, no_chars, &mode,
			pFontBase->pFontArray[FontID].pType1Data,
			kern_pairs, spacewidth);
  /* In all cases, free memory for kerning pairs */
  free(kern_pairs);
  
  /* fill the string_glyph-structure */
  if (mode < 0) {
    sprintf( err_warn_msg_buf, "fontfcnB() set mode=%d", mode);
    T1_PrintLog( "T1_SetChar()", err_warn_msg_buf, T1LOG_WARNING);
    return(NULL);
  }
  if (area == NULL){
    T1_PrintLog( "T1_SetChar()", "area=NULL returned by fontfcnB()", T1LOG_WARNING);
    return(NULL);
  }
  
  h = area->ymax - area->ymin;
  w = area->xmax - area->xmin;
  paddedW = PAD(w, T1_pad);
  if (h > 0 && w > 0) {
    memsize = h * paddedW / 8;
    /* This is for the users copy of the character, for security-reasons
       the original pointer to the cache area is not used. The entry string_glyph.bits
       is free'ed every time this function is called: */
    string_glyph.bits = (char *)malloc(memsize*sizeof( char));
    if (string_glyph.bits == NULL) 
      return(NULL);
  }
  else {
    h = w = 0;
    area->xmin = area->xmax = 0;
    area->ymax = area->ymax = 0;
  }
  
  string_glyph.metrics.leftSideBearing  = area->xmin;
  string_glyph.metrics.characterWidth   = NEARESTPEL(area->ending.x - area->origin.x);
  string_glyph.metrics.rightSideBearing = w + area->xmin;
  string_glyph.metrics.descent          = area->ymax - NEARESTPEL(area->origin.y);
  string_glyph.metrics.ascent           = h - string_glyph.metrics.descent;

  
  if (h > 0 && w > 0) {
    (void) memset(string_glyph.bits, 0, memsize);
    fill(string_glyph.bits, h, paddedW, area, T1_byte, T1_bit, T1_wordsize );
  }

  /* Cache string_glyph if requested */
  /* if (cache_flag){
    memcpy( &(font_ptr->pFontCache[charcode]), &string_glyph, sizeof(GLYPH));
    font_ptr->pFontCache[charcode].bits = (char *)Xalloc(memsize);
    if (font_ptr->pFontCache[charcode].bits == NULL) 
      return(NULL); */
    /* ... and copy bitmap to cache area: */
  /*    memcpy( font_ptr->pFontCache[charcode].bits, string_glyph.bits, memsize);
  }*/
  
  return(&string_glyph);
}



void fill(dest, h, w, area, byte, bit, wordsize)
     register char *dest;  /* destination bitmap                           */
     int h,w;              /* dimensions of 'dest', w padded               */
     register struct region *area;  /* region to write to 'dest'           */
     int byte,bit;         /* flags; LSBFirst or MSBFirst                  */
     int wordsize;         /* number of bits per word for LSB/MSB purposes */
{
  register struct edgelist *edge;  /* for looping through edges         */
  register char *p;     /* current scan line in 'dest'                  */
  register int y;       /* for looping through scans                    */
  register int wbytes = w / 8;  /* number of bytes in width             */
  register pel *leftP,*rightP;  /* pointers to X values, left and right */
  int xmin = area->xmin;  /* upper left X                               */
  int ymin = area->ymin;  /* upper left Y                               */
  
  for (edge = area->anchor; VALIDEDGE(edge); edge = edge->link->link) {
    
    p = dest + (edge->ymin - ymin) * wbytes;
    leftP = edge->xvalues;
    rightP = edge->link->xvalues;
    
    for (y = edge->ymin; y < edge->ymax; y++) {
      fillrun(p, *leftP++ - xmin, *rightP++ - xmin, bit);
      p += wbytes;
    }
  }
  /* 
    Now, as an afterthought, we'll go reorganize if odd byte order requires
    it:
    */
  /* We do not reorganize since t1lib internally always uses LSBFirst */
  /*
  if ( 0 && wordsize != 8) {
    register int i; 
    printf("Reorganizing data ..\n");
    
    switch (wordsize) { 
    case 16: 
      { 
	register unsigned short data,*p; 
	p = (unsigned short *) dest; 
	for (i = h * w /16; --i >= 0;) { 
	  data = *p; 
	  *p++ = (data << 8) + (data >> 8); 
	} 
	break; 
      } 
    case 64: 
    case 32: 
      { 
	register ULONG data,*p; 
	p = (ULONG *) dest; 
	for (i = h * w / 32; --i >= 0;) { 
	  data = *p; 
	  *p++ = (data << 24) + (data >> 24) 
	    + (0xFF00 & (data >> 8)) 
	    + (0xFF0000 & (data << 8)); 
	} 
	if (wordsize == 64) { 
	  p = (ULONG *) dest; 
	  for (i = h * w / 64; --i >= 0;) { 
	    data = *p++; 
	    p[-1] = p[0]; 
	    *p++ = data; 
	  } 
	}
	break; 
      } 
    default: 
      abort("xiFill: unknown format"); 
    } 
  }
  */
}

#define  ALLONES  0xFF

void fillrun(p, x0, x1, bit)
     register char *p;     /* address of this scan line                    */
     pel x0,x1;            /* left and right X                             */
     int bit;              /* format:  LSBFirst or MSBFirst                */
{
  register int startmask,endmask;  /* bits to set in first and last char*/
  register int middle;  /* number of chars between start and end + 1    */
  
  if (x1 <= x0)
    return;
  middle = x1/8 - x0/8;
  p += x0/8;
  x0 &= 7;  x1 &= 7;
  if (bit == LSBFirst) {
    startmask = ALLONES << x0;
    endmask = ~(ALLONES << x1);
  }
  else {
    startmask = ALLONES >> x0;
    endmask = ~(ALLONES >> x1);
  }
  if (middle == 0)
    *p++ |= startmask & endmask;
  else {
    *p++ |= startmask;
    while (--middle > 0)
      *p++ = ALLONES;
    *p |= endmask;
  }
}



/* T1_CopyGlyph(): Make a copy of an existent glyph-structure to save it
   for later usage by the user. */
GLYPH *T1_CopyGlyph( GLYPH *glyph)
{
  GLYPH *dest;
  long size;
  
  
  /* Assign padding value */
  T1_pad=pFontBase->bitmap_pad;
  /* Allocate memory for struct: */
  if ((dest=(GLYPH *)malloc(sizeof(GLYPH)))==NULL)
    return(NULL);
  /* Copy the structure members: */
  *dest=*glyph;
  /* Allocate memory for bitmap, initialize pointer to it and copy bitmap: */
  size=PAD(dest->metrics.rightSideBearing-dest->metrics.leftSideBearing, T1_pad);
  size=size/8*(dest->metrics.ascent+dest->metrics.descent);
  size=size*glyph->bpp;
  /* We must check whether there's actually a bits-pointer different from
     NULL. If not omit the follwoing step: */
  if (glyph->bits!=NULL){
    if ((dest->bits=(unsigned char *)malloc(size*
					    sizeof(unsigned char)))==NULL)
      {
	free(dest);
	return(NULL);
      }
    memcpy(dest->bits,glyph->bits,size);
  }
  return(dest);
}

/* T1_DumpGlyph(): Dump a glyph-representation to stdout: */
void T1_DumpGlyph( GLYPH *glyph)
{
  int i,j,h,w;   /* i=line-index, j=column-index */ 
  long paddedW;
  
  
  printf("Dataformat: T1_bit=%d, T1_byte=%d, T1_wordsize=%d, T1_pad=%d\n",
	 T1_bit, T1_byte, T1_pad, T1_pad);
  
  h=glyph->metrics.ascent+glyph->metrics.descent;
  w=glyph->metrics.rightSideBearing-glyph->metrics.leftSideBearing;
  
  paddedW=PAD(w,T1_pad);
  
  printf("GlyphInfo: h=%d, w=%d, paddedW=%ld\n", h, w, paddedW);
  
  for ( i=0; i<h; i++){
    if (T1_pad==8)
      for (j=0; j<paddedW/T1_pad; j++)
	bin_dump_c(glyph->bits[i*paddedW/T1_pad+j]);
    else if (T1_pad==16)
      for (j=0; j<paddedW/T1_pad; j++){
	bin_dump_s(((unsigned short *)glyph->bits)[i*paddedW/T1_pad+j]);
      }
    else
      for (j=0; j<paddedW/T1_pad; j++){
	bin_dump_l(((unsigned long *)glyph->bits)[i*paddedW/T1_pad+j]);
      }
    printf("\n");
  } 
} 
