/* -*- Mode: C; tab-width: 4 -*-
   hw.c -- common hardware layer.
   Created: Chris Toshok <toshok@hungry.com>, 3-Jun-1997
*/

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "yshm.h"
#include "ylib_p.h"
#include "hw.h"
#include "dihw.h"

static YHW *devices = NULL;
static int num_devices = 0;

void
yhw_addHardware(YHW *hw, int shm_id)
{
  hw->next = devices;
  hw->prev = NULL;

  hw->shm_id = shm_id;

  devices = hw;

  num_devices ++;
}

YHW *
YHW_GetDevices()
{
  return devices;
}

int
YHW_GetNumDevices()
{
  return num_devices;
}

YHW *
YHW_AllocHW()
{
  return (YHW*)calloc(1, sizeof(YHW));
}

YBool
YHW_HWVersionPresent(int32 library_version)
{
  return library_version == YHW_LIBRARY_VERSION;
}

YBool
YHW_Probe(YHW *hw)
{
  return (*hw->vtable.Probe)(hw);
}

const char *
YHW_GetName(YHW *hw)
{
  return (*hw->vtable.GetName)(hw);
}

void
YHW_SetDriver(YHW *hw, const char *driver)
{
  YHW_sharedData *data =
	(YHW_sharedData*)YSHM_getSharedMemoryById(hw->shm_id,
											  sizeof(YHW_sharedData));

  strcpy((char*)data->driver, driver);

  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
}

const char *
YHW_GetDriver(YHW *hw)
{
  static char return_data[200];

  YHW_sharedData *data =
	(YHW_sharedData*)YSHM_getSharedMemoryById(hw->shm_id,
											  sizeof(YHW_sharedData));

  strcpy(return_data, data->driver);

  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
  
  return return_data;
}

void
YHW_GetSize(YHW *hw,
			uint32 *width,
			uint32 *height)
{
  (*hw->vtable.GetCurrentResolution)(hw, width, height);
}

YBool
YHW_CanDoDPMS(YHW *hw)
{
  return (hw->vtable.HandleDPMS != NULL);
}

void
YHW_HandleDPMS(YHW *hw, int dpms_mode)
{
  (*hw->vtable.HandleDPMS)(hw, dpms_mode);
}

static void
yhw_InitDisplay(YHW *hw,
				int argc,
				char *argv[])
{
  (*hw->vtable.InitDisplay)(hw, argc, argv);
}

void
YHW_InitDisplay(int argc,
				char *argv[])
{
  YRect prev_rect;
  YHW *cur;
  YHW *prev = NULL;

  printf ("Inside YHW_InitDisplay()\n");

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	{
	  uint32 width, height;
	  YHW_sharedData *data =
		(YHW_sharedData*)YSHM_getSharedMemoryById(cur->shm_id,
												  sizeof(YHW_sharedData));

	  yhw_InitDisplay(cur, argc, argv);

	  YHW_GetSize(cur, &width, &height);

	  if (prev)
		{
		  data->display_rect.left = prev_rect.right + 1;
		  data->display_rect.top = 0;
		  data->display_rect.right = prev_rect.right + 1 + width;
		  data->display_rect.bottom = height;
		}
	  else
		{
		  data->display_rect.left = 0;
		  data->display_rect.top = 0;
		  data->display_rect.right = width;
		  data->display_rect.bottom = height;
		}

	  prev = cur;
	  prev_rect = data->display_rect;
	  
	  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
	}
}

static void
yhw_DeinitDisplay(YHW *hw)
{
  (*hw->vtable.DeinitDisplay)(hw);
}

void
YHW_DeinitDisplay()
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_DeinitDisplay(cur);
}

static void
yhw_OpenDisplay(YHW *hw)
{
  (*hw->vtable.OpenDisplay)(hw);
}

void
YHW_OpenDisplay()
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_OpenDisplay(cur);
}

static void
yhw_CloseDisplay(YHW *hw)
{
  (*hw->vtable.CloseDisplay)(hw);
}

void
YHW_CloseDisplay()
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_CloseDisplay(cur);
}

static void
yhw_HandleEvents(YHW *hw)
{
  (*hw->vtable.HandleEvents)(hw);
}

void
YHW_HandleEvents()
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_HandleEvents(cur);
}

static void
yhw_DrawRectangle(YHW *hw,
				  YGC* gc,
				  int32 x, int32 y,
				  uint32 width, uint32 height)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  if (hw->vtable.DrawRectangle)
	(*hw->vtable.DrawRectangle)(hw, gc,
								x + info->bounds.left,
								y + info->bounds.top,
								width, height);
  else
	YDIHW_DrawRectangle(hw, gc,
						x + info->bounds.left,
						y + info->bounds.top,
						width, height);

  YSHM_releaseSharedWindowInfo(info);
}

void
YHW_DrawRectangle(YGC *gc,
				  int32 x, int32 y,
				  uint32 width, uint32 height)
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	{
	  YRect rect;
	  YHW_sharedData *data =
		(YHW_sharedData*)YSHM_getSharedMemoryById(cur->shm_id,
												  sizeof(YHW_sharedData));

	  rect.left = x;
	  rect.top = y;
	  rect.right = x + width;
	  rect.bottom = y + height;

	  if (YDoRectsIntersect(&rect, &data->display_rect))
		yhw_DrawRectangle(cur, gc,
						  x - data->display_rect.left,
						  y - data->display_rect.top,
						  width, height);

	  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
	}
}

static void
yhw_FillRectangle(YHW *hw,
				  YGC* gc,
				  int32 x, int32 y,
				  uint32 width, uint32 height)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  if (hw->vtable.FillRectangle)
	(*hw->vtable.FillRectangle)(hw, gc,
								x + info->bounds.left,
								y + info->bounds.top,
								width, height);
  else
	YDIHW_FillRectangle(hw, gc,
						x + info->bounds.left,
						y + info->bounds.top,
						width, height);

  YSHM_releaseSharedWindowInfo(info);
}

void
YHW_FillRectangle(YGC *gc,
				  int32 x, int32 y,
				  uint32 width, uint32 height)
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	{
	  YRect rect;
	  YHW_sharedData *data =
		(YHW_sharedData*)YSHM_getSharedMemoryById(cur->shm_id,
												  sizeof(YHW_sharedData));

	  rect.left = x;
	  rect.top = y;
	  rect.right = x + width;
	  rect.bottom = y + height;

	  if (YDoRectsIntersect(&rect, &data->display_rect))
		yhw_FillRectangle(cur, gc,
						  x - data->display_rect.left,
						  y - data->display_rect.top,
						  width, height);

	  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
	}
}

static void
yhw_DrawPoints(YHW *hw, YGC *gc, YPoint **points, int npoints)
{
  int i;
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  for (i=0; i<npoints; i++)
	{
	  points[ i ]->x += info->bounds.left;
	  points[ i ]->y += info->bounds.top;
	}

  if (hw->vtable.DrawPoints)
	(*hw->vtable.DrawPoints)(hw, gc, points, npoints);
  else
	YDIHW_DrawPoints(hw, gc, points, npoints);

  for (i=0; i<npoints; i++)
	{
	  points[ i ]->x -= info->bounds.left;
	  points[ i ]->y -= info->bounds.top;
	}

  YSHM_releaseSharedWindowInfo(info);
}


void
YHW_DrawPoints(YGC *gc, YPoint **points, int npoints)
{
  YHW *cur;
  int i;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	{
	  YHW_sharedData *data =
		(YHW_sharedData*)YSHM_getSharedMemoryById(cur->shm_id,
												  sizeof(YHW_sharedData));

	  /* yuck */

	  for (i=0; i<npoints; i++)
		{
		  points[ i ]->x -= data->display_rect.left;
		  points[ i ]->y -= data->display_rect.top;
		}

	  /*
	  if (yhw_pointInRect(&p1, &data->display_rect)
		  || yhw_pointInRect(&p2, &data->display_rect))
		  */

	  yhw_DrawPoints (cur, gc, points, npoints);

	  for (i=0; i<npoints; i++)
		{
		  points[ i ]->x += data->display_rect.left;
		  points[ i ]->y += data->display_rect.top;
		}

	  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
	}
}

static void
yhw_DrawLine(YHW *hw,
			 YGC* gc,
			 int32 x1, int32 y1,
			 int32 x2, int32 y2)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  
  if (hw->vtable.DrawLine)
	(*hw->vtable.DrawLine)(hw, gc,
						   x1 + info->bounds.left,
						   y1 + info->bounds.top,
						   x2 + info->bounds.left,
						   y2 + info->bounds.top);
  else
	YDIHW_DrawLine(hw, gc,
				   x1 + info->bounds.left,
				   y1 + info->bounds.top,
				   x2 + info->bounds.left,
				   y2 + info->bounds.top);
  
  YSHM_releaseSharedWindowInfo(info);
}

void
YHW_DrawLine(YGC *gc,
			 int32 x1, int32 y1,
			 int32 x2, int32 y2)
{
  YHW *cur;

  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	{
	  YPoint p1, p2;
	  YHW_sharedData *data =
		(YHW_sharedData*)YSHM_getSharedMemoryById(cur->shm_id,
												  sizeof(YHW_sharedData));

	  p1.x = x1; p1.y = y1;
	  p2.x = x2; p2.y = y2;

	  if (YIsPointInRect(&p1, &data->display_rect)
		  || YIsPointInRect(&p2, &data->display_rect))

		yhw_DrawLine(cur, gc,
					 x1 - data->display_rect.left,
					 y1 - data->display_rect.top,
					 x2 - data->display_rect.left,
					 y2 - data->display_rect.top);

	  YSHM_releaseSharedMemory(data, sizeof(YHW_sharedData));
	}
}

void
YHW_DrawArc(YGC *gc,
			int32 x, int32 y,
			uint32 width, uint32 height)
{
  assert(0);
}

void
YHW_DrawBezier(YGC *gc,
			   YPoint **points,
			   int npoints)
{
  assert(0);
}

static void
yhw_CopyFromWin(YHW *hw, YGC *gc,
				int32 src_x, int32 src_y,
				int32 width, int32 height,
				YColor **buffer)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  (*hw->vtable.CopyFromWin)(hw, gc,
							src_x + info->bounds.left,
							src_y + info->bounds.top,
							width, height,
							buffer);

  YSHM_releaseSharedWindowInfo(info);
}

static void
yhw_CopyToWin(YHW *hw, YGC *gc,
			  int32 dest_x, int32 dest_y,
			  int32 width, int32 height,
			  YColor *buffer)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  (*hw->vtable.CopyToWin)(hw, gc,
						  dest_x + info->bounds.left,
						  dest_y + info->bounds.top,
						  width, height,
						  buffer);

  YSHM_releaseSharedWindowInfo(info);
}

static void
yhw_CopyFromWinToWin(YHW *hw, YGC *gc, YWindow *source,
					 int32 src_x, int32 src_y,
					 int32 dest_x, int32 dest_y,
					 int32 width, int32 height)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  
  (*hw->vtable.CopyFromWinToWin)(hw, gc, source,
								 src_x + info->bounds.left,
								 src_y + info->bounds.top,
								 dest_x + info->bounds.left,
								 dest_y + info->bounds.top,
								 width, height);

  YSHM_releaseSharedWindowInfo(info);
}

void
YHW_CopyToWin(YGC *gc,
			  int32 dest_x, int32 dest_y,
			  int32 width, int32 height,
			  YColor *buffer)
{
  YHW *cur;
  
  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_CopyToWin(cur, gc,
				  dest_x, dest_y,
				  width, height,
				  buffer);
}

void
YHW_CopyFromWin(YGC *gc,
				int32 src_x, int32 src_y,
				int32 width, int32 height,
				YColor **buffer)
{
  YHW *cur;
  
  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_CopyFromWin(cur, gc,
					src_x, src_y,
					width, height,
					buffer);
}

void
YHW_CopyFromWinToWin(YGC *gc, YImage *source,
					 int32 src_x, int32 src_y,
					 int32 dest_x, int32 dest_y,
					 uint32 width, uint32 height)
{
  /* this routine needs to be a *lot* smarter.  For example, if we're blitting between hardware
	 types, we need to call the CopyFromWin/CopyToWin pairs.  Also, if a window is stradling the
	 boundary between hardware devices, we need to cope with that. */

  /* gross, icky hack. XXXX fix me please oh please. */
  YHW *cur;
  
  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_CopyFromWinToWin(cur, gc, source,
						 src_x, src_y,
						 dest_x, dest_y,
						 width, height);
}

static void
yhw_MoveCursor(YHW *hw, int32 x, int32 y)
{
  (*hw->vtable.MoveCursor)(hw, x, y);
}

void
YHW_MoveCursor(int32 x, int32 y)
{
}

static void
yhw_ShowCursor(YHW *hw)
{
  (*hw->vtable.ShowCursor)(hw);
}

void
YHW_ShowCursor()
{
}

static void
yhw_HideCursor(YHW *hw)
{
  (*hw->vtable.ShowCursor)(hw);
}

void
YHW_HideCursor()
{
  YHW *cur;
  
  for (cur = devices;
	   cur != NULL;
	   cur = cur->next)
	yhw_HideCursor(cur);
}
