/* -*- Mode: C; tab-width: 4 -*-
   scan_hw.c -- finds a list of applicable drivers and runs down the list,
                probing them.
   Created: Chris Toshok <toshok@hungry.com>, 30-Jun-1997
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>

#include "yshm.h"
#include "hw.h"
#include "ytypes.h"
#include "ydll.h"
#include "pci_hw_p.h"

/* everyone gets the misc drivers... */
#define MISC_DRIVERS

static int hw_shm_id = 0;

/* Return true if file is dynamic library. Only skips directories */
static int
yhw_dynalib(
#if defined(linux) || defined(hpux)
			const
#endif
			struct dirent *file)
{
  int len = strlen(file->d_name);

  if (len > 3
	  && file->d_name[len - 3] == '.'
	  && file->d_name[len - 2] == 's'
	  && file->d_name[len - 1] == 'o')
	return 1;
  else
	return 0;
}

static char **
yhw_generateDriverPaths(char *driver_type)
{
  char *dir_path;
  struct dirent **namelist;
  int i;
  int num_entries;
  char **driver_paths;

  dir_path = malloc(strlen(DRIVER_PATH "/")+strlen(driver_type)+1);
  strcpy(dir_path, DRIVER_PATH "/");
  strcat(dir_path, driver_type);

  num_entries = scandir(dir_path, &namelist, yhw_dynalib, alphasort);

  if (num_entries <= 0)
    return NULL;

  driver_paths = (char**)calloc(num_entries + 1, sizeof(char*));
  for (i = 0; i < num_entries; i ++)
    {
      driver_paths[i] = (char*)calloc(strlen(dir_path) + strlen(namelist[i]->d_name) + 1, sizeof(char));
      strcpy(driver_paths[i], dir_path);
      strcat(driver_paths[i], namelist[i]->d_name);
    }

  for (i = 0; i < num_entries; i ++)
	free(namelist[i]);

  free(namelist);
  free(dir_path);

  return driver_paths;
}

void
yhw_freeDriverPaths(char **paths)
{
  int i = 0;

  while (paths[i] != 0)
    free (paths[i++]);

  free(paths);
}

typedef YHW* (*YHW_initfunc)();

static YBool
yhw_probeDrivers(char **drivers, YBool find_all)
{
  YBool found = Y_FALSE;
  int i = 0;
  char *driver;
  
  while ((driver = drivers[i++]) != NULL)
    {
      YDLL *driver_lib = YDLL_LoadLibrary(driver);
      YHW_initfunc yhw_init = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
															  "YHW_init");
      YHW_initfunc yhw_deinit;
      YHW *hw;

      if (!driver_lib || !yhw_init)
		continue;
	  
      hw = (*yhw_init)();

	  if (!hw) continue;

      printf ("+Probing for -- %s....", YHW_GetName(hw));
      fflush(stdout);
      found = YHW_Probe(hw);
      
      if (found)
		{
		  YSHM_allocateSharedMemoryById(hw_shm_id,
										sizeof(YHW_sharedData));

		  printf (" found.\n");

		  yhw_addHardware(hw, hw_shm_id);

		  hw_shm_id++;

		  YHW_SetDriver(hw, driver);

		  if (!find_all)
			break;
		}
      else
		{
		  printf (" not found.\n");
		  yhw_deinit = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
													   "YHW_deinit");
		  (*yhw_deinit)(hw);
		  
		  YDLL_UnloadLibrary(driver_lib);
		}
    }
  
  return found;
}

/*
** this one loads a particular driver.
*/
void
YHW_LoadHardware(char *driver)
{
  char full_driver_path[1024];
  YDLL *driver_lib;
  YHW_initfunc yhw_init;
  YHW_initfunc yhw_deinit;
  YBool found = Y_FALSE;
  YHW *hw;

#ifdef HAVE_PCI
  YPCI_init();
#endif

  strcpy (full_driver_path, DRIVER_PATH "/");
  strcat (full_driver_path, driver);

  driver_lib = YDLL_LoadLibrary(full_driver_path);

  if (!driver_lib)
	{
	  char *dir_name, *driver_name;
	  char *slash = NULL;

	  slash = strrchr(driver, '/');
	  
	  if (slash)
		{
		  dir_name = driver;
		  *slash = 0;
		  driver_name = slash + 1;
		}
	  else
		{
		  dir_name = "";
		  driver_name = driver;
		}

	  fprintf (stderr, "attemping to load: '%s/%s/libyhw_%s.so'\n",
			   DRIVER_PATH, dir_name, driver_name);

	  sprintf(full_driver_path, "%s/%s/libyhw_%s.so",
			  DRIVER_PATH, dir_name, driver_name);
	  
	  driver_lib = YDLL_LoadLibrary(full_driver_path);
	  
	  if(slash)
		*slash = '/';
	}

  yhw_init = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
											 "YHW_init");
  
  if (!driver_lib || !yhw_init)
	{
	  printf ("Not a well formed driver.\n");
	  exit(1);
	}
  
  hw = (*yhw_init)();
  if (!hw)
	{
	  printf ("Not a well formed driver.\n");
	  exit(1);
	}

  printf ("+Sanity probing for -- %s....", YHW_GetName(hw));
  fflush(stdout);
  found = YHW_Probe(hw);

  if (found)
	{
	  YSHM_allocateSharedMemoryById(hw_shm_id,
									sizeof(YHW_sharedData));
	  
	  printf (" probing successful.\n");
	  
	  yhw_addHardware(hw, hw_shm_id);

	  hw_shm_id ++;

	  YHW_SetDriver(hw, full_driver_path);
	}
  else
	{
	  printf (" probing failed.\n");
	  yhw_deinit = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
												   "YHW_deinit");
	  (*yhw_deinit)(hw);
	  
	  YDLL_UnloadLibrary(driver_lib);
	}
}

/*
** This function is called by the clients to initialize an already existing
** hardware structure from it's shared memory id.
*/
void
YHW_InitHardware(int shm_id)
{
  YHW *hw;
  YDLL *driver_lib;
  YHW_initfunc yhw_init;
  YHW_initfunc yhw_deinit;
  YHW_sharedData *data =
	(YHW_sharedData*)YSHM_getSharedMemoryById(shm_id,
											  sizeof(YHW_sharedData));
  YBool found;

  driver_lib = YDLL_LoadLibrary(data->driver);

  yhw_init = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
											 "YHW_init");
  
  if (!driver_lib || !yhw_init)
	{
	  printf ("Not a well formed driver.\n");
	  exit(1);
	}
  
  hw = (*yhw_init)();
  if (!hw)
	{
	  printf ("Not a well formed driver.\n");
	  exit(1);
	}

  printf ("+Sanity probing for -- %s....", YHW_GetName(hw));
  fflush(stdout);
  found = YHW_Probe(hw);

  if (found)
	{
	  printf (" probing successful.\n");
	  
	  hw->shm_id = shm_id;
	  
	  yhw_addHardware(hw, shm_id);
	}
  else
	{
	  printf (" probing failed.\n");
	  yhw_deinit = (YHW_initfunc)YDLL_LookupSymbol(driver_lib,
												   "YHW_deinit");
	  (*yhw_deinit)();
	  
	  YDLL_UnloadLibrary(driver_lib);
	  exit(1); /* should we exit here? */
	}
}

/*
** this one finds the first available hardware driver -- the first one that
** probes TRUE, and returns.
*/
void
YHW_FindHardware(YBool find_all)
{
  YBool found = Y_FALSE;

  if (!found || find_all)
  {
    char **drivers = yhw_generateDriverPaths("");

    if (drivers)
      {
		found = yhw_probeDrivers(drivers, find_all);
		
		yhw_freeDriverPaths(drivers);
      }
  }

#ifdef HAVE_PCI

  YPCI_init();

  if (!found || find_all)
  {
    char **drivers = yhw_generateDriverPaths("pci/");

    if (drivers)
      {
		found = yhw_probeDrivers(drivers, find_all);
		
		yhw_freeDriverPaths(drivers);
      }
  }
#endif /* HAVE_PCI */
  
#ifdef MISC_DRIVERS
  if (!found || find_all)
	{
	  char **drivers = yhw_generateDriverPaths("misc/");
	  
	  if (drivers)
		{
		  found = yhw_probeDrivers(drivers, find_all);
		  
		  yhw_freeDriverPaths(drivers);
		}
	}
#endif
  
  if (!found)
    {
      printf ("Couldn't find any hardware.\n");
      exit(1);
    }
}

void
YHW_CleanupSharedMemory()
{
  int i;

  for (i = 0; i < hw_shm_id; i ++)
	YSHM_destroySharedMemoryById(i);
}

