/* -*- Mode: C; tab-width: 4 -*-
   hw.h -- hardware interface.
   Created: Chris Toshok <toshok@hungry.com>, 11-May-1997
 */

#ifndef _y_hw_h
#define _y_hw_h

#include "ytypes.h"

#define YHW_LIBRARY_VERSION 2

typedef struct YHW YHW;

typedef struct {
  uint32 width;
  uint32 height;
} YHW_resolution;

/* The values for the DPMS mode. */
enum {
  DPMSModeOn,
  DPMSModeStandby,
  DPMSModeSuspend,
  DPMSModeOff
};

typedef struct YHW_vtable {
  /*
  ** Return a pretty name for the driver.
  */
  const char* (*GetName)(YHW *hw);

  /*
  ** Returns Y_TRUE if the hardware is found.  No persistent initialization
  ** should be done in this routine, since the hardware type might not be
  ** used.
  */
  YBool (*Probe)(YHW *hw);

  /*
  ** Resolution negotiation.
  **
  ** GetResolutions returns a list of width x height pairs that the hardware
  **                recommends.
  ** SetResolution  takes a pair of width x height.  If the resolution is
  **                acceptable, the hardware makes the change and returns
  **                Y_TRUE.  Otherwise, it does nothing but return Y_FALSE.
  **                It always returns Y_TRUE for one of the resolutions listed
  **                in the return value of GetResolutions.
  ** GetCurrentResolution Gets the current resolution of the driver, or the
  **                      default if the hardware has not been set to a 
  **                      particular width/height yet.
  */
  void (*GetResolutions)(YHW *hw, YHW_resolution** resolutions, int *num_res);
  YBool (*SetResolution)(YHW *hw, uint32 width, uint32 height);
  void (*GetCurrentResolution)(YHW *hw, uint32* width, uint32* height);

  /*
  ** Display stuff.
  ** InitDisplay - called by the microserver to actually initialize the
  **               the hardware.
  ** DeinitDisplay - called by the microserver to actually clean up the hardware.
  ** OpenDisplay - called by all the clients to attach to the hardware.
  ** CloseDisplay - called by everyone to signal that they are done with
  **                the display.
  */
  void (*InitDisplay)(YHW *hw, int argc, char *argv[]);
  void (*DeinitDisplay)(YHW *hw);

  void (*OpenDisplay)(YHW *hw);
  void (*CloseDisplay)(YHW *hw);

  void (*HandleEvents)(YHW *hw);

  /*
  ** Cursor functions (NOTE:  the cursor_image can contain alpha. deal.)
  */
  void (*SetCursor)(YHW *hw,
					YImage *cursor_image,
					int32 hot_x, int32 hot_y);
  void (*MoveCursor)(YHW *hw, int32 x, int32 y);
  void (*ShowCursor)(YHW *hw);
  void (*HideCursor)(YHW *hw);

  /*
  ** Display Power Management -- screen blanking. NULL if the driver
  ** doesn't support it.
  */
  void (*HandleDPMS)(YHW *hw, int dpms_mode);

  /* Blitting routines.  For blits to different display types, an intervening
	 copy needs to take place.  This is done by the CopyFromWin and CopyToWin
	 calls. The CopyFromWinToWin call is for windows that are being handled
	 by the same display type and hardware device. */
  
  void (*CopyFromWin)(YHW *hw, YGC *gc,
					  int32 src_x, int32 src_y,
					  int32 width, int32 height,
					  YColor **buffer);
  void (*CopyToWin)(YHW *hw, YGC *gc,
					int32 dest_x, int32 dest_y,
					int32 width, int32 height,
					YColor *buffer);
  void (*CopyFromWinToWin)(YHW *hw, YGC *gc, YWindow *source,
						   int32 src_x, int32 src_y,
						   int32 dest_x, int32 dest_y,
						   int32 width, int32 height);

  /*
  ** The only required hardware drawing entry.
  */
  void (*DrawSpan)(YHW *hw, YGC *gc, int32 x1, int32 x2, int32 y);

  /*
  ** These are optional.  The HW library will use device independent versions
  ** of these if they are NULL, which call DrawSpan.
  */
  void (*DrawRectangle)	(YHW *hw, YGC *gc, int32 x, int32 y,
						 uint32 width, uint32 height);
  void (*FillRectangle)	(YHW *hw, YGC *gc, int32 x, int32 y,
						 uint32 width, uint32 height);
  void (*DrawPoints)	(YHW *hw, YGC *gc,
						 YPoint **points, int npoints);
  void (*DrawLine)		(YHW *hw, YGC *gc,
						 int32 x1, int32 y1, int32 x2, int32 y2);
  void (*DrawArc)		(YHW *hw, YGC *gc,
						 int32 x, int32 y, uint32 width, uint32 height);
  void (*DrawBezier)	(YHW *hw, YGC *gc,
						 YPoint **points, int npoints);
} YHW_vtable;

/*
** The structure kept in a shared memory segment.  This allows clients to
** get access to the display rectangle and driver .so path without requiring
** changes to the microserver/client protocol.
*/
typedef struct {
  int sem_id;

  /* this display's area of the "virtual" screen */
  YRect display_rect;

  /* the shared object to load for this device. */
  const char driver[300];
} YHW_sharedData;

struct YHW {
  YHW_vtable vtable;

  void *hw_data;

  int shm_id;

  /* used by the hardware management routines */
  YHW *next;
  YHW *prev;
};

extern YHW *YHW_GetDevices();
extern int YHW_GetNumDevices();

/* These two functions are implemented in each of the respective hardware 
   implementations, and is responsible for initializing the hardware vtable
   -- listed below. */
extern YHW *YHW_init();
extern void YHW_deinit(YHW *hw);

/* Calls to make sure drivers work with multiple versions of the library.
   YHW_allowHW() allocates the YHW struct (as well as the vtable.)  The 
   individual drivers should memcpy their vtable into the space for it.
   YHW_HWVersionPresent() lets the HW driver writer specify that the driver
   should not be run with an older version of the driver. */
extern YHW *YHW_AllocHW();
extern YBool YHW_HWVersionPresent(int32 library_version);

extern YBool YHW_Probe(YHW *hw);
extern const char *YHW_GetName(YHW *hw);
extern void YHW_SetDriver(YHW *hw, const char *driver);
extern const char *YHW_GetDriver(YHW *hw);
extern void YHW_GetSize(YHW *hw, uint32 *width, uint32 *height);

extern void YHW_InitDisplay(int argc, char *argv[]);
extern void YHW_DeinitDisplay();
extern void YHW_OpenDisplay();
extern void YHW_CloseDisplay();
extern void YHW_HandleEvents();

extern void YHW_SetCursor(YImage *cursor_image,
						  int32 hot_x, int32 hot_y);
extern void YHW_MoveCursor(int32 x, int32 y);
extern void YHW_ShowCursor();
extern void YHW_HideCursor();

extern void YHW_DrawRectangle(YGC *gc, int32 x, int32 y,
							  uint32 width, uint32 height);
extern void YHW_FillRectangle(YGC *gc, int32 x, int32 y,
							  uint32 width, uint32 height);
extern void YHW_DrawPoints(YGC *gc, YPoint **points, int npoints);
extern void YHW_DrawLine(YGC *gc, int32 x1, int32 y1, int32 x2, int32 y2);
extern void YHW_DrawArc	(YGC *gc,
						 int32 x, int32 y, uint32 width, uint32 height);
extern void YHW_DrawBezier(YGC *gc,
						   YPoint **points, int npoints);

extern void YHW_CopyFromWin(YGC *gc,
							int32 src_x, int32 src_y,
							int32 width, int32 height,
							YColor **buffer);
extern void YHW_CopyToWin(YGC *gc,
						  int32 dest_x, int32 dest_y,
						  int32 width, int32 height,
						  YColor *buffer);
extern void YHW_CopyFromWinToWin(YGC *gc, YWindow *source,
								 int32 src_x, int32 src_y,
								 int32 dest_x, int32 dest_y,
								 uint32 width, uint32 height);

extern YBool YHW_CanDoDPMS(YHW *hw);
extern void YHW_HandleDPMS(YHW *hw, int dpms_mode);

extern void YHW_FindHardware(YBool find_all);
extern void YHW_LoadHardware(char *driver);
extern void YHW_InitHardware(int shm_id);
extern void YHW_CleanupSharedMemory();

void yhw_addHardware(YHW *hw, int shm_id);


#endif /* _y_hw_h */
