/* -*- Mode: C; tab-width: 4 -*-
   xhw.c -- X11 hardware layer.
   Created: Chris Toshok <toshok@hungry.com>, 3-Jun-1997
*/

#include "hw.h"
#include "ylib_p.h"
#include "ytypes.h"
#include "yshm.h"
#include "ymessages.h"

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/fcntl.h>

#define X_ROOT_WINDOW_SHM_ID "x-root-window"
#define X_HW_WIDTH	800
#define X_HW_HEIGHT 600

#define DRIVER_VERSION "$Revision: 1.21 $"

/* our vtable */
static YBool x_probe(YHW *hw);
static const char *x_getName(YHW *hw);
static void x_getResolutions(YHW *hw,
							 YHW_resolution** resolutions, int *num_res);
static YBool x_setResolution(YHW *hw, uint32 width, uint32 height);
static void x_getCurrentResolution(YHW *hw, uint32 *width, uint32 *height);
static void x_initDisplay(YHW *hw, int argc, char *argv[]);
static void x_deinitDisplay(YHW *hw);
static void x_openDisplay(YHW *hw);
static void x_closeDisplay(YHW *hw);
static void x_handleEvents(YHW *hw);
static void x_drawRectangle(YHW *hw, YGC* gc, int32 x, int32 y,
							uint32 width, uint32 height);
static void x_fillRectangle(YHW *hw, YGC* gc, int32 x, int32 y,
							uint32 width, uint32 height);
static void x_drawPoints(YHW *hw, YGC *gc, YPoint **points, int npoints);
static void x_drawLine(YHW *hw, YGC* gc, int32 x1, int32 y1,
					   int32 x2, int32 y2);
static void x_drawArc(YHW *hw, YGC *gc,
					  int32 center_x, int32 center_y,
					  uint32 width, uint32 height);

static void x_drawSpan(YHW *hw, YGC *gc, int32 x1, int32 x2, int32 y);

static void x_setcursor(YHW *hw,
						YImage *cursor_image,
						int32 hot_x, int32 hot_y);
static void x_movecursor(YHW *hw, int32 x, int32 y);
static void x_showcursor(YHW *hw);
static void x_hidecursor(YHW *hw);

static unsigned long ycolor_to_pixel(YHW *hw, YColor ycolor);

static void x_copyFromWin(YHW *hw, YGC *gc,
						  int32 src_x, int32 src_y,
						  int32 width, int32 height,
							YColor **buffer);
static void x_copyToWin(YHW *hw, YGC *gc,
						int32 dest_x, int32 dest_y,
						int32 width, int32 height,
						YColor *buffer);
static void x_copyFromWinToWin(YHW *hw, YGC *gc, YWindow *source,
							   int32 src_x, int32 src_y,
							   int32 dest_x, int32 dest_y,
							   int32 width, int32 height);

static YHW_vtable x_hw_vtable = {
  x_getName,
  x_probe,

  x_getResolutions,
  x_setResolution,
  x_getCurrentResolution,

  x_initDisplay,
  x_deinitDisplay,
  x_openDisplay,
  x_closeDisplay,


  x_handleEvents,

  x_setcursor,
  x_movecursor,
  x_showcursor,
  x_hidecursor,

  NULL,					 /* no dpms stuff */

  x_copyFromWin,
  x_copyToWin,
  x_copyFromWinToWin,

  x_drawSpan,

  x_drawRectangle,
  x_fillRectangle,
  x_drawPoints,
  x_drawLine,
  x_drawArc,
  NULL					 /* inherit the device independent bezier function. */
};

typedef struct {
  Display *x_display;	/* our X display */
  Window x_window; 		/* our X root window */
  Colormap x_colormap;	/* our screen's colormap */
  GC x_gc;				/* our X gc. */

  uint32 width;
  uint32 height;

  YHW_sharedData *shared_data;
} x_hw_data;

static int num_x_devices = 0;

YHW *
YHW_init()
{
  YHW *hw = NULL;

 /* we need to match the current hw library version. */
  if (YHW_HWVersionPresent(YHW_LIBRARY_VERSION))
	{
	  hw = YHW_AllocHW();
	  
	  memcpy(&hw->vtable, &x_hw_vtable, sizeof(x_hw_vtable));
	  
	  hw->hw_data = (void*)calloc(1, sizeof(x_hw_data));
	}
	  
  return hw;
}

void
YHW_deinit(YHW *hw)
{
  free(hw->hw_data);
  free(hw);
}

static unsigned long
ycolor_to_pixel(YHW *hw, YColor ycolor)
{
  XColor color;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  color.pixel = 0;

  /* skip the alpha */
  ycolor >>= 8;

  color.blue = ycolor & 0xFF;
  color.blue = color.blue << 8 | color.blue;
  ycolor >>= 8;

  color.green = ycolor & 0xFF;
  color.green = color.green << 8 | color.green;
  ycolor >>= 8;

  color.red = ycolor & 0xFF;
  color.red = color.red << 8 | color.red;

  XAllocColor(xhw->x_display, xhw->x_colormap, &color);

  return color.pixel;
}

static YColor
pixel_to_ycolor(YHW *hw, long pixel)
{
  YColor ycolor;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Visual *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(xhw->x_display));

  ycolor = (pixel & v->red_mask) >> 8;

  ycolor <<= 8;

  ycolor |= (pixel & v->green_mask) >> 8;

  ycolor <<= 8;

  ycolor |= (pixel & v->blue_mask) >> 8;

  ycolor <<= 8;

  return ycolor;
}

static YBool
x_probe(YHW *hw)
{
  Display *foo = XOpenDisplay(NULL);
  YBool result = Y_FALSE;

  if (foo)
	{
	  result = Y_TRUE;
	  XCloseDisplay(foo);
	}

  return result;
}

static const char *
x_getName(YHW *hw)
{
  return "X11 Driver :" DRIVER_VERSION;
}

static YHW_resolution x_resolutions[] = {
  { 640, 480 },
  { 800, 600 },
  { 1024, 768 },
  { 1280, 1024 }
};
static int x_num_res = sizeof(x_resolutions) / sizeof(x_resolutions[0]);

static void
x_getResolutions(YHW *hw,
				 YHW_resolution** resolutions,
				 int *num_res)
{
  *resolutions = x_resolutions;
  *num_res = x_num_res;
}

static YBool
x_setResolution(YHW *hw,
				uint32 width,
				uint32 height)
{
  if (width < 320 || width > 1600)
	return Y_FALSE;
  else if (height < 200 || height > 1200)
	return Y_FALSE;


  /* do something cool, like XResizeWindow. */

  return Y_TRUE;
}

static void
x_getCurrentResolution(YHW *hw,
					   uint32 *width,
					   uint32 *height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  
  *width = xhw->width;
  *height = xhw->height;
}

static void
x_handleEvents(YHW *hw)
{
  XEvent event;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  while (XCheckMaskEvent(xhw->x_display,
						 ~0L,
						 &event))
	{
	  /*YWindow *target_window;*/ /* unused...? */
	  static Y_Event y_event;

	  printf ("got an X event\n");

	  y_event.any_event.target = 0/*target_window->shared_window; XXXX */;

	  switch(event.xany.type)
		{
		case Expose:
		  y_event.expose.tag = Y_ExposeEvent;
		  y_event.expose.target = (YWindow_id)~0;
		  y_event.expose.bounds.left = (event.xexpose.x 
										+ xhw->shared_data->display_rect.left);
		  y_event.expose.bounds.right = (event.xexpose.x 
									 + event.xexpose.width 
									 + xhw->shared_data->display_rect.left);
		  y_event.expose.bounds.top = (event.xexpose.y 
									   + xhw->shared_data->display_rect.top);
		  y_event.expose.bounds.bottom = (event.xexpose.y 
									  + event.xexpose.height 
									  + xhw->shared_data->display_rect.top);
		  y_event.expose.timestamp = 0; /* XXX */
		  break;
		case KeyPress:
		  y_event.key_down.tag = Y_KeyDownEvent;
		  y_event.key_down.timestamp = event.xkey.time;
		  y_event.key_down.x = event.xkey.x;
		  y_event.key_down.y = event.xkey.y;
		  break;
		case KeyRelease:
		  y_event.key_up.tag = Y_KeyUpEvent;
		  y_event.key_up.timestamp = event.xkey.time;
		  y_event.key_up.x = event.xkey.x;
		  y_event.key_up.y = event.xkey.y;
		  break;
		case ButtonPress:
		  y_event.mouse_down.tag = Y_MouseDownEvent;
		  y_event.mouse_down.button = event.xbutton.button;
		  y_event.mouse_down.x = event.xbutton.x;
		  y_event.mouse_down.y = event.xbutton.y;
		  y_event.mouse_down.timestamp = event.xbutton.time;
		  break;
		case ButtonRelease:
		  y_event.mouse_up.tag = Y_MouseUpEvent;
		  y_event.mouse_up.button = event.xbutton.button;
		  y_event.mouse_up.x = event.xbutton.x;
		  y_event.mouse_up.y = event.xbutton.y;
		  y_event.mouse_up.timestamp = event.xbutton.time;
		  break;
		case MotionNotify:
		  y_event.mouse_motion.tag = Y_MouseMotionEvent;
		  y_event.mouse_motion.timestamp = event.xbutton.time;
		  y_event.mouse_motion.x = event.xmotion.x;
		  y_event.mouse_motion.y = event.xmotion.y;
		  y_event.mouse_motion.timestamp = event.xmotion.time;
		  break;
		default:
		  printf ("Unhandled event\n");
		  return;
		  break;
		}

	  Y_handleEvent(&y_event);
	}
}

static void
x_initializeRootWindow(YHW *hw)
{
  char name[100];

  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  xhw->x_window = XCreateWindow(xhw->x_display,
								DefaultRootWindow(xhw->x_display),
								0, 0,
								xhw->width,
								xhw->height,
								0, CopyFromParent,
								InputOutput,
								CopyFromParent,
								0L, NULL);

  XSelectInput(xhw->x_display, xhw->x_window, 0xffffff /*everything*/);

  sprintf(name, "Y Display Server %d", num_x_devices);
  XStoreName(xhw->x_display, xhw->x_window, name);

  XMapRaised(xhw->x_display, xhw->x_window);
}

static void
_handle_xevent(void *arg)
{
  x_handleEvents((YHW*)arg);
}

static int (*old_x_error_handler)(Display *, XErrorEvent *);

static int
xhw_error_handler(Display *d, XErrorEvent *event)
{
  (*old_x_error_handler)(d, event);

  abort();
}

static int
xhw_ioerror_handler(Display *d)
{
  fprintf (stderr, "x io error.\n");
  abort();
}

/* called by the microserver */
static void
x_initDisplay(YHW *hw, int argc, char *argv[])
{
  Window *window;
  int i;
  int status, dummy_1, dummy_2;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  xhw->shared_data = YSHM_getSharedMemoryById(hw->shm_id,
											  sizeof(YHW_sharedData));

  printf ("inside x_initDisplay()\n");

  xhw->x_display = XOpenDisplay(NULL);
  assert(xhw->x_display != NULL);

  XSynchronize(xhw->x_display, True);

  xhw->x_colormap =
	DefaultColormapOfScreen(DefaultScreenOfDisplay(xhw->x_display));

  old_x_error_handler = XSetErrorHandler(xhw_error_handler);
  XSetIOErrorHandler(xhw_ioerror_handler);

  xhw->width = X_HW_WIDTH;
  xhw->height = X_HW_HEIGHT;

  i = 0;
  while (i < argc && strcmp(argv[i++], "-geometry") != 0)
	;

  if (i < argc)
	{
	  status = XParseGeometry(argv[i], &dummy_1, &dummy_2,
							  &xhw->width, &xhw->height);
	  if ((status & WidthValue) == 0)
		{
		  xhw->width = X_HW_WIDTH;
		  if (xhw->width < 320 || xhw->width > 1600)
			xhw->width = X_HW_WIDTH;
		}
	  if ((status & HeightValue) == 0)
		{
		  xhw->height = X_HW_HEIGHT;
		  if (xhw->height < 200 || xhw->height > 1200)
			xhw->height = X_HW_HEIGHT;
		}
	  if ((status & (WidthValue | HeightValue)) == 0)
		{
		  printf("Malformed -geometry argument, ignored.\n");
		  xhw->height = X_HW_HEIGHT;
		  xhw->width = X_HW_WIDTH;
		}
	  
	  printf("geometry is %dx%d\n", xhw->width, xhw->height);
	}

  x_initializeRootWindow(hw);
	  
  YAddInput(XConnectionNumber(xhw->x_display),
			_handle_xevent,
			hw);

  {
	char name[100];

	sprintf(name, "%s-%d", X_ROOT_WINDOW_SHM_ID, num_x_devices++);

	window = YSHM_allocateSharedMemory(name,
									   sizeof(Window));
  }

  if (window == NULL)
	{
	  printf ("Failed to allocate shared segment for X root window.\n");
	  exit(1);
	}

  *window = xhw->x_window;

  YSHM_releaseSharedMemory(window, sizeof(Window));

  printf ("leaving x_initDisplay()\n");
}

static void
x_deinitDisplay(YHW *hw)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  XFreeGC(xhw->x_display, xhw->x_gc);
  XFreeColormap(xhw->x_display, xhw->x_colormap);
  XDestroyWindow(xhw->x_display, xhw->x_window);
  XCloseDisplay(xhw->x_display);
}

/* called by the client side libraries */
static void
x_openDisplay(YHW *hw)
{
  Window *window;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  printf ("inside x_openDisplay()\n");

  xhw->x_display = XOpenDisplay(NULL);

  assert(xhw->x_display != NULL);

  XSynchronize(xhw->x_display, True);

  xhw->shared_data = YSHM_getSharedMemoryById(hw->shm_id,
											  sizeof(YHW_sharedData));

  xhw->x_colormap =
	DefaultColormapOfScreen(DefaultScreenOfDisplay(xhw->x_display));

  {
	char name[100];

	sprintf(name, "%s-%d", X_ROOT_WINDOW_SHM_ID, num_x_devices++);

	window = YSHM_getSharedMemory(name,
								  sizeof(Window));
  }

  if (window == NULL)
	{
	  printf ("Failed to get shared segment for X root window.\n");
	  exit(1);
	}

  xhw->x_window = *window;

  YSHM_releaseSharedMemory(window, sizeof(Window));

  printf ("leaving x_openDisplay()\n");
}

static void
x_closeDisplay(YHW *hw)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  YSHM_releaseSharedMemory(xhw->shared_data, sizeof(YHW_sharedData));

  XCloseDisplay(xhw->x_display);
}

static Region
x_getClipRegionFromWindowBounds(YHW *hw, YWindow *window)
{
  Region region = XCreateRegion();
  Region region_return = XCreateRegion();
  XRectangle rect;
  int id;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  rect.x = info->bounds.left;
  rect.y = info->bounds.top;
  rect.width = info->bounds.right - info->bounds.left;
  rect.height = info->bounds.bottom - info->bounds.top;

  if (rect.x > xhw->shared_data->display_rect.left)
	rect.x -= xhw->shared_data->display_rect.left;
  else
	rect.x = 0;

  if (rect.y > xhw->shared_data->display_rect.top)
	rect.y -= xhw->shared_data->display_rect.top;
  else
	rect.y = 0;

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);

  XUnionRectWithRegion(&rect, region, region_return);

  XDestroyRegion(region);

  return region_return;
}

static Region
x_getClipRegionFromObscuringWindows(YHW *hw, YWindow *window)
{
  Region region = XCreateRegion();
  XRectangle rect;
  YWindow_id shared_window_id = window->shared_window;
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(shared_window_id);
  int id = YSHM_lockSemaphore(info->sem_id);

  shared_window_id = info->prev_sibling;

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
  
  while (shared_window_id != 0)
	{
	  Region composite_region = XCreateRegion();
	  info = YSHM_getSharedWindowInfo(shared_window_id);
	  id = YSHM_lockSemaphore(info->sem_id);
	  
	  shared_window_id = info->prev_sibling;

	  rect.x = info->bounds.left;
	  rect.y = info->bounds.top;
	  rect.width = info->bounds.right - info->bounds.left;
	  rect.height = info->bounds.bottom - info->bounds.top;

	  if (rect.x > xhw->shared_data->display_rect.left)
		rect.x -= xhw->shared_data->display_rect.left;
	  else
		rect.x = 0;

	  if (rect.y > xhw->shared_data->display_rect.top)
		rect.y -= xhw->shared_data->display_rect.top;
	  else
		rect.y = 0;

	  YSHM_unlockSemaphore(info->sem_id, id);
	  YSHM_releaseSharedWindowInfo(info);
	  
	  XUnionRectWithRegion(&rect, region, composite_region);

	  XDestroyRegion(region);

	  region = composite_region;
	}

  return region;
}

static void
x_drawRectangle(YHW *hw, YGC* gc,
				int32 x, int32 y, uint32 width, uint32 height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw, gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);

  XSubtractRegion(win_bounds, obscuring_bounds, real_region);

  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XSetForeground(xhw->x_display, xhw->x_gc,
				 ycolor_to_pixel(hw, gc->foreground));
  XDrawRectangle(xhw->x_display, xhw->x_window, xhw->x_gc,
				 x, y,
				 width, height);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

static void
x_fillRectangle(YHW *hw,
				YGC* gc,
				int32 x, int32 y, uint32 width, uint32 height)
{
  Region win_bounds = x_getClipRegionFromWindowBounds(hw, gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);

  XSubtractRegion(win_bounds, obscuring_bounds, real_region);

  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XSetForeground(xhw->x_display, xhw->x_gc,
				 ycolor_to_pixel(hw, gc->foreground));
  XFillRectangle(xhw->x_display, xhw->x_window, xhw->x_gc,
				 x, y, width, height);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

static void
x_drawPoints(YHW *hw, YGC *gc,
			 YPoint **points,
			 int npoints)
{
  XPoint *xpoints;
  int i;

  Region win_bounds = x_getClipRegionFromWindowBounds(hw, gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;


  xpoints = (XPoint *) malloc (sizeof (XPoint) * npoints);
  for (i=0; i<npoints; i++)
	{
	  xpoints[ i ].x = points[ i ]->x;
	  xpoints[ i ].y = points[ i ]->y;
	}

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);

  XSubtractRegion(win_bounds, obscuring_bounds, real_region);

  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XSetForeground(xhw->x_display, xhw->x_gc,
				 ycolor_to_pixel(hw, gc->foreground));

  XDrawPoints (xhw->x_display, xhw->x_window, xhw->x_gc,
			   xpoints, npoints, CoordModeOrigin);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);

  free (xpoints);
}

static void
x_drawArc(YHW *hw, YGC *gc,
		  int32 center_x, int32 center_y, uint32 width, uint32 height)
{
}

static void
x_drawLine(YHW *hw, YGC* gc,
		   int32 x1, int32 y1,
		   int32 x2, int32 y2)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw, gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);

  XSubtractRegion(win_bounds, obscuring_bounds, real_region);

  XSetRegion(xhw->x_display, xhw->x_gc, real_region);
  
  XSetForeground(xhw->x_display, xhw->x_gc,
				 ycolor_to_pixel(hw, gc->foreground));
  XDrawLine(xhw->x_display, xhw->x_window, xhw->x_gc,
			x1, y1, x2, y2);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

static void
x_drawSpan(YHW *hw, YGC *gc,
		   int32 x1, int32 x2, int32 y)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw, gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);

  XSubtractRegion(win_bounds, obscuring_bounds, real_region);

  XSetRegion(xhw->x_display, xhw->x_gc, real_region);
  
  XSetForeground(xhw->x_display, xhw->x_gc,
				 ycolor_to_pixel(hw, gc->foreground));

  XDrawLine(xhw->x_display, xhw->x_window, xhw->x_gc,
			x1, y, x2, y);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

static void
x_setcursor(YHW *hw,
			YImage *cursor_image,
			int32 hot_x, int32 hot_y)
{
}

static void
x_movecursor(YHW *hw,
			 int32 x,
			 int32 y)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;

  XWarpPointer(xhw->x_display, None, xhw->x_window, 0, 0, 0, 0, x, y);
}

static void
x_showcursor(YHW *hw)
{
  /*
   * show a stupid cursor.  should use XChangeWindowAttributes?
   * some of this code should be in x_setcursor.
   */
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Cursor cur = XCreateFontCursor(xhw->x_display, XC_arrow);

  XDefineCursor(xhw->x_display, xhw->x_window, cur);
}

static void
x_hidecursor(YHW *hw)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  XUndefineCursor(xhw->x_display, xhw->x_window);
}

/* the next two methods, x_copyFromWin and x_copyToWin, are pretty icky.  They need to convert 
   the 32 bit Y data into data that the X server can deal with.  This pretty much means doing 
   completely fucked up things and special casing the kitchen sink. hopefully, it'll still end
   up being kinda fast when I'm done. */
static void
x_copyFromWin(YHW *hw, YGC *gc,
			  int32 src_x, int32 src_y,
			  int32 width, int32 height,
			  YColor **buffer)
{
}

static void
x_copyToWin(YHW *hw, YGC *gc,
			int32 dest_x, int32 dest_y,
			int32 width, int32 height,
			YColor *buffer)
{
#if 0
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw,gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  int x, y;
  int index = 0;

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);
  
  XSubtractRegion(win_bounds, obscuring_bounds, real_region);
  
  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  for (y = 0; y < height; y ++)
	for (x = 0; x < width; x ++)
	  {
		YColor color = buffer[index++];
		XPoint point;

		point.x = dest_x + x;
		point.y = dest_y + y;

		XSetForeground(xhw->x_display, xhw->x_gc,
					   ycolor_to_pixel(hw, color));

		XDrawPoint (xhw->x_display, xhw->x_window, xhw->x_gc,
					point.x, point.y);
	  }

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
#else
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw,gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  XImage *ximage;
  char *image_buffer;
  char *image_buf_pos;
  YColor *buf_pos;
  int depth = XDefaultDepthOfScreen(DefaultScreenOfDisplay(xhw->x_display));
  int i;

  image_buffer = (char*)malloc(width * height * depth);

  buf_pos = buffer;
  image_buf_pos = image_buffer;

  for (i = 0; i < width * height; i ++)
	{
	  uint8 red, green, blue;

	  YColorGetRGB(*buf_pos, &red, &green, &blue);

	  switch (depth)
		{
		case 16:
		  ;
		case 24:
		  /* assumes:
			 redmask 0xff0000
			 greenmask 0x00ff00
			 bluemask 0x0000ff
		  */
		  *(image_buf_pos ++) = red;
		  *(image_buf_pos ++) = green;
		  *(image_buf_pos ++) = blue;
		  break;
		default:
		  abort();
		}

	  buf_pos++;
	}

  ximage = XCreateImage(xhw->x_display,
						DefaultVisualOfScreen(DefaultScreenOfDisplay(xhw->x_display)),
						depth,
						ZPixmap, 0,
						(void*)image_buffer,
						width, height,
						24, 0);

  if (ximage == NULL)
	{
	  printf ("problem with toshok's x_copyToWin\n");
	  return;
	}

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);
  
  XSubtractRegion(win_bounds, obscuring_bounds, real_region);
  
  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XPutImage(xhw->x_display,
			xhw->x_window,
			xhw->x_gc,
			ximage,
			0,0,
			dest_x, dest_y,
			width, height);
			
  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);

  ximage->data = 0; /* we don't want them to free our image data. */

  XDestroyImage(ximage);
#endif
}

static void
x_copyFromWinToWin(YHW *hw, YGC *gc, YWindow *source,
				   int32 src_x, int32 src_y,
				   int32 dest_x, int32 dest_y,
				   int32 width, int32 height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw,gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  
  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);
  
  XSubtractRegion(win_bounds, obscuring_bounds, real_region);
  
  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XCopyArea(xhw->x_display,
			xhw->x_window, xhw->x_window,
			xhw->x_gc,
			src_x, src_y,
			width, height,
			dest_x, dest_y);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

#if 0

static void
x_OnOnBlit(YHW *hw, YGC *gc, YWindow *source, YWindow *dest,
		   int32 src_x, int32 src_y,
		   int32 dest_x, int32 dest_y,
		   uint32 width, uint32 height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw,gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  
  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);
  
  XSubtractRegion(win_bounds, obscuring_bounds, real_region);
  
  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XCopyArea(xhw->x_display,
			xhw->x_window, xhw->x_window,
			xhw->x_gc,
			src_x, src_y,
			width, height,
			dest_x, dest_y);

  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);
}

static void
x_OffOnBlit(YHW *hw, YGC *gc, YWindow *source, YWindow *dest,
			int32 src_x, int32 src_y,
			int32 dest_x, int32 dest_y,
			uint32 width, uint32 height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  Region win_bounds = x_getClipRegionFromWindowBounds(hw,gc->window);
  Region obscuring_bounds = x_getClipRegionFromObscuringWindows(hw,gc->window);
  Region real_region = XCreateRegion();
  XImage *ximage;
  int x, y;

  ximage = XCreateImage(xhw->x_display,
						DefaultVisualOfScreen(DefaultScreenOfDisplay(xhw->x_display)),
						16,
						ZPixmap, 0,
						source->image_data,						
						width, height, 32, width * 4);

  ximage->red_mask = 0xff000000;
  ximage->green_mask = 0x00ff0000;
  ximage->blue_mask = 0x0000ff00;

  XInitImage(ximage);

  if (!xhw->x_gc)
	xhw->x_gc = XCreateGC(xhw->x_display, xhw->x_window, 0, NULL);
  
  XSubtractRegion(win_bounds, obscuring_bounds, real_region);
  
  XSetRegion(xhw->x_display, xhw->x_gc, real_region);

  XPutImage(xhw->x_display,
			xhw->x_window,
			xhw->x_gc,
			ximage,
			0,0,
			dest_x, dest_y,
			width, height);
			
  XDestroyRegion(win_bounds);
  XDestroyRegion(obscuring_bounds);
  XDestroyRegion(real_region);

  ximage->data = 0; /* we don't want them to free our image data. */

  XDestroyImage(ximage);
}

static void
x_OnOffBlit(YHW *hw, YGC *gc, YWindow *source, YWindow *dest,
			int32 src_x, int32 src_y,
			int32 dest_x, int32 dest_y,
			uint32 width, uint32 height)
{
  x_hw_data *xhw = (x_hw_data*)hw->hw_data;
  XImage *ximage = XGetImage(xhw->x_display,
							 xhw->x_window,
							 src_x, src_y,
							 width, height,
							 AllPlanes, ZPixmap);
  int x, y;
  YGC *dest_gc = YGetGC(dest);
  YPoint *point = (YPoint*)malloc(sizeof(YPoint));

  for (y = 0; y < height; y ++)
	for (x = 0; x < width; x ++)
	  {
		long pixel = XGetPixel(ximage,
							   x, y);
		YColor color = pixel_to_ycolor(hw, pixel);

		YSetForeground(dest_gc, color);

		point->x = x;
		point->y = y;

		YDrawPoints(dest_gc, &point, 1);
	  }

  XDestroyImage(ximage);
  YDestroyGC(dest_gc);
  free(point);
}

static void
x_OffOffBlit(YHW *hw, YGC *gc, YWindow *source, YWindow *dest,
			 int32 src_x, int32 src_y,
			 int32 dest_x, int32 dest_y,
			 uint32 width, uint32 height)
{
}


static void
x_blit(YHW *hw, YGC *gc, YWindow *source, YWindow *dest,
	   int32 src_x, int32 src_y,
	   int32 dest_x, int32 dest_y,
	   uint32 width, uint32 height)
{
  if (y_isWindowActuallyImage(source))
	{
	  if (y_isWindowActuallyImage(dest))
		{
		  x_OffOffBlit(hw, gc, source, dest, src_x, src_y,
					   dest_x, dest_y, width, height);
		}
	  else
		{
		  x_OffOnBlit(hw, gc, source, dest, src_x, src_y,
					  dest_x, dest_y, width, height);
		}
	}
  else if (y_isWindowActuallyImage(dest))
	{
	  x_OnOffBlit(hw, gc, source, dest, src_x, src_y,
				  dest_x, dest_y, width, height);
	}
  else
	x_OnOnBlit(hw, gc, source, dest, src_x, src_y,
			   dest_x, dest_y, width, height);
}
#endif
