/* -*- Mode: C; tab-width: 4 -*-
   yshm.c -- Shared memory operations.
   Created: Chris Toshok <toshok@hungry.com>, 12-Sep-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

/*#define SYSV_SHM 1*/
#define MMAP_SHM 1

#include "yshm.h"

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#if defined ( MMAP_SHM )
#include <sys/mman.h>
#include <sys/fcntl.h>
#elif defined ( SYSV_SHM )
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#endif
#include <errno.h>

#if defined ( MMAP_SHM )
/* Linux seem to miss this definition */
#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

#define NESTED_LOCKS

static uint8 *_locks = NULL;
static uint32 _alloced_locks = 0;
#endif /* MMAP_SHM */

typedef enum {
  LOCK_EXCLUSIVE,
  LOCK_SHARED,
  UNLOCK
} lock_operation;

int
lock_file(int fd, lock_operation op)
{
  int real_op;

#ifdef HAVE_FLOCK
  switch (op)
    {
    case LOCK_EXCLUSIVE:
	  real_op = LOCK_EX;
	  break;
    case LOCK_SHARED:
	  real_op = LOCK_SH;
	  break;
    case UNLOCK:
	  real_op = LOCK_UN;
	  break;
	default:
	  fprintf (stderr, "lock_file (flock) -- bad operation: %d\n", op);
	  abort ();
    }

  return flock(fd, real_op);
#else
  switch (op)
    {
    case LOCK_EXCLUSIVE:
	  real_op = F_LOCK;
	  break;
    case LOCK_SHARED:
	  real_op = F_LOCK;
	  break;
    case UNLOCK:
	  real_op = F_ULOCK;
	  break;
	default:
	  fprintf (stderr, "lock_file (lockf) -- bad operation: %d\n", op);
	  abort ();
    }

  return lockf(fd, real_op, 0);
#endif
}

#if defined ( MMAP_SHM )
#ifdef DEBUG
void
print_locks()
{
  int i;

  printf ("Monitor\t\tDepth\n");
  printf ("=======\t\t=====\n");

  for (i = 0; i < _alloced_locks; i ++)
	if (_locks[i] > 0)
	  {
		printf ("%6d\t\t%5d\n", i, _locks[i]);
	  }
}
#endif

void
ENTER_MONITOR(int sem_id, int fd)
{
#ifdef NESTED_LOCKS
  if (!_locks)
	{
	  _alloced_locks = 100;
	  _locks = (uint8*)calloc(_alloced_locks, sizeof(uint8));
	  if (!_locks)
		{
		  fprintf(stderr, "couldn't allocate space for locks.\n");
		  exit(1);
		}
#ifdef DEBUG
	  signal(SIGUSR1, print_locks);
#endif
	}
  else if (_alloced_locks < sem_id)
	{
	  uint32 old_alloced_locks = _alloced_locks;

	  while (_alloced_locks < sem_id)
		_alloced_locks *= 2;

	  _locks = (uint8*)realloc(_locks, sizeof(uint8) * _alloced_locks);

	  memset(&_locks[old_alloced_locks], 0,
			 sizeof(uint8) * _alloced_locks - old_alloced_locks);
	}
  

  /* now we actually do the locking.  We check our array, and if we've
	 previously locked this semaphore, just bump up the count.  Otherwise,
	 someone else has (possibly) locked it, so we flock blocking, and set
	 our count to 1 after we gain the lock. */
  if (_locks[sem_id] > 0)
	{
	  _locks[sem_id] ++;
	  return;
	}
  else
	{
	  lock_file(fd, LOCK_EXCLUSIVE);
	  _locks[sem_id] = 1;
	}

#else /* ! NESTED_LOCKS */
  lock_file(fd, LOCK_EXCLUSIVE);
#endif
}

void
LEAVE_MONITOR(int sem_id, int fd)
{
#ifdef NESTED_LOCKS

  _locks[sem_id] --;
  if (_locks[sem_id] == 0)
	lock_file(fd, UNLOCK);

#else
  lock_file(fd, UNLOCK);
#endif
}

#endif /* MMAP_SHM */

static void
getNewSem(unsigned int *sem)
{
  int fd;
  struct stat stat_buf;
  char *buf;
  unsigned int shm;

  fd = open(Y_SHM_CONTEXT_FILE, O_RDWR | O_CREAT, 0777);

  if (fd < 0)
	{
	  perror ("getNewShmAndSem: open(2)");
	  shm = -1;
	  *sem = -1;
	  return;
	}
 
  lock_file(fd, LOCK_EXCLUSIVE); 

  fstat(fd, &stat_buf);
  buf = (char*)malloc(stat_buf.st_size);

  /* read in the current values from the file. */
  read(fd, buf, stat_buf.st_size);
  if (sscanf(buf, "%d %d", &shm, sem) != 2)
	{
	  *sem = 0;
	}

  /* rewind to the beginning. */
  lseek(fd, 0, SEEK_SET);

  /* increment them and rewrite the buffer. */
  sprintf(buf, "%d %d", shm, *sem + 1);
  write(fd, buf, strlen(buf));

  lock_file(fd, UNLOCK);

  close(fd);
}

static void
getNewShmAndSem(unsigned int *shm, unsigned int *sem)
{
  int fd;
  struct stat stat_buf;
  char *buf;

  fd = open(Y_SHM_CONTEXT_FILE, O_RDWR | O_CREAT, 0777);

  if (fd < 0)
	{
	  perror ("getNewShmAndSem: open(2)");
	  *shm = -1;
	  *sem = -1;
	  return;
	}
 
  lock_file(fd, LOCK_EXCLUSIVE); 

  fstat(fd, &stat_buf);
  buf = (char*)malloc(stat_buf.st_size);

  /* read in the current values from the file. */
  read(fd, buf, stat_buf.st_size);
  if (sscanf(buf, "%d %d", shm, sem) != 2)
	{
	  *shm = 0;
	  *sem = 0;
	}

  /* rewind to the beginning. */
  lseek(fd, 0, SEEK_SET);

  /* increment them and rewrite the buffer. */
  sprintf(buf, "%d %d", *shm + 1, *sem + 1);
  write(fd, buf, strlen(buf));

  lock_file(fd, UNLOCK);

  close(fd);
}

static char *
build_shm_filename(char *name)
{
  char buf[100];

  sprintf (buf, Y_SERVER_DIR "shm/%s", name);

  return strdup(buf);
}

static char *
build_sem_filename(char *name)
{
  char buf[100];
  
  sprintf (buf, Y_SERVER_DIR "sem/%s", name);

  return strdup(buf);
}

static char *
build_sem_filename_from_num(unsigned int id)
{
  char buf[100];

  sprintf (buf, Y_SERVER_DIR "sem/%d", id);

  return strdup(buf);
}

void
ensure_dir_exists(char *dir)
{
  char buf[100];

  strcpy (buf, Y_SERVER_DIR);

  if (mkdir(buf, 0777) == -1)
	{
	  if (errno != EEXIST)
		{
		  perror("ensure_dir_exists(mkdir1)");
		  exit(1);
		}
	}

  strcat (buf, dir);

  if (mkdir(buf, 0777) == -1)
	{
	  if (errno != EEXIST)
		{
		  perror("ensure_dir_exists(mkdir2)");
		  exit(1);
		}
	}
}

YWindow_id
YSHM_allocateSharedWindowInfo()
{
  YWindow_id id;
  YSharedWindowInfo *info;
  char window_filename[100];
  char *sem_filename;
  unsigned int sem_id;
  unsigned int sem_fd;

  getNewShmAndSem(&id, &sem_id);

  sprintf(window_filename, "win-%d", id);

  ensure_dir_exists("shm");
  info = YSHM_allocateSharedMemory(window_filename, sizeof(YSharedWindowInfo));

  /* fill in the semaphore field of the shared into */
  info->sem_id = sem_id;

  ensure_dir_exists("sem");
  /* create the semaphore file -- doesn't have to have anything in it. */
  sem_filename = build_sem_filename_from_num(sem_id);
  sem_fd = creat(sem_filename, 0777);
  free(sem_filename);
  close(sem_fd);

  return id;
}

YSharedWindowInfo *
YSHM_getSharedWindowInfo(YWindow_id id)
{
  char window_filename[100];
  YSharedWindowInfo *addr;

  sprintf(window_filename, "win-%d", id);
  addr = YSHM_getSharedMemory(window_filename, sizeof(YSharedWindowInfo));

  return addr;
}

void
YSHM_releaseSharedWindowInfo(YSharedWindowInfo *info)
{
  YSHM_releaseSharedMemory(info, sizeof(YSharedWindowInfo));
}

void
YSHM_destroySharedWindowInfo(YWindow_id win_id)
{
  char win_filename[100];
  char *sem_filename;
  unsigned int sem_id;
  YSharedWindowInfo *info;

  info = YSHM_getSharedWindowInfo(win_id);
  if (info == NULL) return; /* punt */

  sprintf(win_filename, "win-%d", win_id);
  sem_id = info->sem_id;
  sem_filename = build_sem_filename_from_num(sem_id);

  YSHM_releaseSharedWindowInfo(info);

  YSHM_destroySharedMemory(win_filename);

  if (unlink(sem_filename) < 0)
	perror("unlink(semaphore)\n");

  free(win_filename);
  free(sem_filename);
}

unsigned int
YSHM_allocateSemaphore()
{
  char *sem_filename;
  unsigned int sem_id;
  unsigned int sem_fd;
  
  getNewSem(&sem_id);

  ensure_dir_exists("sem");
  /* create the semaphore file -- doesn't have to have anything in it. */
  sem_filename = build_sem_filename_from_num(sem_id);
  sem_fd = creat(sem_filename, 0777);
  free(sem_filename);
  close(sem_fd);

  return sem_id;
}

static void
ysem_destroySemaphore(char *name)
{
  char *shm_name = build_shm_filename(name);
  if (unlink(shm_name) < 0)
	perror("unlink(shared_memory)\n");
  free(shm_name);
}

void
YSHM_destroySemaphoreById(unsigned int id)
{
  char sem_name[100];
  
  sprintf(sem_name, "sem-%d", id);
  
  ysem_destroySemaphore(sem_name);
}

int
YSHM_lockSemaphore(unsigned int semaphore_id)
{
  char *sem_filename;
  int sem_fd = 0;

  sem_filename = build_sem_filename_from_num(semaphore_id);

#if defined ( MMAP_SHM )
  sem_fd = open(sem_filename, O_RDWR);

  ENTER_MONITOR(semaphore_id, sem_fd);
#elif defined ( SYSV_SHM )
  {
	struct sembuf ops[2];
	key_t sem_key = ftok(sem_filename, 0);
	int sem_id = semget(sem_key, 1, IPC_CREAT | SEM_R | SEM_A);

	/* wait until the semaphore is 0 */
	ops[0].sem_num = sem_id; 
	ops[0].sem_op = 0;
	ops[0].sem_flg = 0;

	/* and then increment it by one. */
	ops[1].sem_num = 0;
	ops[1].sem_op = 1;
	ops[1].sem_flg = SEM_UNDO;

	semop(sem_id, ops, 2);
  }  
#endif

  free(sem_filename);


  return sem_fd;
}

void
YSHM_unlockSemaphore(unsigned int semaphore_id, unsigned int sem_fd)
{
#if defined ( MMAP_SHM )
  LEAVE_MONITOR(semaphore_id, sem_fd);
  close(sem_fd);
#elif defined ( SYSV_SHM )
  {
	char *sem_filename = build_sem_filename_from_num(semaphore_id);
	struct sembuf ops[1];
	key_t sem_key = ftok(sem_filename, 0);
	int sem_id = semget(sem_key, 1, IPC_CREAT | SEM_R | SEM_A);

	/* decrement the semaphore by 1. */
	ops[0].sem_num = sem_id; 
	ops[0].sem_op = -1;
	ops[0].sem_flg = 0;

	semop(sem_id, ops, 1);

	free (sem_filename);
  }
#endif
}

void*
YSHM_allocateSharedMemory(char *name, size_t size)
{
  int fd;
  char *shm_name = build_shm_filename(name);
  void *addr;

  /*  printf ("opening %s\n", shm_name);*/

  fd = open(shm_name, O_WRONLY | O_CREAT, 0777);

  if (fd < 0)
	{
	  perror("YSHM_allocateSharedMemory(open)");
	  fprintf (stderr, " file was '%s'\n", shm_name);
	  abort();
	}

#if defined ( MMAP_SHM )
  {
	char *buf;
	
	buf = calloc(1, size);
	write(fd, buf, size);
	
	free(buf);
	close(fd);
	
	fd = open(shm_name, O_RDWR /*| O_TRUNC*/);
	
	addr = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (addr == (void*)MAP_FAILED)
	  {
		perror("YSHM_allocateSharedMemory(mmap)");
		close(fd);
		free(shm_name);
		return NULL;
	  }
	
	close(fd);
  }
#elif defined ( SYSV_SHM )
  {
	key_t shm_key;
	int shm_id;
	
	close(fd);
	
	shm_key = ftok(shm_name, 0);
	
	shm_id = shmget(shm_key, size, IPC_CREAT | SHM_R | SHM_W);
	
	addr = shmat(shm_id, NULL, 0);
  }
#endif

  free(shm_name);

  return addr;
}

void*
YSHM_allocateSharedMemoryById(int id,
							  size_t size)
{
  char shm_name[100];

  sprintf(shm_name, "shm-%d", id);

  return YSHM_allocateSharedMemory(shm_name, size);
}

void
YSHM_releaseSharedMemory(void* addr,size_t size)
{
#if defined ( MMAP_SHM )
  int error = munmap((caddr_t)addr, size);
#elif defined ( SYSV_SHM )
  int error = shmdt(addr);
#endif

  if (error != 0)
	{
	  perror ("YSHM_releaseSharedMemory");
	  abort();
	}
}

void*
YSHM_getSharedMemory(char *name, size_t size)
{
  void *addr;
  char *shm_name = build_shm_filename(name);

#if defined ( MMAP_SHM )
  {
	int fd;
	fd = open(shm_name, O_RDWR);
	if (fd < 0)
	  {
		perror("YSHM_getSharedMemory - open(2)");
		return NULL;
	  }
	
	addr = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (addr == (void*)MAP_FAILED)
	  {
		perror("YSHM_getSharedMemory(mmap)");
		abort();
	  }
	
	/* do we want to do this here? */
	close(fd);
  }
#elif defined ( SYSV_SHM )
  {
	key_t shm_key;
	int shm_id;

	shm_key = ftok(shm_name, 0);
	
	shm_id = shmget(shm_key, size, SHM_R | SHM_W);
	
	addr = shmat(shm_id, NULL, 0);
  }
#endif

  free(shm_name);

  return addr;
}

void*
YSHM_getSharedMemoryById(int id,
						 size_t size)
{
  char shm_name[100];
  
  sprintf(shm_name, "shm-%d", id);
  
  return YSHM_getSharedMemory(shm_name, size);
}

void
YSHM_destroySharedMemory(char *name)
{
  char *shm_name = build_shm_filename(name);
  if (unlink(shm_name) < 0)
	perror("unlink(shared_memory)\n");
  free(shm_name);
}

void
YSHM_destroySharedMemoryById(int id)
{
  char shm_name[100];
  
  sprintf(shm_name, "shm-%d", id);
  
  YSHM_destroySharedMemory(shm_name);
}

