/* -*- Mode: C; tab-width: 4 -*-
   ydsloc.c -- Y window system local display connections.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ylib_p.h"
#include "yshm.h"
#include "hw.h"
#include "ymessages.h"

#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

typedef struct LocalDisplayData {
  int sock;
} LocalDisplayData;

extern YWindow_vtable yLocalWindow_vtable;

#define RECV_MSG(sock, buf) do { \
    int errval = recv(sock, buf, sizeof(Y_Reply), 0); \
    if (errval < 0) {\
		perror("RECV_MSG"); \
	    exit(0);\
    }} while (0);


/* send a message to the microserver, with all that error handling jazz... */
int
SEND_MSG(int sock, void *buf)
{

  D(("buffer size is %d\n", sizeof(Y_Request)));

  return send(sock, buf, sizeof(Y_Request),0);
}

static YWindow*
LocalCreateWindow(YDisplay *display,
				  YWindow *parent_window,
				  int x, int y, int width, int height,
				  YColor bg_color)
{
  YWindow *new_window;
  YSharedWindowInfo *info;
  YSharedWindowInfo *parent_info;
  int parent_sem, sem;

  /*  D(("Inside LocalCreateWindow()\n"));*/

  new_window = (YWindow*)calloc(1, sizeof(YWindow));

  new_window->display = display;
  new_window->vtable = yLocalWindow_vtable;

  new_window->shared_window = YSHM_allocateSharedWindowInfo();
  info = YSHM_getSharedWindowInfo(new_window->shared_window);

  sem = YSHM_lockSemaphore(info->sem_id);

  memset(info, 0, sizeof(YSharedWindowInfo));

  info->parent_id = parent_window->shared_window;

  info->ref_count = 1;

  info->bounds.left = x;
  info->bounds.top = y;
  info->bounds.right = width + x;
  info->bounds.bottom = height + y;

  /* Now we have to lock the parent's semaphore and add the
	 new window to its list of children. */
  parent_info = YSHM_getSharedWindowInfo(parent_window->shared_window);
  parent_sem = YSHM_lockSemaphore(parent_info->sem_id);

  info->prev_sibling = 0;
  info->next_sibling = parent_info->top_child;
 
  if(info->next_sibling != 0)
	{
	  YSharedWindowInfo *sib_info;
	  int sib_sem;

	  sib_info = YSHM_getSharedWindowInfo(info->next_sibling);
	  sib_sem = YSHM_lockSemaphore(sib_info->sem_id);

	  sib_info->prev_sibling = new_window->shared_window;

	  YSHM_unlockSemaphore(sib_info->sem_id, sib_sem);
	  YSHM_releaseSharedWindowInfo(sib_info);
	}

  parent_info->top_child = new_window->shared_window;

  if (parent_info->bottom_child == 0)
	parent_info->bottom_child = new_window->shared_window;

  YSHM_unlockSemaphore(parent_info->sem_id, parent_sem);
  YSHM_releaseSharedWindowInfo(parent_info);

  YSHM_unlockSemaphore(info->sem_id, sem);
  YSHM_releaseSharedWindowInfo(info);

  return new_window;
}

static void
LocalDestroyWindow(YDisplay *display, YWindow *window)
{
  YSharedWindowInfo *info;
  int id;

  D(("Inside LocalDestroyWindow()\n"));

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  info->ref_count --;

  if (info->ref_count == 0)
	{
	  YWindow_id prev_sibling, next_sibling;

	  YSHM_unlockSemaphore(info->sem_id, id);

	  prev_sibling = info->prev_sibling;
	  next_sibling = info->next_sibling;
	  
	  YSHM_releaseSharedWindowInfo(info);
	  
	  if (prev_sibling != 0)
		{
		  info = YSHM_getSharedWindowInfo(prev_sibling);
		  
		  info->next_sibling = next_sibling;
		  
		  YSHM_releaseSharedWindowInfo(info);
		}
	  
	  if (next_sibling != 0)
		{
		  info = YSHM_getSharedWindowInfo(next_sibling);
		  
		  info->prev_sibling = prev_sibling;
		  
		  YSHM_releaseSharedWindowInfo(info);
		}
	  
	  YSHM_destroySharedWindowInfo(window->shared_window);
	  free(window);
	}
  else
	YSHM_unlockSemaphore(info->sem_id, id);
}

static void
LocalOpenConnection(YDisplay *display)
{
  LocalDisplayData *data;
  int servlen, clientlen;
  struct sockaddr_un serv_addr, client_addr;

  ensure_dir_exists ("c");

  D(("Inside LocalOpenConnection()\n"));

  data = (LocalDisplayData*)calloc(1, sizeof(LocalDisplayData));

  data->sock = socket(AF_UNIX, SOCK_DGRAM, 0);

  if (data->sock < 0)
	{
	  perror("LocalOpenConnection(socket)");
	  abort();
	}

  memset((char*)&client_addr, 0, sizeof(client_addr));
  client_addr.sun_family = AF_UNIX;
  sprintf(client_addr.sun_path, Y_CLIENT_SOCK_PATH, getpid());
  clientlen = sizeof(struct sockaddr_un);

  unlink(client_addr.sun_path);
  if (bind(data->sock, (struct sockaddr*)&client_addr, clientlen) < 0)
	{
	  perror("LocalOpenConnection(bind)");
	  abort();
	}

  memset((char*)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sun_family = AF_UNIX;
  strcpy(serv_addr.sun_path, Y_SERVER_SOCK_PATH);
  servlen = sizeof(struct sockaddr_un);

  if (connect(data->sock, (struct sockaddr*)&serv_addr, servlen) < 0)
	{
	  perror("LocalOpenConnection(connect)");
	  abort();
	}

  display->connection_data = data;
  
  /* get the root window */
  {
    Y_Request root_req;
    Y_Reply root_rep;
	YWindow *root;

    root_req.tag = Y_GetRootWindowReq;

    root = (YWindow*)calloc(1, sizeof(YWindow));
    root->display = display;
    root->vtable = yLocalWindow_vtable;

    SEND_MSG(data->sock, &root_req);

    /* now get the reply */
    RECV_MSG(data->sock, &root_rep);

	printf ("Received the message\n");

    root->shared_window = root_rep.getroot_rep.root_id;

    display->root = root;
  }

  /* get the current hardware driver */
  {
    Y_Request driver_req;
    Y_Reply driver_rep;
	int cur_hw_shm_id;

    driver_req.tag = Y_GetDriverReq;

    SEND_MSG(data->sock, &driver_req);

	/* now get the reply */
	RECV_MSG(data->sock, &driver_rep);
	
	for (cur_hw_shm_id = 0;
		 cur_hw_shm_id < driver_rep.getdriver_rep.num_ids;
		 cur_hw_shm_id ++)
	  {
		YHW_InitHardware(driver_rep.getdriver_rep.first_id + cur_hw_shm_id);
	  }
  }

  /* and open the display connection */
  YHW_OpenDisplay();
}

static void
LocalCloseConnection(YDisplay *display)
{
  LocalDisplayData *data = display->connection_data;

  /* let the hardware layer know we're closing the display */
  YHW_CloseDisplay();

  /* do something else, probably... */

  close(data->sock);
}

static int
LocalGetConnectionFD(YDisplay *display)
{
  LocalDisplayData *data = display->connection_data;

  return data->sock;
}

static YWindow *
LocalGetWindowById(YDisplay *display, YWindow_id id)
{
  YWindow *new_window = (YWindow*)calloc(1, sizeof(YWindow));
  int sem_id;
  YSharedWindowInfo *info;

  new_window->shared_window = id;
  new_window->display = display;
  new_window->vtable = yLocalWindow_vtable;

  info = YSHM_getSharedWindowInfo(id);
  sem_id = YSHM_lockSemaphore(info->sem_id);

  info->ref_count ++;

  YSHM_unlockSemaphore(info->sem_id, sem_id);
  YSHM_releaseSharedWindowInfo(info);

  return new_window;
}

static int
LocalListFonts(YDisplay *display, const char **font_specs)
{
  /* help? */
  return 0;
}

static void
LocalGetNextEvent(YDisplay *display, Y_Event *event)
{
  LocalDisplayData *data = display->connection_data;

  printf ("in LocalGetNextEvent...\n");
  
  RECV_MSG(data->sock, event);

  printf ("leaving LocalGetNextEvent...\n");
}

void
y_sendEvent(YDisplay *display, Y_Event *event)
{
  LocalDisplayData *data = display->connection_data;

  SEND_MSG(data->sock, event);
}

YDisplay_vtable yLocalDisplay_vtable = {
  LocalOpenConnection,
  LocalCloseConnection,
  LocalGetConnectionFD,
  LocalCreateWindow,
  LocalDestroyWindow,
  LocalGetWindowById,
  LocalListFonts,
  LocalGetNextEvent
};
