/* -*- Mode: C; tab-width: 4 -*-
   ydsnet.c -- Y window system net display connections.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ylib_p.h"
#include "ynet_p.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

typedef struct NetDisplayData {
  int sock;
} NetDisplayData;

extern YWindow_vtable yNetWindow_vtable;

static int
net_RECV_MSG(int sock, void *buf, size_t size)
{
  int errval = read(sock, buf, size);
  if (errval < 0) {
	perror("RECV_MSG");
	abort();
  }

  return errval;
}

/* send a message to the microserver, with all that error handling jazz... */
static int
net_SEND_MSG(int sock, void *buf, size_t size)
{

  /*  D(("buffer size is %d\n", size));*/

  return write(sock, buf, size);
}

int
NET_RECV_MSG(YDisplay *display, void *buf)
{
  NetDisplayData *data = (NetDisplayData*)display->connection_data;

  return net_RECV_MSG(data->sock, buf, sizeof(YNetRequest/*XXX*/));
}

int
NET_RECV_DATA(YDisplay *display, void *buf, uint32 length)
{
  NetDisplayData *data = (NetDisplayData*)display->connection_data;

  return net_RECV_MSG(data->sock, buf, length);
}

int
NET_SEND_MSG(YDisplay *display, YNetRequest *req)
{
  NetDisplayData *data = (NetDisplayData*)display->connection_data;

  return net_SEND_MSG(data->sock, req, sizeof(YNetRequest));
}

int
NET_SEND_DATA(YDisplay *display, void *buffer, uint32 length)
{
  NetDisplayData *data = (NetDisplayData*)display->connection_data;

  return net_SEND_MSG(data->sock, buffer, length);
}

static YWindow*
NetCreateWindow(YDisplay *display,
				YWindow *parent_window,
				int x, int y, int width, int height,
				YColor bg_color)
{
  YWindow *new_window;
  YNetCreateWinReq create_win;
  YNetReply create_win_reply;

  D(("Inside NetCreateWindow()\n"));

  create_win.tag = CREATE_WIN;
  create_win.parent_id = htonl(parent_window->shared_window);
  create_win.x = htonl(x);
  create_win.y = htonl(y);
  create_win.width = htonl(width);
  create_win.height = htonl(height);
  create_win.bg_color = htonl(bg_color);

  NET_SEND_MSG(display, (YNetRequest*)&create_win);

  NET_RECV_MSG(display, &create_win_reply);

  assert(create_win_reply.tag == CREATE_WIN);

  new_window = (YWindow*)calloc(1, sizeof(YWindow));

  new_window->shared_window = ntohl(create_win_reply.create_win.win_id);

  new_window->display = display;
  new_window->vtable = yNetWindow_vtable;
  
  return new_window;
}

static void
NetDestroyWindow(YDisplay *display, YWindow *window)
{
  D(("Inside NetDestroyWindow()\n"));
}

static void
NetOpenConnection(YDisplay *display)
{
  NetDisplayData *data;
  int servlen;
  struct sockaddr_in serv_addr;
  char *colon;
  struct hostent *hostent;

  D(("Inside NetOpenConnection()\n"));

  data = (NetDisplayData*)calloc(1, sizeof(NetDisplayData));

  data->sock = socket(AF_INET, SOCK_STREAM, 0);

  if (data->sock < 0)
	{
	  perror("NetOpenConnection(socket)");
	  abort();
	}

  memset((char*)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;

  colon = strchr(display->display_name, ':');
  if (colon)
	*colon = '\0';

  hostent = gethostbyname(display->display_name);
  if (!hostent)
	hostent = gethostbyaddr(display->display_name, strlen(display->display_name),
				AF_INET);

  if (!hostent)
	{
	  fprintf(stderr, "Unable to resolve hostname %s\n", display->display_name);
	  exit(1);
	}

  memcpy(&serv_addr.sin_addr.s_addr, hostent->h_addr, sizeof(long));

  if (colon)
	*colon = ':';

  serv_addr.sin_port = htons(6500);
  servlen = sizeof(struct sockaddr_in);

  if (connect(data->sock, (struct sockaddr*)&serv_addr, servlen) < 0)
	{
	  perror("NetOpenConnection(connect)");
	  abort();
	}

  display->connection_data = data;
  
  /* get the root window */
  {
	YNetGetRootWinReq root_req;
    Y_Reply root_rep;
	YWindow *root;

    root_req.tag = Y_GetRootWindowReq;

    root = (YWindow*)calloc(1, sizeof(YWindow));
    root->display = display;
    root->vtable = yNetWindow_vtable;

    NET_SEND_MSG(display, (YNetRequest*)&root_req);

    /* now get the reply */
    NET_RECV_MSG(display, &root_rep);

	printf ("Received the message\n");

    root->shared_window = ntohl(root_rep.getroot_rep.root_id);

    display->root = root;
  }
}

static void
NetCloseConnection(YDisplay *display)
{
  NetDisplayData *data = display->connection_data;

  /* do something else, probably... */

  close(data->sock);
}

static int
NetGetConnectionFD(YDisplay *display)
{
  NetDisplayData *data = display->connection_data;

  return data->sock;
}

static YWindow *
NetGetWindowById(YDisplay *display, YWindow_id id)
{
  YWindow *new_window = (YWindow*)calloc(1, sizeof(YWindow));

  /* do we need anything more than this here? */
  new_window->shared_window = id;
  new_window->display = display;
  new_window->vtable = yNetWindow_vtable;

  return new_window;
}

static int
NetListFonts(YDisplay *display, const char **font_specs)
{
  /* help? */
  return 0;
}

static void
NetGetNextEvent(YDisplay *display, Y_Event *event)
{
  /* uh... huh huh :) */
}

YDisplay_vtable yNetDisplay_vtable = {
  NetOpenConnection,
  NetCloseConnection,
  NetGetConnectionFD,
  NetCreateWindow,
  NetDestroyWindow,
  NetGetWindowById,
  NetListFonts,
  NetGetNextEvent
};
