/* -*- Mode: C; tab-width: 4 -*-
   ygcmem.c -- Y window system graphics contexts, rendering directly to RAM --
               to implement things like pixmaps.
   Created: Chris Toshok <toshok@hungry.com>, 21-Sep-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ylib_p.h"
#include "yfonts_p.h"
#include "yshm.h"
#include "hw.h"

#include <stdio.h>
#include <stdlib.h>

static void
MemCopyFromWin(YGC *gc,
			   YPoint *src_loc, YRect *size, YColor **buffer)
{
  YWindow *window = gc->window;
  YColor *buf = (YColor*)calloc(1,sizeof(YColor) * size->right * size->bottom);
  YColor *off_data;
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);

  if (src_loc->x == 0 && size->right == info->bounds.right)
	{
	  off_data = window->image_data;
	  off_data += src_loc->y * (info->bounds.right - info->bounds.left);

	  memcpy(buf, off_data, size->bottom * size->right * sizeof(YColor));
	}
  else
	{
	  int row;
	  int index = 0;

	  off_data = window->image_data;
	  off_data += src_loc->y * (info->bounds.right - info->bounds.left) + src_loc->x;

	  for (row = src_loc->y; row < size->bottom + src_loc->y; row ++)
		{
		  memcpy(&(buf[index]), off_data, size->right * sizeof(YColor));

		  off_data += (info->bounds.right - info->bounds.left);
		  index += size->right * sizeof(YColor);
		}
	}

  YSHM_releaseSharedWindowInfo(info);
  *buffer = buf;
}

static void
MemCopyToWin(YGC *gc,
			 YPoint *dest_loc, YRect *size, YColor *buffer)
{
}

static void
MemCopyFromWinToWin(YGC *gc,
					YWindow *source, YPoint *src_loc, YPoint *dest_loc,
					YRect *size)
{
  YWindow *dest = gc->window;
  YSharedWindowInfo *src_info =
	YSHM_getSharedWindowInfo(source->shared_window);
  YSharedWindowInfo *dest_info =
	YSHM_getSharedWindowInfo(dest->shared_window);
  int x, y;
  int src_initial_index = (src_loc->x
						   + (src_loc->y
							  *(src_info->bounds.right - src_info->bounds.left)));
  int dest_initial_index = (dest_loc->x
							+ (dest_loc->y
							   *(dest_info->bounds.right - dest_info->bounds.left)));

  for (y = 0; y < size->bottom; y ++)
	for (x = 0; x < size->right; x ++)
	  {
		((YColor*)dest->image_data)[dest_initial_index 
								   + x 
								   + y 
								   * (dest_info->bounds.right - dest_info->bounds.left)] 
		  =	((YColor*)source->image_data)[src_initial_index 
										 + x 
										 + y 
										 * (src_info->bounds.right - src_info->bounds.left)];
	  }

  YSHM_releaseSharedWindowInfo(dest_info);
  YSHM_releaseSharedWindowInfo(src_info);
}

static void
draw_midpoint_line(YGC *gc,
				   YPoint *p1,
				   YPoint *p2)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  int width = info->bounds.right - info->bounds.left;
  int x;
  double dy = p2->y - p1->y;
  double dx = p2->x - p1->x;
  double m = dy / dx;
  double y = p1->y;
  
  for (x = p1->x; x <= p2->x; x ++)
	{
	  ((YColor*)gc->window->image_data)[ (int)(y * width + x) ]
		= gc->foreground;
	  y += m;
	}
  

  YSHM_releaseSharedWindowInfo(info);
}

static void
draw_horizontal_line(YGC *gc,
					 YPoint *p1,
					 YPoint *p2)
{
  int x = p1->x;
  int y = p1->y;
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  int width = info->bounds.right - info->bounds.left;

  while (x < p2->x)
	{
	  ((YColor*)gc->window->image_data)[ y * width + x ] = gc->foreground;
	  x++;
	}
}

static void
draw_vertical_line(YGC *gc,
					 YPoint *p1,
					 YPoint *p2)
{
  int x = p1->x;
  int y = p1->y;
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  int width = info->bounds.right - info->bounds.left;

  while (y < p2->y)
	{
	  ((YColor*)gc->window->image_data)[ y * width + x ] = gc->foreground;
	  y++;
	}

  YSHM_releaseSharedWindowInfo(info);
}

static void
MemDrawLine(YGC *gc,
			YPoint *p1,
			YPoint *p2)
{
  if (p1->x == p2->x)
	{
	  if (p1->y > p2->y)
		{
		  YPoint tmp = *p1;
		  *p1 = *p2;
		  *p2 = tmp;
		}

	  draw_vertical_line(gc, p1, p2);
	}
  else if (p1->y == p2->y)
	{
	  if (p1->x > p2->x)
		{
		  YPoint tmp = *p1;
		  *p1 = *p2;
		  *p2 = tmp;
		}
	  draw_horizontal_line(gc, p1, p2);
	}
  else
	{
	  if (p1->x > p2->x)
		{
		  YPoint tmp = *p1;
		  *p1 = *p2;
		  *p2 = tmp;
		}

	  draw_midpoint_line(gc, p1, p2);
	}
	
}

static void
MemDrawAsciiText(YGC *gc,
				 YPoint *where,
				 char *ascii_text)
{
  YFont *font = gc->font;

  YFontRenderer_renderCharacter(font->renderer,
								ascii_text, strlen(ascii_text),
								gc, where);
}

static void
MemDrawPoints(YGC *gc,
			  YPoint **points, int npoints)
{
  YSharedWindowInfo *info =
	YSHM_getSharedWindowInfo(gc->window->shared_window);
  int width = info->bounds.right - info->bounds.left;
  int i;

  for (i = 0; i < npoints; i ++)
	((YColor*)gc->window->image_data)[ points[i]->y * width + points[i]->x ]
	  = gc->foreground;

  YSHM_releaseSharedWindowInfo(info);
}

static void
MemDrawRect(YGC *gc,
			YRect *rect)
{
}

static void
MemFillRect(YGC *gc,
			YRect *rect)
{
}

static void
MemSetForeground(YGC *gc,
				 YColor foreground)
{
  gc->foreground = foreground;
}

static void
MemSetBackground(YGC *gc,
				 YColor background)
{
  gc->background = background;
}

static void
MemSetFont(YGC *gc,
		   YFont *font)
{
  gc->font = font;
}

static void
MemDestroy(YGC *gc)
{
  /* free things. */
  free(gc);
}

YGC_vtable yMemGC_vtable = {
  MemCopyFromWin,
  MemCopyToWin,
  MemCopyFromWinToWin,
  MemDrawLine,
  MemDrawAsciiText,
  MemDrawPoints,
  MemDrawRect,
  MemFillRect,
  MemSetForeground,
  MemSetBackground,
  MemSetFont,
  MemDestroy
};
