/* -*- Mode: C; tab-width: 4 -*-
   ygcnet.c -- Y window system graphics contexts, rendering across the network.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ylib_p.h"
#include "ynet_p.h"

#include <stdio.h>
#include <stdlib.h>
#if defined(linux) || defined(hpux)
#include <netinet/in.h>
#endif

extern int NET_RECV_MSG(YDisplay *display, void *buf);
extern int NET_SEND_MSG(YDisplay *display, YNetRequest *req);
extern int NET_RECV_DATA(YDisplay *display, void *buf, uint32 length);
extern int NET_SEND_DATA(YDisplay *display, void *buf, uint32 length);

static void
NetCopyFromWin(YGC *gc,
			   YPoint *src_loc, YRect *size, YColor **buffer)
{
  YNetCopyFromWinReq req;
  YNetCopyFromWinRep rep;

  req.tag = COPY_FROM_WIN;
  req.gc_id = htonl(gc->gc_id);
  req.src_x = htonl(src_loc->x);
  req.src_y = htonl(src_loc->y);
  req.width = htonl(size->right);
  req.height = htonl(size->bottom);

  NET_SEND_MSG(gc->display,
			   (YNetRequest*)&req);

  NET_RECV_DATA(gc->display,
				&rep, sizeof(YNetCopyFromWinRep));
  (*buffer)[0] = rep.buffer[0];
  NET_RECV_DATA(gc->display,
				&buffer[1], rep.buffer_size);
}

void 
NetCopyToWin(YGC *gc,
			 YPoint *dest_loc, YRect *size, YColor *buffer)
{
  YNetCopyToWinReq *req;
  uint32 buffer_size;
  uint32 req_size;

  buffer_size = size->right * size->bottom * sizeof(YColor);
  req_size = sizeof(YNetCopyToWinReq) + buffer_size - sizeof(YColor);

  req = (YNetCopyToWinReq*)malloc(req_size);

  req->tag = COPY_TO_WIN;
  req->gc_id = htonl(gc->gc_id);
  req->dest_x = htonl(dest_loc->x);
  req->dest_y = htonl(dest_loc->y);
  req->width = htonl(size->right);
  req->height = htonl(size->bottom);
  req->buffer_size = buffer_size;

  memcpy(&req->buffer, buffer, buffer_size);

  NET_SEND_DATA(gc->display,
				req, req_size);

  free(req);
}

void 
NetCopyFromWinToWin(YGC *dest_gc,
					YWindow *source, YPoint *src_loc, YPoint *dest_loc,
					YRect *size)
{
  YNetCopyFromWinToWinReq req;

  req.tag = COPY_FROM_WIN_TO_WIN;
  req.gc_id = htonl(dest_gc->gc_id);
  req.src_id = htonl(source->shared_window);
  req.src_x = htonl(src_loc->x);
  req.src_y = htonl(src_loc->y);
  req.dest_x = htonl(src_loc->x);
  req.dest_x = htonl(src_loc->y);
  req.width = htonl(size->right);
  req.height = htonl(size->bottom);

  NET_SEND_MSG(dest_gc->display,
			   (YNetRequest*)&req);

  /* no reply */
}

static void
NetDrawLine(YGC *gc, YPoint *p1, YPoint *p2)
{
  YNetDrawLineReq req;

  req.tag = DRAW_LINE;
  req.gc_id = htonl(gc->gc_id);
  req.x1 = htonl(p1->x);
  req.y1 = htonl(p1->y);
  req.x2 = htonl(p2->x);
  req.y2 = htonl(p2->y);

  NET_SEND_MSG(gc->display,
			   (YNetRequest*)&req);

  /* no reply */
}

static void
NetDrawAsciiText(YGC *gc, YPoint *where, char *text)
{
  YNetDrawAsciiTextReq *req;
  int buffer_size;
  int req_size;

  buffer_size = strlen(text);
  req_size = sizeof(YNetDrawAsciiTextReq) + buffer_size - sizeof(char);

  req = (YNetDrawAsciiTextReq*)malloc(req_size);
  req->tag = DRAW_ASCII_TEXT;
  req->gc_id = gc->gc_id;
  req->x = htonl(where->x);
  req->y = htonl(where->y);
  strcpy(req->txt, text);

  NET_SEND_DATA(gc->display,
				req, req_size);

  free(req);
}

static void
NetDrawPoints(YGC *gc, YPoint **points, int npoints)
{
}

static void
NetDrawRect(YGC *gc, YRect *rect)
{
}

static void
NetFillRect(YGC *gc, YRect *rect)
{
}

static void
NetSetForeground(YGC *gc, YColor foreground)
{
  YNetSetForegroundReq req;

  req.tag = SET_FOREGROUND;
  req.gc_id = htonl(gc->gc_id);
  req.foreground = htonl(foreground);
  
  NET_SEND_MSG(gc->display,
			   (YNetRequest*)&req);

  /* no reply */
}

static void
NetSetBackground(YGC *gc, YColor background)
{
}

static void
NetSetFont(YGC *gc, YFont *font)
{
}

static void
NetDestroy(YGC *gc)
{
  free(gc);
}

YGC_vtable yNetGC_vtable = {
  NetCopyFromWin,
  NetCopyToWin,
  NetCopyFromWinToWin,
  NetDrawLine,
  NetDrawAsciiText,
  NetDrawPoints,
  NetDrawRect,
  NetFillRect,
  NetSetForeground,
  NetSetBackground,
  NetSetFont,
  NetDestroy
};
