/* -*- Mode: C; tab-width: 4 -*-
   ywinloc.c -- Local YWindow vtable.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ylib_p.h"
#include "yshm.h"
#include "ymessages.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

extern YGC_vtable yLocalGC_vtable;

static void LocalMapWindow(YWindow *window);
static void LocalUnmapWindow(YWindow *window);
static void LocalClearWindow(YWindow *window);
static void LocalResizeWindow(YWindow *window, int width, int height);
static void LocalMoveWindow(YWindow *window, int x, int y);
static void LocalRaiseWindow(YWindow *window);
static void LocalLowerWindow(YWindow *window);
static void LocalSetBackground(YWindow *window, YColor background);
static YGC *LocalGetGC(YWindow *window);
static void LocalAddInterest(YWindow *window);
static void LocalNotifyInterested(YWindow *window, Y_Event *event);
static YRect *LocalGetBounds(YWindow *window);
static int LocalGetWindowId(YWindow *window);
static int LocalGetSubwindows(YWindow *window, YWindow ***subwindows);
static void LocalReleaseSubwindows(YWindow *window, YWindow **subwindows, int length);

YWindow_vtable yLocalWindow_vtable = {
  LocalMapWindow,
  LocalUnmapWindow,
  LocalMoveWindow,
  LocalResizeWindow,
  LocalRaiseWindow,
  LocalLowerWindow,
  LocalClearWindow,
  LocalGetGC,
  LocalSetBackground,
  LocalAddInterest,
  LocalNotifyInterested,
  LocalGetBounds,
  LocalGetWindowId,
  LocalGetSubwindows,
  LocalReleaseSubwindows
};

static YBool
AnyWindowOverlapsMe(YWindow *window)
{
  YWindow_id id;
  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(window->shared_window);
  YRect bounds;

  bounds = info->bounds;
  id = info->prev_sibling;

  YSHM_releaseSharedWindowInfo(info);

  while (id != 0)
	{
	  YSharedWindowInfo *prev = YSHM_getSharedWindowInfo(id);
	  int sem_id = YSHM_lockSemaphore(prev->sem_id);
	  YWindow_id new_id;

	  if (prev->map_state == Y_MAPPED
		  && YDoRectsIntersect(&prev->bounds, &bounds))
		{

		  YSHM_unlockSemaphore(prev->sem_id, sem_id);
		  if (prev != info) YSHM_releaseSharedWindowInfo(prev);

		  return Y_TRUE;
		}
	  
	  new_id = prev->prev_sibling;

	  YSHM_unlockSemaphore(prev->sem_id, sem_id);
	  if (prev != info) YSHM_releaseSharedWindowInfo(prev);

	  id = new_id;
	}

  return Y_FALSE;
}

static YBool
IOverlapAnyWindow(YWindow *window)
{
  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(window->shared_window);
  YWindow_id id;
  YRect bounds;

  if (info->map_state == Y_UNMAPPED)
	{
	  YSHM_releaseSharedWindowInfo(info);
	  return Y_FALSE;
	}

  bounds = info->bounds;
  id = info->next_sibling;

  YSHM_releaseSharedWindowInfo(info);

  while (id != 0)
	{
	  YSharedWindowInfo *next = YSHM_getSharedWindowInfo(id);
	  int sem_id = YSHM_lockSemaphore(next->sem_id);
	  YWindow_id new_id;

	  if (next->map_state == Y_MAPPED
		  && YDoRectsIntersect(&next->bounds, &bounds))
		{

		  YSHM_unlockSemaphore(next->sem_id, sem_id);
		  if (next != info) YSHM_releaseSharedWindowInfo(next);

		  return Y_TRUE;
		}
	  
	  new_id = next->next_sibling;

	  YSHM_unlockSemaphore(next->sem_id, sem_id);
	  if (next != info) YSHM_releaseSharedWindowInfo(next);

	  id = new_id;
	}

  return Y_FALSE;
}

static void
LocalMapWindow(YWindow *window)
{
  YSharedWindowInfo *info;
  int id;
  Y_Message expose_event;

  /*  D(("Inside LocalMapWindow()\n"));*/

  info = YSHM_getSharedWindowInfo(window->shared_window);

  /* the root window is always mapped. */
  if (info->parent_id == 0)
	goto done;

  id = YSHM_lockSemaphore(info->sem_id);

  if (info->map_state == Y_MAPPED)
	goto done;

  info->map_state = Y_MAPPED;

  {
	YWindow_id child;

	/* paint the background for the client */
	YGC *tmp_gc = YGetGC(window);
	YRect bounds = info->bounds;

	YSetForeground(tmp_gc, info->background);

	bounds.right -= bounds.left;
	bounds.bottom -= bounds.top;
	bounds.left = bounds.top = 0;

	YFillRectangle(tmp_gc, &bounds);
	
	expose_event.expose.tag = Y_ExposeEvent;
	expose_event.expose.bounds = bounds;
	expose_event.expose.target = window->shared_window;
	expose_event.expose.timestamp = 0; /* XXX */

	y_sendEvent(window->display, &expose_event);

	YDestroyGC(tmp_gc);

	/* first we map all the children of this window */

	child = info->bottom_child;
	while (child != 0)
	  {
		YSharedWindowInfo *child_info = YSHM_getSharedWindowInfo(child);
		YWindow_id prev_child;
		YWindow *child_window;

		prev_child = child_info->prev_sibling;
		
		YSHM_releaseSharedWindowInfo(child_info);

		child_window = YGetWindowById(window->display,
									  child);

		YMapWindow(child_window);

		free(child_window);

		child = prev_child;
	  }
  }

  YSHM_unlockSemaphore(info->sem_id, id);

  /* after we've unlocked the semaphore, send an expose event to all that are
	 interested. */

 done:
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalUnmapWindow(YWindow *window)
{
  YSharedWindowInfo *info;
  int id;
  D(("Inside LocalUnmapWindow()\n"));

  info = YSHM_getSharedWindowInfo(window->shared_window);

  /* the root window can't be unmapped. */
  if (info->parent_id == 0)
	goto done;

  id = YSHM_lockSemaphore(info->sem_id);

  if (info->map_state != Y_UNMAPPED)
	info->map_state = Y_UNMAPPED;

  /* XXX do some more stuff here. */

  YSHM_unlockSemaphore(info->sem_id, id);

 done:
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalResizeWindow(YWindow *window, int width, int height)
{
  YSharedWindowInfo *info;
  YBool smaller = Y_FALSE; /* if the window is smaller after the resize. */
  int id;

  D(("Inside LocalResizeWindow()\n"));

  info = YSHM_getSharedWindowInfo(window->shared_window);

  /* the root window can't be resized. */
  if (info->parent_id == 0)
	goto done;

  id = YSHM_lockSemaphore(info->sem_id);

  if (width < info->bounds.right - info->bounds.left
	  || height < info->bounds.bottom - info->bounds.top)
	smaller = Y_TRUE;

  info->bounds.right = info->bounds.left + width;
  info->bounds.bottom = info->bounds.top + height;

  /* figure out things about windows newly unobscured, if we need to */
  if (smaller) /* XXX */;

#if 0 /* this needs to be reworked. */
  for (i = 0; i < info->num_interested; i ++)
	{
	  Y_Event event;

	  event.any_event.dest = info->interested_parties[i];
	  event.any_event.tag = Y_ExposeEvent;
	  event.expose.bounds = info->bounds;
	  event.expose.timestamp = 0;

	  if (info->interested_parties[i] == getpid())
		{
		  Y_handleEvent(&event);
		}
	  else
		{
		  /*		  Y_sendEvent(&event);*/
		}
	}
#endif

  YSHM_unlockSemaphore(info->sem_id, id);

 done:
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalMoveWindow(YWindow *window, int x, int y)
{
  int deltax, deltay;
  YSharedWindowInfo *info;
  int id;
  D(("Inside LocalMoveWindow()\n"));

  info = YSHM_getSharedWindowInfo(window->shared_window);

  /* the root window can't be moved. */
  if (info->parent_id == 0)
	goto done;

  id = YSHM_lockSemaphore(info->sem_id);

  deltax = x - info->bounds.left;
  deltay = y - info->bounds.top;

  info->bounds.left += deltax;
  info->bounds.right += deltax;
  info->bounds.top += deltay;
  info->bounds.bottom += deltay;

  /* XXX do some more stuff here. */

  YSHM_unlockSemaphore(info->sem_id, id);

 done:
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalRaiseWindow(YWindow *window)
{
}

static void
LocalLowerWindow(YWindow *window)
{
}

static void
LocalClearWindow(YWindow *window)
{
  YGC *tmp_gc;
  YSharedWindowInfo *info;
  int id;
  /*  D(("Inside LocalClearWindow()\n"));*/

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  tmp_gc = YGetGC(window);

  YSetForeground(tmp_gc, info->background);

  YFillRectangle(tmp_gc, &info->bounds);

  YDestroyGC(tmp_gc);
  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalSetBackground(YWindow *window, YColor background)
{
  YSharedWindowInfo *info;
  int id;
  /*D(("Inside LocalSetBackground()\n"));*/

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  info->background = background;

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
}

static YGC *
LocalGetGC(YWindow *window)
{
  YGC *new_gc;
 
  /*  D(("Inside LocalGetGC()\n"));*/

  new_gc = (YGC*)calloc(1, sizeof(YGC));

  new_gc->vtable = yLocalGC_vtable;

  new_gc->window = window;

  return new_gc;
}

static void
LocalAddInterest(YWindow *window)
{
  YSharedWindowInfo *info;
  int i;
  int id;
  int pid = getpid ();
  D(("Inside LocalAddInterest()\n"));

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  for (i = 0; i < info->num_interested; i ++)
	if (pid == info->interested_parties[i])
	  goto done;

  info->interested_parties[info->num_interested++] = pid;

 done:

  printf ("LocalAddInterest... window ID = %d, info->num_interested = %d\n",
		  window->shared_window, info->num_interested);

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
}

static void
LocalNotifyInterested(YWindow *window, Y_Event *event)
{
  YSharedWindowInfo *info;
  int id;

  printf ("in LocalNotifyInterested...\n");

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  if (info->num_interested)
	y_sendEvent(window->display, event);

  YSHM_unlockSemaphore (info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);

  printf ("leaving LocalNotifyInterested.\n");
}

static YRect *
LocalGetBounds(YWindow *window)
{
  YSharedWindowInfo *info;
  YRect *rect = (YRect*)calloc(1, sizeof(YRect));
  int id;

  /*  D(("Inside LocalGetBounds()\n"));*/

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  *rect = info->bounds;

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
  
  return rect;
}

static int
LocalGetWindowId(YWindow *window)
{
  return window->shared_window;
}

static int
LocalGetSubwindows(YWindow *window, YWindow ***subwindows)
{
  /* XXXX needs reworking. */
#if 1
  return 0;
#else
  YSharedWindowInfo *info;
  int num_subwindows;
  int i;
  int id;

  info = YSHM_getSharedWindowInfo(window->shared_window);
  id = YSHM_lockSemaphore(info->sem_id);

  num_subwindows = info->num_children;

  *subwindows = (YWindow**)calloc(num_subwindows, sizeof(YWindow*));

  for (i = 0; i < num_subwindows; i ++)
	{
	  (*subwindows)[i] = (YWindow*)calloc(1, sizeof(YWindow));

	  (*subwindows)[i]->display = window->display;
	  (*subwindows)[i]->shared_window = info->children[i];
	  (*subwindows)[i]->vtable = yLocalWindow_vtable;
	}

  YSHM_unlockSemaphore(id);
  YSHM_releaseSharedWindowInfo(info);

  return num_subwindows;
#endif
}

static void
LocalReleaseSubwindows(YWindow *window, YWindow **subwindows, int length)
{
  int i;

  for (i = 0; i < length; i ++)
	{
	  free(subwindows[i]);
	}

  free(subwindows);
}
