/* -*- Mode: C; tab-width: 4 -*-
   main.c -- the initialization of the microserver
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ymicro.h"
#if notyet
#include "os.h"
#endif
#include "ytypes.h"
#include "hw.h"

#include "wintree.h"

#if notyet
#include "gccache.h"
#endif
#include "yshm.h"
#include "ylib_p.h"
#include "ymessages.h"
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define OsSignal signal

extern int _y_sock;
extern int _y_num_fds;
extern fd_set _y_readfds;

void
abort()
{
  kill(SIGBUS, getpid());
  exit (1);
}

void
cleanupAndExit()
{
  OsSignal(SIGINT,	SIG_DFL);
  OsSignal(SIGTERM,	SIG_DFL);
  OsSignal(SIGSEGV,	SIG_DFL);
  OsSignal(SIGBUS,	SIG_DFL);

  Y_cleanupShmStuff();
  YHW_DeinitDisplay();
  exit(0);
}

static void sig_int_handler()
{
  printf ("* caught SIGINT *\n");
  cleanupAndExit();
}

static void sig_term_handler()
{
  printf ("* caught SIGTERM *\n");
  cleanupAndExit();
}

static void sig_segv_handler()
{
  printf ("* caught SIGSEGV *\n");
  cleanupAndExit();
}

static void sig_bus_handler()
{
  printf ("* caught SIGBUS *\n");
  cleanupAndExit();
}

void
main_loop()
{
  printf ("Inside main_loop\n");

  while (1) 
    {
	  y_addInputFDsToFDSet(&_y_readfds, &_y_num_fds);

	  YHW_HandleEvents();

	  select(256, &_y_readfds, NULL, NULL, NULL);

	  y_dispatchInputForFDSet(&_y_readfds);
    }
}

int
main(int argc,
     char **argv)
{
  int i;
  YBool driver_set = Y_FALSE;
  YBool probe_only = Y_FALSE;

  /* make sure the /tmp/Ylocal directory is there. */
  ensure_dir_exists ("");
  ensure_dir_exists ("shm");

  for (i = 1; i < argc; i ++)
	{
	  if (argv[i][0] != '-')
		{
		  printf ("Malformed command line option %s, ignored\n", argv[i]);
		}
	  else if (!strcmp(argv[i], "-probeonly")
			   || !strcmp(argv[i], "--probeonly"))
		{
		  probe_only = Y_TRUE;
		}
	  else if (!strcmp(argv[i], "-hw")
			   || !strcmp(argv[i], "-hardware")
			   || !strcmp(argv[i], "--hw")
			   || !strcmp(argv[i], "--hardware"))
		{
		  if (i == argc - 1)
			{
			  printf ("%s switch requires an argument\n", argv[i]);
			}
		  else
			{
			  char *driver = argv[i + 1];
			  
			  YHW_LoadHardware(driver);

			  driver_set = Y_TRUE;

			  i++;
			}
		}
	}

#if notyet
  /* make any special OS initializations required. */
  Y_initOS();
#endif
  
  OsSignal(SIGINT,	sig_int_handler);
  OsSignal(SIGTERM,	sig_term_handler);
  OsSignal(SIGSEGV,	sig_segv_handler);
  OsSignal(SIGBUS,	sig_bus_handler);

  if (!probe_only)
	{
	  /* initialize the shared memory stuff, including our semaphores */
	  Y_initShmStuff();
	  
	  /* initialize the window tree data structures */
	  Y_initWindowTree();
	}

  /* probe for hardware (if the user hasn't set it with command line flags
	 or if none of them could be found), and initialize the HW vtable. */
  if (!driver_set || YHW_GetDevices() == NULL)
	YHW_FindHardware(probe_only);

  if (probe_only)
	return 0;

  /* initialize the graphics hw */
  YHW_InitDisplay(argc, argv);

  main_loop();

  cleanupAndExit();
  return 0;
}
