/* -*- Mode: C; tab-width: 4 -*-
   wintree.c -- window tree data structures.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "ymicro.h"
#include "ylib_p.h"
#include "wintree.h"
#include "yshm.h"
#include "ymessages.h"
#include "hw.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include <sys/socket.h>
#include <sys/un.h>

/** XXXX this must match the definition in ydsploc.c */

typedef struct LocalDisplayData {
  int sock;
} LocalDisplayData;

/* special pointer the toplevel window on a display. */
YWindow *root_window;

void
Y_initWindowTree()
{
  YSharedWindowInfo *root_info;
  int sem_id;

  Y_createRootWindow(0);

  root_info = YSHM_getSharedWindowInfo(root_window->shared_window);
  sem_id = YSHM_lockSemaphore(root_info->sem_id);

  root_info->bounds.left = root_info->bounds.top = 0;
  root_info->bounds.bottom = root_info->bounds.right = 1000; /* XXX */

  YSHM_unlockSemaphore(root_info->sem_id, sem_id);
  YSHM_releaseSharedWindowInfo(root_info);
}

void
Y_mapRootWindow()
{
}

YWindow *
Y_getRootWindow()
{
  return root_window;
}

extern YDisplay_vtable yLocalDisplay_vtable;
extern YWindow_vtable yLocalWindow_vtable;

extern int _y_sock;

void
Y_createRootWindow(YWindow_id parent_id)
{
  YSharedWindowInfo *info;
  LocalDisplayData *data = (LocalDisplayData*)calloc(1,
													 sizeof(LocalDisplayData));

  root_window = (YWindow*)calloc(1, sizeof(YWindow));

  root_window->display = (YDisplay*)calloc(1, sizeof(YDisplay));
  root_window->display->vtable = yLocalDisplay_vtable;

  data->sock = _y_sock;
  root_window->display->connection_data = data;

  root_window->vtable = yLocalWindow_vtable;

  root_window->shared_window = YSHM_allocateSharedWindowInfo();
  info = YSHM_getSharedWindowInfo(root_window->shared_window);

  info->parent_id = parent_id;

  YSHM_releaseSharedWindowInfo(info);
}

static void
send_event_to_client(Y_Event *event,
					 int32 client)
{
  struct sockaddr_un client_addr;
  int clientlen;

  printf ("sending event to client %d\n", client);

  memset((char*)&client_addr, 0, sizeof(client_addr));
  client_addr.sun_family = AF_UNIX;
  sprintf(client_addr.sun_path, Y_CLIENT_SOCK_PATH, client);
  clientlen = sizeof(struct sockaddr_un);
  
  Y_sendMessage(_y_sock, event,
				&client_addr, clientlen);
}

static YGC *root_gc = NULL;

static YWindow_id
get_deepest_window_for_point(YWindow_id rooted_at,
							 int32 x, int32 y)
{
  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(rooted_at);
  int id = YSHM_lockSemaphore(info->sem_id);
  YWindow_id return_value = (YWindow_id)~0;
  YWindow_id cur_child;
  YPoint point;

  point.x = x;
  point.y = y;

  cur_child = info->top_child;

  while (cur_child)
	{
	  YSharedWindowInfo *child_info =
		YSHM_getSharedWindowInfo(cur_child);
	  int child_id = YSHM_lockSemaphore(child_info->sem_id);
	  YWindow_id next_child = child_info->next_sibling;

	  if (YIsPointInRect(&point,
						 &child_info->bounds))
		{
		  return_value =
			get_deepest_window_for_point(cur_child,
										 x - child_info->bounds.left,
										 y - child_info->bounds.top);

		  if (return_value == (YWindow_id)~0)
			return_value = cur_child;
		  
		  next_child = 0;
		}

	  YSHM_unlockSemaphore(child_info->sem_id, child_id);
	  YSHM_releaseSharedWindowInfo(child_info);

	  cur_child = next_child;
	}

  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);

  return return_value;
}

static void
notify_interested(Y_Event *event)
{
  YWindow_id target = event->any_event.target;
  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(target);
  int id = YSHM_lockSemaphore(info->sem_id);
  int32 *clients;
  int num;

  num = info->num_interested;

  if (num)
	{
	  int i;

	  clients = (int32*)calloc(num, sizeof(int32));
	  memcpy(clients, info->interested_parties, num * sizeof(int32));

	  for (i = 0; i < num; i ++)
		send_event_to_client(event, clients[i]);

	  free(clients);
	}

  YSHM_unlockSemaphore (info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);

}

static void
handle_exposure_recurse(YWindow_id target, Y_Event *event)
{
  YRect *intersection;
  YSharedWindowInfo *info = YSHM_getSharedWindowInfo(target);
  int id;

  id = YSHM_lockSemaphore(info->sem_id);

  intersection = YRectIntersect(&info->bounds, &event->expose.bounds);

  if (!intersection)
	{
	  /* the exposure rectangle doesn't touch this window.  So, it can't touch
		 any of this window's children either, so we just return. */
	  goto done;
	}
  else
	{
	  /* the exposure rectangle intersects this window's bounds.  So, we save
		 off the original exposure rectangle and replace it with the intersection,
		 then call notify_interested().  We then revert the exposure bounds and
		 loop over this window's children, calling this function on each of them. */
	  YWindow_id cur_child = info->top_child;
	  YRect event_rect = event->expose.bounds;
	  
	  event->expose.bounds = *intersection;
	  free(intersection);

	  event->expose.target = target;

	  notify_interested(event);

	  event->expose.bounds = event_rect;

	  while(cur_child)
		{
		  YSharedWindowInfo *child_info =
			YSHM_getSharedWindowInfo(cur_child);
		  int child_id = YSHM_lockSemaphore(child_info->sem_id);
		  YWindow_id next_child = child_info->next_sibling;

		  handle_exposure_recurse(cur_child, event);

		  YSHM_unlockSemaphore(child_info->sem_id, child_id);
		  YSHM_releaseSharedWindowInfo(child_info);

		  cur_child = next_child;
		}
	}

 done:
  YSHM_unlockSemaphore(info->sem_id, id);
  YSHM_releaseSharedWindowInfo(info);
}

static void
handle_expose_event(Y_Event *event)
{
  YWindow_id target = event->any_event.target;

  if (target == (YWindow_id)~0)
	{
	  handle_exposure_recurse(root_window->shared_window, event);
	}
  else
	{
	  /* the target is specified in the event already, it came from a client
		 and just needs to be sent along. */

	  notify_interested(event);
	}
}

void
Y_handleEvent(Y_Event *event)
{
  int32 x;
  int32 y;
  YWindow_id target;

  if (root_gc == NULL)
	root_gc = YGetGC(root_window);

  switch (event->any_event.tag)
	{
	case Y_ExposeEvent:
	  printf ("(Y_ExposeEvent)\n");

	  YSetForeground(root_gc, YGetColorRGB(40, 40, 40));
	  
	  YFillRectangle(root_gc, &event->expose.bounds);

	  /* special things are needed for expose events... */
	  handle_expose_event(event);
	  return;
	case Y_KeyDownEvent:
	  printf ("(Y_KeyDownEvent ");
	  x = event->key_down.x;
	  y = event->key_down.y;
	  break;
	case Y_KeyUpEvent:
	  printf ("(Y_KeyUpEvent ");
	  x = event->key_up.x;
	  y = event->key_up.y;
	  break;
	case Y_MouseDownEvent:
	  printf ("(Y_MouseDownEvent (%d) ",
			  event->mouse_down.button);
	  x = event->mouse_down.x;
	  y = event->mouse_down.y;

	  break;
	case Y_MouseUpEvent:
	  printf ("(Y_MouseUpEvent (%d) ",
			  event->mouse_up.button);
	  x = event->mouse_up.x;
	  y = event->mouse_up.y;
	  break;
	case Y_MouseMotionEvent:
	  printf ("(Y_MouseMotionEvent)\n");
	  x = event->mouse_motion.x;
	  y = event->mouse_motion.y;
	  break;
	}

  target = get_deepest_window_for_point(root_window->shared_window, x, y);

  if (target == (YWindow_id)~0)
	target = root_window->shared_window;

  printf (" [%d,%d], target = %d\n", x, y, target);

  event->any_event.target = target;

  notify_interested(event);

  /* do we need this?? */
#if 0
  YNotifyInterested(root_window, event);
#endif
}
