/* -*- Mode: C; tab-width: 4 -*-
   shm.c -- shared memory code for the microserver.
   Created: Chris Toshok <toshok@hungry.com>, 8-May-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include <stdio.h>

#include "hw.h"
#include "yshm.h"
#include "ymicro.h"
#include "ymessages.h"
#include "ylib_p.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

int _y_sock = -1;
int _y_num_fds;
fd_set _y_readfds;

/* Keep in sync with the function definition further down. */
void Y_handleRequest(int, Y_Request *, struct sockaddr_un *, int clientaddrlen);

static void
_y_handleEvent(void *arg)
{
  Y_Request request;
  struct sockaddr_un clientaddr;
  int clientaddrlen;

  if (Y_getMessage(_y_sock, &request, &clientaddr, &clientaddrlen) < 0)
	{
	  printf ("huh... got a -1 response.\n");
	  return;
	}

  printf ("Got a request, handling it.\n");
  Y_handleRequest(_y_sock, &request, &clientaddr, clientaddrlen);
}

static void
init_socket()
{
  struct sockaddr_un serv_addr;
  int servlen;

  printf ("Inside init_msg_queue()\n");

  memset((char *)&serv_addr, '\0', sizeof(serv_addr));
  serv_addr.sun_family = AF_UNIX;
  strcpy(serv_addr.sun_path, Y_SERVER_SOCK_PATH);
  servlen = sizeof(struct sockaddr_un);

  if (mkdir(Y_SERVER_DIR, 0700) == -1)
	{
	  if (errno != EEXIST)
		{
		  perror("init_socket(mkdir)");
		  cleanupAndExit();
		}
	}

  _y_sock = socket(AF_UNIX, SOCK_DGRAM, 0);

  if (_y_sock < 0)
    {
	  perror("init_socket(socket)");
	  cleanupAndExit();
    }

  /* make sure the socket isn't there from a previous microserver. */
  unlink(Y_SERVER_SOCK_PATH);
  if (bind(_y_sock, (struct sockaddr *) &serv_addr, servlen) < 0)
	{
	  perror("init_socket(bind)");
	  cleanupAndExit();
	}

  printf("socket initialized (_y_sock = %d)\n", _y_sock);

  YAddInput(_y_sock, _y_handleEvent, NULL);
}

static void
destroy_socket()
{
  if (_y_sock != -1)
	{
	  YRemoveInput(_y_sock, _y_handleEvent, NULL);
	  close(_y_sock);
	  unlink(Y_SERVER_SOCK_PATH);
	}
}

void
Y_initShmStuff()
{
  printf ("Initializing shared memory stuff.\n");

  init_socket();
}

void
Y_cleanupShmStuff()
{
  printf ("Cleaning up shared memory stuff.\n");
  
  destroy_socket();
  YHW_CleanupSharedMemory();
}

int
Y_getMessage(int sock, Y_Message *request,
			 struct sockaddr_un *from, int *addrlen)
{
  *addrlen = sizeof(struct sockaddr_un);
  return recvfrom(sock, request, sizeof(Y_Message),
				  0, (struct sockaddr*)from, addrlen);
}

int
Y_sendMessage(int sock, Y_Message *reply, struct sockaddr_un *to, int addrlen)
{
  return sendto(sock, reply, sizeof(Y_Message),
				0, (struct sockaddr*)to, addrlen);
}

/* needed for the GetRootWindow request */
extern YWindow *root_window;

void
Y_handleRequest(int sock,
				Y_Request *request,
				struct sockaddr_un *client,
				int clientaddrlen)
{
  if (request->tag > Y_LastRequest) /* it's actually an event */
	{
	  printf ("It's actually an X request\n");
	  Y_handleEvent((Y_Event*)request);
	}
  else
	{
	  printf ("It's actually an Y request ");
	  switch (request->tag)
		{
		case Y_GetRootWindowReq:
		  {
			Y_Reply reply;

			printf ("(Y_GetRootWindowReq)\n");
			reply.any_rep.tag = Y_GetRootWindowRep;
			reply.getroot_rep.root_id = root_window->shared_window;

			Y_sendMessage(sock, &reply, client, clientaddrlen);
		  }
		  break;
		case Y_GetDriverReq:
		  {
			Y_Reply reply;
			int num_devices = YHW_GetNumDevices();
			YHW *hw = YHW_GetDevices();

			printf ("(Y_GetDriverReq)\n");
			reply.any_rep.tag = Y_GetDriverRep;

			reply.getdriver_rep.first_id = hw->shm_id - num_devices + 1;
			reply.getdriver_rep.num_ids = num_devices;

			Y_sendMessage(sock, &reply, client, clientaddrlen);
		  }
		  break;
		default:
		  printf ("(Unhandled request)\n");
		  abort();
		}
	}
}
