/* -*- Mode: C; tab-width: 4 -*-
   client.c -- stuff dealing with the clients of the network server.
   Created: Chris Toshok <toshok@hungry.com>, 21-Sep-1997
 */
/* 
   Copyright (C) 1996, 1997 The Hungry Programmers.
   
   This file is part of The Y Window System.
   
   Y is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
   
   Y is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with GLD; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307 , USA.
*/

#include "client.h"
#include "ylib_p.h"
#include "ynet_p.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#if defined(linux) || defined(hpux)
#include <netinet/in.h>
#endif


typedef struct YWindow_cons {
  struct YWindow_cons *prev;
  struct YWindow_cons *next;

  YWindow *win;
} YWindow_cons;

typedef struct YGC_cons {
  struct YGC_cons *prev;
  struct YGC_cons *next;
  
  YGC *gc;
} YGC_cons;

typedef struct YNetClient {
  struct YNetClient *prev;
  struct YNetClient *next;

  int client_fd;

  YWindow_cons *window_list;
  YGC_cons *gc_list;

  YGC_id next_gc_id;

} YNetClient;

static YNetClient *client_list;

extern YDisplay *_ydisplay;

static YNetClient *
find_client_by_fd(int fd)
{
  YNetClient *cur;

  for (cur = client_list;
	   cur != NULL;
	   cur = cur->next)
	{
	  if (cur->client_fd == fd)
		return cur;
	}

  return NULL;
}

void 
ynet_addClient(int client)
{
  YNetClient *new_client = (YNetClient*)calloc(1, sizeof(YNetClient));

  new_client->client_fd = client;

  /* link the new client into the client_list */
  new_client->next = client_list;
  if (client_list)
	client_list->prev = new_client;
  client_list = new_client;
}

void 
ynet_clientHasDied(int fd)
{
  YNetClient *client = find_client_by_fd(fd);
  YWindow_cons *win_cons;
  YGC_cons *gc_cons;

  if (!client)
	{
	  fprintf (stderr, "Nonexistent client dying.... ?\n");
	  return;
	}

  if (client->next)
	client->next->prev = client->prev;
  if (client->prev)
	client->prev->next = client->next;

  if (client_list == client)
	client_list = client->next;

  win_cons = client->window_list;
  while (win_cons)
	{
	  YWindow_cons *next_cons;

	  next_cons = win_cons->next;

	  YDestroyWindow(win_cons->win);
	  free(win_cons);

	  win_cons = next_cons;
	}

  gc_cons = client->gc_list;
  while (gc_cons)
	{
	  YGC_cons *next_cons;

	  next_cons = gc_cons->next;

	  YDestroyGC(gc_cons->gc);
	  free(gc_cons);

	  gc_cons = next_cons;
	}

  close(client->client_fd);

  free(client);
}

static void 
ynet_addWindowToClient(YWindow *window,
					   YNetClient *client)
{
}

static YGC_id
ynet_addGCToClient(YGC *gc,
				   YNetClient *client)
{
  YGC_cons *cons = (YGC_cons*)calloc(1, sizeof(YGC_cons));

  cons->next = client->gc_list;
  if (client->gc_list)
	client->gc_list->prev = cons;
  client->gc_list = cons;

  cons->gc = gc;
  cons->gc->gc_id = client->next_gc_id ++;

  return cons->gc->gc_id;
}

static YGC*
find_gc_by_id(YNetClient *client,
			  YGC_id gc_id)
{
  YGC_cons *cur;

  for (cur = client->gc_list;
	   cur != NULL;
	   cur = cur->next)
	if (cur->gc->gc_id == gc_id)
	  return cur->gc;

  return NULL;
}

void
ynet_addClientsToFDSet(fd_set *fds)
{
  YNetClient *cur;
  
  for (cur = client_list;
	   cur != NULL;
	   cur = cur->next)
	FD_SET(cur->client_fd, fds);
}

static YBool
handle_get_root(YNetClient *client,
				YNetRequest *request,
				YNetReply *reply)
{
  YWindow *root = YRootWindowOfDisplay(_ydisplay);
  reply->tag = GET_ROOT_WIN;
  reply->get_root.win_id = htonl(root->shared_window);

  return Y_TRUE;
}

static YBool
handle_create_win(YNetClient *client,
				  YNetRequest *request,
				  YNetReply *reply)
{
  YWindow *parent;
  YWindow *new_window;

  parent = YGetWindowById(_ydisplay, ntohl(request->create_win.parent_id));
  new_window = YCreateWindow(_ydisplay, parent,
							 ntohl(request->create_win.x),
							 ntohl(request->create_win.y),
							 ntohl(request->create_win.width),
							 ntohl(request->create_win.height),
							 ntohl(request->create_win.bg_color));

  ynet_addWindowToClient(new_window,
						 client);

  reply->tag = CREATE_WIN;
  reply->create_win.win_id = htonl(new_window->shared_window);

  return Y_TRUE;
}

static YBool
handle_map_win(YNetClient *client,
			   YNetRequest *request,
			   YNetReply *reply)
{
  YWindow *win = YGetWindowById(_ydisplay, ntohl(request->map_win.win_id));
  YMapWindow(win);

  return Y_FALSE;
}

static YBool
handle_get_gc(YNetClient *client,
			  YNetRequest *request,
			  YNetReply *reply)
{
  YWindow *win = YGetWindowById(_ydisplay, ntohl(request->get_gc.win_id));
  YGC *gc;
  YGC_id gc_id;

  gc = YGetGC(win);

  gc_id = ynet_addGCToClient(gc, client);

  reply->tag = GET_GC;
  reply->get_gc.gc_id = htonl(gc_id);

  return Y_TRUE;
}

static YBool
handle_draw_line(YNetClient *client,
				 YNetRequest *request,
				 YNetReply *reply)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->draw_line.gc_id));
  YPoint p1, p2;

  p1.x = ntohl(request->draw_line.x1);
  p1.y = ntohl(request->draw_line.y1);
  p2.x = ntohl(request->draw_line.x2);
  p2.y = ntohl(request->draw_line.y2);

  YDrawLine(gc, &p1, &p2);

  return Y_FALSE;
}

static void
handle_copy_to_win(YNetClient *client,
				   YNetRequest *request)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->copy_to.gc_id));
  YColor *buffer;
  YPoint p1;
  YRect size;

  p1.x = ntohl(request->copy_to.dest_x);
  p1.y = ntohl(request->copy_to.dest_y);

  size.right = ntohl(request->copy_to.width);
  size.bottom = ntohl(request->copy_to.height);

  buffer = (YColor*)malloc(request->copy_to.buffer_size);
  buffer[0] = request->copy_to.buffer[0];

  read(client->client_fd,
	   &(buffer[1]),
	   (request->copy_to.buffer_size / sizeof(YColor)) - 1);

  (*gc->vtable.CopyToWin)(gc, &p1, &size, buffer);

  free(buffer);
}

static void
handle_copy_from_win(YNetClient *client,
					 YNetRequest *request)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->copy_from.gc_id));
  YPoint p1;
  YRect size;
  YWindow *src;
  uint32 buffer_size;
  YNetCopyFromWinRep *copy_from_rep;
  uint32 reply_size;

  src = gc->window;

  p1.x = ntohl(request->copy_from.src_x);
  p1.y = ntohl(request->copy_from.src_y);

  size.right = ntohl(request->copy_from.width);
  size.bottom = ntohl(request->copy_from.height);

  buffer_size = size.right * size.bottom * sizeof(YColor);

  reply_size = (sizeof(YNetCopyFromWinRep) + buffer_size 
				- sizeof(YColor) /* to get rid of the YColor[1]
									in YNetCopyFromWinRep */);

  copy_from_rep = (YNetCopyFromWinRep*)malloc(reply_size);

  copy_from_rep->tag = COPY_FROM_WIN;
  copy_from_rep->buffer_size = buffer_size;
  (*gc->vtable.CopyFromWin)(gc, &p1, &size, (YColor**)&copy_from_rep->buffer);

  write(client->client_fd,
		&copy_from_rep,
		reply_size);
}

static void
handle_copy_from_win_to_win(YNetClient *client,
							YNetRequest *request)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->copy_fromto.gc_id));
  YPoint p1, p2;
  YRect size;
  YWindow *src;

  src = YGetWindowById(_ydisplay, ntohl(request->copy_fromto.src_id));
  
  p1.x = ntohl(request->copy_fromto.src_x);
  p1.y = ntohl(request->copy_fromto.src_y);
  p2.x = ntohl(request->copy_fromto.dest_x);
  p2.y = ntohl(request->copy_fromto.dest_y);

  size.right = ntohl(request->copy_fromto.width);
  size.bottom = ntohl(request->copy_fromto.height);

  YWinBlit(gc, src, &p1, &p2, &size);
}

static void
handle_draw_ascii_text(YNetClient *client,
					   YNetRequest *request)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->draw_ascii.gc_id));
  YPoint p1;
  char *txt;

  txt = (char*)malloc(request->draw_ascii.txt_length);
  txt[0] = request->draw_ascii.txt[0];

  read(client->client_fd,
	   &(txt[1]),
	   request->draw_ascii.txt_length - 1);
  
  p1.x = ntohl(request->draw_ascii.x);
  p1.y = ntohl(request->draw_ascii.y);

  YDrawAsciiText(gc, &p1, txt);
}

static YBool
handle_set_foreground(YNetClient *client,
					  YNetRequest *request,
					  YNetReply *reply)
{
  YGC *gc = find_gc_by_id(client, ntohl(request->set_fore.gc_id));

  YSetForeground(gc, ntohl(request->set_fore.foreground));

  return Y_FALSE;
}

static void
RECV_REQUEST(int fd, YNetRequest *request)
{
  int num_remaining;
  char *buf;

  num_remaining = sizeof(YNetRequest); 
  buf = (char*)request;

  while (num_remaining)
	{
	  int errval = read(fd,
						buf,
						num_remaining);
	  
	  if (errval < 0) 
		{
		  perror ("ynet_handleClientInput(read)");
		  abort();
		}

	  num_remaining -= errval;
	  buf += errval;
	}
}


static void
ynet_handleClientInput(YNetClient *client)
{
  YNetRequest request;
  YNetReply reply;
  YBool send_reply = Y_FALSE;

  RECV_REQUEST(client->client_fd, &request);

  switch(request.tag)
	{
	case GET_ROOT_WIN:
	  send_reply = handle_get_root(client, &request, &reply);
	  break;
	case CREATE_WIN:
	  send_reply = handle_create_win(client, &request, &reply);
	  break;
	case MAP_WIN:
	  send_reply = handle_map_win(client, &request, &reply);
	  break;
	case GET_GC:
	  send_reply = handle_get_gc(client, &request, &reply);
	  break;
	case DRAW_LINE:
	  send_reply = handle_draw_line(client, &request, &reply);
	  break;
	case SET_FOREGROUND:
	  send_reply = handle_set_foreground(client, &request, &reply);
	  break;
	case COPY_FROM_WIN_TO_WIN:
	  handle_copy_from_win_to_win(client, &request);
	  send_reply = Y_FALSE;
	  break;
	case COPY_FROM_WIN:
	  handle_copy_from_win(client, &request);
	  send_reply = Y_FALSE;
	  break;
	case COPY_TO_WIN:
	  handle_copy_to_win(client, &request);
	  send_reply = Y_FALSE;
	  break;
	case DRAW_ASCII_TEXT:
	  handle_draw_ascii_text(client, &request);
	  break;
	default:
	  assert(0);
	}

  if (send_reply)
	write(client->client_fd,
		  &reply,
		  sizeof (YNetReply));
}

void
ynet_handleClientInputFromFDs(fd_set *fds)
{
  YNetClient *cur;

  for (cur = client_list;
	   cur != NULL;
	   cur = cur->next)
	{
	  if (FD_ISSET(cur->client_fd, fds))
		ynet_handleClientInput(cur);
	}
}
